#ifdef ROCKALL_DB_EXAMPLES
#define main Boggle
#endif
	//
	//  Header files.
	//
	//  The header files provide access to functionality in the
	//  related classes.
	//
#include <ctype.h>
#include <stdio.h>
#include <string.h>

#include "AutomaticHeapScope.hpp"
#include "FindWords.hpp"
#include "FixedString.hpp"
#include "MultiThreadedHeap.hpp"

	//
	//  Private types.
	//
	//  The private types are local to the current class.
	//
static const int MaxTextSize = 128;

typedef FIXED_STRING<MULTI_THREADED_HEAP,char,MaxTextSize> STRING;

	//
	//  Read Board.
	//
	//  Read the 'Boggle' board.
	//
bool ReadBoard( char **NewBoard,int *NewHeight,int *NewWidth )
	{
	AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> Scope;
#ifdef ROCKALL_DB_EXAMPLES
	char *Board;

	//
	//  We start by allocating a small amount of space 
	//  to the 'Board'.
	//
	if ( Scope.New( ((void**) & Board),1 ) )
		{
		int Height = 0;
		int Width = 0;

		//
		//  We cycle round reading the in the 'Board'.
		//
		while ( true )
			{
			static const int MaxInputSize = 4096;
			char Text[ MaxInputSize ];

			//
			//  We read the next line of text.
			//
			if ( fgets( Text,MaxInputSize,stdin ) != NULL )
				{
				int Length = ((int) (strlen( Text )-1));
				int Count;

				//
				//  We do not know the 'Length' of each line
				//  on the 'Board' so we assume it is the 'Length'
				//  of the first line of the 'Board'.
				//
				if ( Width == 0 )
					{ Width = Length; }
				else
					{
					//
					//  We ensure that the 'Length' of all the 
					//  lines is the same.
					//
					if ( (Length != 0) && (Length != Width) )
						{ printf( "Error: All lines must have the same number of letters.\n" ); break; }
					}

				//
				//  We process each line unless it is empty in
				//  which case we can exit.
				//
				if ( Length > 0 )
					{
					//
					//  We ensure all characters are lower case 
					//  and fail if we find any non-alphabetic
					//  characters.
					//
					for ( Count=0;Count < Length;Count ++ )
						{
						char *Current = & Text[ Count ];

						//
						//  Convert any upper case characters
						//  to lower case.
						//
						if ( isupper( (*Current) ) )
							{ (*Current) = tolower( (*Current) ); }

						//
						//  We reject any non-alphabetic characters.
						//
						if ( ((*Current)) < 'a' || ((*Current) > 'z') )
							{
							printf
								( 
								"Warning: Skipping non-alphabetic character '%c'.\n",
								(*Current) 
								);
							}
						}

					//
					//  We 'Resize()' the 'Board' to the correct
					//  size and copy in the new line.
					//
					if ( Scope.Resize( ((void**) & Board),((Height+1) * Width) ) )
						{ strncpy( & Board[ ((Height ++) * Width) ],Text,Width ); }
					else
						{ printf( "Error: Unable to read the 'Boggle' board.\n" ); break; }
					}
				else
					{
					//
					//  We update the parameters and exit.
					//
					(*NewBoard) = Board;
					(*NewHeight) = Height;
					(*NewWidth) = Width;

					return true; 
					}
				}
			else
				{ printf( "Error: Unable to read a line of text.\n" ); break; }
			}

		//
		//  We need to delete the 'Board'.
		//
		Scope.Delete( Board );
		}
	else
		{ printf( "Error: Unable to read the 'Boggle' board.\n" ); }
#else
	static char *TestString = 
		{ 
		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"

		"abcdefghijklmnopqrstuvwxyz" "bcdefghijklmnopqrstuvwxyza"
		"cdefghijklmnopqrstuvwxyzab" "defghijklmnopqrstuvwxyzabc"
		"efghijklmnopqrstuvwxyzabcd" "fghijklmnopqrstuvwxyzabcde"
		"ghijklmnopqrstuvwxyzabcdef" "hijklmnopqrstuvwxyzabcdefg"
		"ijklmnopqrstuvwxyzabcdefgh" "jklmnopqrstuvwxyzabcdefghi"
		"klmnopqrstuvwxyzabcdefghij" "lmnopqrstuvwxyzabcdefghijk"
		"mnopqrstuvwxyzabcdefghijkl" "nopqrstuvwxyzabcdefghijklm"
		"opqrstuvwxyzabcdefghijklmn" "pqrstuvwxyzabcdefghijklmno"
		"qrstuvwxyzabcdefghijklmnop" "rstuvwxyzabcdefghijklmnopq"
		"stuvwxyzabcdefghijklmnopqr" "tuvwxyzabcdefghijklmnopqrs"
		"uvwxyzabcdefghijklmnopqrst" "vwxyzabcdefghijklmnopqrstu"
		"wxyzabcdefghijklmnopqrstuv" "xyzabcdefghijklmnopqrstuvw"
		"yzabcdefghijklmnopqrstuvwy" "zabcdefghijklmnopqrstuvwxy"
		};
	int Length = ((int) (strlen( TestString )));
	int MaxSize = ((Length * Length) + 1);

	//
	//  We allocate space for the 'Board'. 
	//
	if ( Scope.New( ((void**) NewBoard),MaxSize ) )
		{
		int Count;

		//
		//  We update the sizes.
		//
		(*NewHeight) = Length;
		(*NewWidth) = Length;

		//
		//  We copy the 'TestString' to the 'Board'.
		//
		for ( Count=0;Count < Length;Count ++ )
			{
			int Current = (Count * Length);

			strcpy( & (*NewBoard)[ Current ],TestString );
			}

		(*NewBoard)[ (MaxSize-1) ] = '\0';

		return true;
		}
#endif

	return false;
	}

	//
	//  Unique Words
	//
	//  All the unique words are passed to this function
	//  for processing.
	//
static bool UniqueWords( STRING & String,short & Score,void *UserData )
	{
	char *Text;

	//
	//  We ned to view the 'String'.
	//
	if ( String.View( & Text ) )
		{
		//
		//  Print out the string and update the 'Score'.
		//
		printf( "Score %d for '%s'\n",Score,Text );

		(*((int*) UserData)) += Score;

		String.EndView();

		return true;
		}

	return false;
	}

	//
	//  Main.
	//
	//  The main function simply processes the command line
	//  options and calls the approriate functions.
	//
int main( int Count,char *Argument[] )
	{
	MULTI_THREADED_HEAP Heap;

	//
	//  We process a request to build a new directory.
	//
	if ( (Count == 4) && (strcmp( "-c",Argument[1] ) == 0) )
		{
		AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> Scope( & Heap );
		BOGGLE::FIND_WORDS FindWords;

		//
		//  We try to build a new database.
		//
		if ( FindWords.CreateFindWords( Argument[2],Argument[3] ) )
			{
			FindWords.CloseFindWords();

			return 0;
			}
		else
			{
			printf
				( 
				"Error: Unable to build database '%s' from '%s'\n",
				Argument[2], 
				Argument[3] 
				);

			return 1;
			}
		}

	//
	//  We process a request to play the game.
	//
	if ( (Count == 3) && (strcmp( "-p",Argument[1] ) == 0) )
		{
		AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> Scope( & Heap );
		BOGGLE::FIND_WORDS FindWords;

		//
		//  We try to load the database.
		//
		if ( FindWords.LoadFindWords( Argument[2] ) )
			{
			char *Board;
			int Height;
			int Width;

			//
			//  Output the description of how to use the program.
			//
			printf
				(
				"\n"
				"Please type in the 'Boggle' board.  All you need to type in is the letters\n"
				"on each line of the board without any spaces.  The letters can be in either\n"
				"capitals or lower case.  A new line should be typed at the end of each\n" 
				"row of the board.  All the rows should contain the same number of letters.\n"
				"When you have typed in all the rows on the board simply enter a blank line\n"
				"to denote the end of the board.\n"   
				"\n" 
				"Please enter a 'Boggle' board now:\n" 
				"\n" 
				);

			//
			//  We now read the new 'Board'.
			//
			if ( ReadBoard( & Board,& Height,& Width ) )
				{
				int Score = 0;

				//
				//  Output the size of the 'Board'.
				//
				printf
					( 
					"The 'Boggle' board size is %d x %d:\n"
					"\n"
					"The 'Boggle' words are:\n"
					"\n",
					Height,
					Width
					);

				//
				//  We can now compute all the words.
				//
				if 
						(
						(FindWords.FindAllWords( Board,Height,Width,UniqueWords,& Score ))
							&&
						(Score > 0) 
						)
					{ printf( "\nThe total score is %d.\n",Score ); }
				else
					{ printf( "*** No words found. The score is zero ***.\n" ); }

				//
				//  Delete the 'Board'.
				//
				Heap.Delete( Board );

				//
				//  Close the database.
				//
				FindWords.CloseFindWords();

				return 0;
				}
			else
				{
				//
				//  Close the database.
				//
				FindWords.CloseFindWords();

				return 1;
				}	
			}
		else
			{
			printf
				( 
				"Error: Unable to load dictionary from '%s'\n",
				Argument[2] 
				);

			return 1;
			}
		}

	printf( "Usage: Boggle [-c <Database Name> <New Dictionary Name> | -p <Database Name>]\n" );

	return 1;
	}
