#ifndef _DATA_STORE_HPP_
#define _DATA_STORE_HPP_ 1

    //
	//  Header files.
	//
	//  The header files provide access to functionality in the
	//  Rockall-DB library and other classes.
	//
#include "DatabaseHeap.hpp"
#include "RockallTypes.hpp"

	//
	//  The 'BOOGLE' namespace contains all the related code.
	//
namespace BOGGLE
	{
	//
	//  Public classes.
	//
	//  The public classes are part of the interface for the
	//  'DATA_STORE' class.
	//
	typedef struct
		{
		short								Character;
		short								Score;
		int									State;
		}
	NEXT_STATE;

	typedef struct
		{
		int									Start;
		int									End;
		}
	STATE;

	typedef struct
		{
		int									MaxNextStates;
		int									MaxStates;
		int									MaxWordSize;

		NEXT_STATE							*NextStates;
		STATE								*States;
		}
	STATE_TABLE;

	//
	//  Data Store.
	//
	//  The 'DATA_STORE' class manages all the interactions with
	//  the related dictionary database.
	//
	class DATA_STORE
		{
			//
			//  Private constants.
			//
			static const int				MagicNumber = 0x12345678;

			//
			//  Private data.
			//
			bool							Active;
			bool							Open;

			DATABASE_HEAP					*Database;

		public:
			//
			//  Public functions.
			//
			DATA_STORE( void );

			bool CreateDataStore( char *DatabaseName,STATE_TABLE *StateTable );

			bool LoadDataStore( char *DatabaseName,STATE_TABLE *StateTable );

			bool CloseDataStore( void );

			~DATA_STORE( void );
		};
	}
#endif