#ifndef _DICTIONARY_HPP_
#define _DICTIONARY_HPP_ 1

    //
	//  Header files.
	//
	//  The header files provide access to functionality in the
	//  Rockall-DB library and other classes.
	//
#include "DataStore.hpp"
#include "MemoryLock.hpp"
#include "MultiThreadedHeap.hpp"
#include "Tree.hpp"

	//
	//  The 'BOOGLE' namespace contains all the related code.
	//
namespace BOGGLE
	{
	//
	//  Dictionary.
	//
	//  The 'DICTIONARY' class manages all the interactions with the
	//  related 'STATE_TABLE'.
	//
	class DICTIONARY
		{
			//
			//  Private constants.
			//
			static const int				MinStates = 1;
			static const int				ViewSize = 1024;

			//
			//  Private types.
			//
			typedef long long KEY;

			typedef TREE<MULTI_THREADED_HEAP,MEMORY_LOCK,KEY,NEXT_STATE> STATES;

			typedef struct
				{
				KEY							Character : 16;
				KEY							State : 32;
				}
			NEW_STATE;

			typedef struct
				{
				int							Character;
				NEW_STATE					NewState;
				NEXT_STATE					NextState;
				int							Length;
				}
			CONTEXT;

			typedef struct
				{
				int							MaxNextStates;
				int							MaxStates;

				NEXT_STATE					*NextStates;
				STATE						*States;
				}
			COPY_DATA;

			typedef struct
				{
				int							MaxStates;
				int							MaxWords;
				int							MaxWordSize;
				}
			COUNTERS;

			//
			//  Private data.
			//
			bool							Active;

			DATA_STORE						DataStore;

		public:
			//
			//  Public functions.
			//
			DICTIONARY( void );

			bool CreateDictionary( char *DatabaseName,char *FileName,STATE_TABLE *StateTable );

			bool LoadDictionary( char *DatabaseName,STATE_TABLE *StateTable );

			bool CloseDictionary( void );

			~DICTIONARY( void );

		private:
			//
			//  Private static functions.
			//
			static bool CopyStatesIntoDatabase( KEY & Key,NEXT_STATE & NextState,void *UserData );
		};
	}
#endif