#ifndef _FIND_WORDS_HPP_
#define _FIND_WORDS_HPP_ 1
	//
	//  Header files.
	//
	//  The header files provide access to functionality in the
	//  Rockall-DB library and other classes.
	//
#include "AutomaticHeapScope.hpp"
#include "Dictionary.hpp"
#include "FixedString.hpp"
#include "MemoryLock.hpp"
#include "Tree.hpp"

	//
	//  The 'BOOGLE' namespace contains all the related code.
	//
namespace BOGGLE
	{
		//
		//  Find Words.
		//
		//  The 'FIND_WORDS' class manages all analysis and searching
		//  for words.
		//
	class FIND_WORDS
		{
			//
			//  Private constants.
			//
			static const int				MaxFindTime = ((3 * 60) - 5);
			static const int				MaxTextSize = 128;

			//
			//  Private types.
			//
			typedef AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> SCOPE;

			typedef FIXED_STRING<MULTI_THREADED_HEAP,char,MaxTextSize> STRING;

			typedef TREE<MULTI_THREADED_HEAP,MEMORY_LOCK,STRING,short> UNIQUE;

			typedef struct
				{
				char						*Board;
				FIND_WORDS					*Class;
				int							Height;
				int							MaxThreads;
				bool						*Processing;
				bool						*Result;
				SCOPE						*Scope;
				int							Size;
				char						Text[ MaxTextSize ];
				int							ThreadID;
				UNIQUE						*Unique;
				int							Width;
				}
			CONTEXT;

			typedef bool (*FUNCTION)
				(
				STRING						& String,
				short						& Score,
				VOID						*UserValue
				);

			//
			//  Private data.
			//
			bool							Active;

			DICTIONARY						Dictionary;
			STATE_TABLE						StateTable;

		public:
			//
			//  Public functions.
			//
			FIND_WORDS( void );

			bool CreateFindWords( char *DatabaseName,char *FileName );

			bool LoadFindWords( char *DatabaseName );

			bool FindAllWords( char *Board,int Height,int Width,FUNCTION Function,void *UserData );

			bool CloseFindWords( void );

			~FIND_WORDS( void );

		private:
			//
			//  Private functions
			//
			static void FindAllWordsInParallel( CONTEXT *Context );

			bool FindNearby( CONTEXT *Context,int State,int X,int Y );

			bool FindNextState( CONTEXT *Context,NEXT_STATE **NextState,int State,int X,int Y );

			bool FindSingle( CONTEXT *Context,int State,int X,int Y );

			static void FindTimer( bool **Processing );

		};
	}
#endif
