#ifdef ROCKALL_DB_EXAMPLES
#define main Book
#endif
    //
    //  Header files.
    //
    //  The header files provide access to functionality in the
    //  Rockall-DB library.
    //
#include <stdio.h>
#include <string.h>

#include "Database.hpp"
#include "DatabaseHeap.hpp"
#include "RockallTypes.hpp"

    //
    //  Display Reference.
    //
    //  The display reference function outputs the results to
    //  the user.
    //
void DisplayReference
        (
        FILE_SIZE                           MaxSize,
        char                                *Paragraph,
        FILE_SIZE                           Size1,
        char                                *Title,
        FILE_SIZE                           Size2
        )
    {
    if ( (MaxSize != 0) || (Size1 != 0) || (Size2 != 0) )
        {
        if ( MaxSize > 0 )
            { printf( "Matches: %d matching reference(s) found.\n",((int) MaxSize) ); }

        if ( Size2 > 0 )
            { printf( "\n\nTitle: '%.*s'\n",((int) Size2),Title ); }

        if ( Size1 > 0 )
            { printf( "\nReference:\n'%.*s'\n",((int) Size1),Paragraph ); }
        }
    else
        { printf( "Error: No matching references found.\n" ); }
    }

    //
    //  Display Word.
    //
    //  The display word function outputs the results to the
    //  user.
    //
void DisplayWord( char *Word,FILE_SIZE References,BOOLEAN Newline )
    {
    if ( ! Newline )
        { printf( "%19s%6d ",Word,((int) References) ); }
    else
        { printf( "%19s%6d\n",Word,((int) References) ); }
    }

    //
    //  Main.
    //
    //  The main function simply processes the command line
    //  options and calls the approriate functions.
    //
int main( int Count,char *Argument[] )
    {
    //
    //  We process a request to build a new database.
    //
    if ( (Count == 4) && (strcmp( Argument[1],"-c" ) == 0) )
        {
        AUTO DATABASE Database;

        if ( Database.Create( Argument[2],Argument[3] ) )
            {
            Database.Close();

            return 0;
            }
        else
            {
            printf( "Error: Open or create fails.\n" );

            return 1;
            }
        }

    //
    //  We process a request to search an existing
    //  database.
    //
    if ( (Count >= 4) && (strcmp( Argument[1],"-f" ) == 0) )
        {
        AUTO DATABASE Database;

        if ( Database.Open( Argument[2] ) )
            {
            if ( Database.Find( DisplayReference,(Count-3),& Argument[3] ) )
                {
                Database.Close();

                return 0;
                }
            else
                {
                Database.Close();

                printf( "Error: Search fails.\n" );

                return 1;
                }
            }
        }

    //
    //  We process a request to walk an existing
    //  database.
    //
    if ( (Count == 3) && (strcmp( Argument[1],"-w" ) == 0) )
        {
        AUTO DATABASE Database;

        if ( Database.Open( Argument[2] ) )
            {
            if ( Database.Walk( DisplayWord ) )
                {
                Database.Close();

                return 0;
                }
            else
                {
                Database.Close();

                printf( "Error: walk fails.\n" );

                return 1;
                }
            }
        }

    printf
        (
        "Usage: book [-c <DB name> <file name> | "
        "-f <DB name> <search words> ... | -w <DB name>]\n"
        );

    return 1;
    }
