    //
    //  Header files.
    //
    //  The header files provide access to functionality in the
    //  Rockall-DB library.
    //
#include "Words.hpp"

    //
    //  Words (Constructor).
    //
    //  The words class does not use Rockall-DB and exists
    //  purely to process the words of the book into a collection
    //  of references that are stored in the database.
    //
WORDS::WORDS( void )
    { ZeroAll(); }

    //
    //  Analyse.
    //
    //  The analyse function processes all the text in the book
    //  and into a number of references which it stores by using
    //  callbacks to the database class.
    //
BOOLEAN WORDS::Analyse
        (
        void                                *Context,
        FUNCTION1                           NewParagraph,
        FUNCTION2                           NewWord,
        FILE_SIZE                           FileSize,
        char                                *Text
        )
    {
    //
    //  Clean up.
    //
    ZeroAll();

    //
    //  We start by expecting a 'Title'.
    //
    NewLine = MinTitleLines;

    //
    //  We process the entire text file.
    //
    while ( EndOfWord < FileSize )
        {
        //
        //  We process the file a character at a time.
        //
        switch ( Text[ EndOfWord ] )
            {
            default:
                {
                //
                //  Process any white space.
                //
                if ( ! WhiteSpace( Context,FileSize,NewWord,Text ) )
                    { return False; }

                break;
                }

            case 'A': case 'a':
            case 'B': case 'b':
            case 'C': case 'c':
            case 'D': case 'd':
            case 'E': case 'e':
            case 'F': case 'f':
            case 'G': case 'g':
            case 'H': case 'h':
            case 'I': case 'i':
            case 'J': case 'j':
            case 'K': case 'k':
            case 'L': case 'l':
            case 'M': case 'm':
            case 'N': case 'n':
            case 'O': case 'o':
            case 'P': case 'p':
            case 'Q': case 'q':
            case 'R': case 'r':
            case 'S': case 's':
            case 'T': case 't':
            case 'U': case 'u':
            case 'V': case 'v':
            case 'W': case 'w':
            case 'X': case 'x':
            case 'Y': case 'y':
            case 'Z': case 'z':
                {
                //
                //  Process a character based on the number of
                //  new lines that immediately preceed it.
                //
                switch ( NewLine )
                    {
                    default:
                        {
                        if ( NewLine >= MinTitleLines )
                            { Title( FileSize,Text ); }
                        else
                            {
                            if ( ! Paragraph( Context,FileSize,NewParagraph,Text ) )
                                { return False; }
                            }

                        break;
                        }

                    case 1:
                        {
                        if ( NewLine >= MinParagraphLines )
                            {
                            if ( ! Paragraph( Context,FileSize,NewParagraph,Text ) )
                                { return False; }
                            }
                        else
                            { Word(); }

                        break;
                        }

                    case 0:
                        { Word(); break; }
                    }

                //
                //  Zero the white space counters.
                //
                NewLine = 0;
                Space = 0;

                break;
                }
            }
        }

    return True;
    }

    //
    //  Paragraph.
    //
    //  When we see a paragraph we create a reference for it
    //  before we process any of the words in the paragraph.
    //
BOOLEAN WORDS::Paragraph( void *Context,FILE_SIZE FileSize,FUNCTION1 NewParagraph,char *Text )
    {
    AUTO FILE_ADDRESS Count1;
    AUTO FILE_ADDRESS Count2;
    AUTO FILE_SIZE Count3;
    AUTO FILE_SIZE Count4;

    //
    //  A paragraph starts at the beginning of a line so skip
    //  backwards until we find the new line.
    //
    for ( Count1=EndOfWord;Count1 > 0;Count1 -- )
        {
        if ( Text[ Count1 ] == '\n' )
            { Count1 ++; break; }
        }

    //
    //  A paragraph ends will a number of new lines so skip
    //  forwards until we find the end of the paragraph.
    //
    for ( Count2=EndOfWord,Count3=0,Count4=0;Count2 < FileSize;Count2 ++ )
        {
        AUTO BOOLEAN Eof = (Count2 >= (FileSize-1));
        AUTO BOOLEAN Found = False;

        //
        //  We carefully count any spaces, returns, tabs and
        //  new lines while we look for the end of the paragraph.
        //
        switch ( Text[ Count2 ] )
            {
            case ' ':
            case '\r':
            case '\t':
                { Count4 ++; break; }

            case '\n':
                { Found = ((++ Count3) >= MinParagraphLines); break; }

            default:
                { Count3=0; Count4=0; break; }
            }

        //
        //  When we have find the start and end of the paragragh
        //  we create a new reference and make a callback to the
        //  database class to store it.
        //
        if ( (Eof) || (Found) )
            {
            AUTO FILE_SIZE Adjustment = ((Count3 + Count4) - 1);
            AUTO REFERENCE Reference;

            StartOfWord = EndOfWord;
            NewLine = 0;
            ParagraphStart = Count1;
            ParagraphSize = (Count2 - Count1 - Adjustment);

            Reference.ParagraphStart = ParagraphStart;
            Reference.ParagraphSize = ParagraphSize;
            Reference.TitleStart = TitleStart;
            Reference.TitleSize = TitleSize;

            return (NewParagraph( Context,& Reference ));
            }
        }

    return True;
    }

    //
    //  Title.
    //
    //  We keep track of any titles so that we can include
    //  them in any new references.
    //
void WORDS::Title( FILE_SIZE FileSize,char *Text )
    {
    TitleStart = EndOfWord;

    //
    //  We skip to the end of the line.
    //
    for
        (
        /* void */;
        (EndOfWord < FileSize)
            &&
        Text[ EndOfWord ] != '\r'
            &&
        Text[ EndOfWord ] != '\n';
        EndOfWord ++
        );

    StartOfWord = EndOfWord;
    TitleSize = (EndOfWord - TitleStart);
    }

    //
    //  White space.
    //
    //  We skip all white space.
    //
BOOLEAN WORDS::WhiteSpace( void *Context,FILE_SIZE FileSize,FUNCTION2 NewWord,char *Text )
    {
    //
    //  When we find a complete word we add it to the index
    //  by doing a callback to the database class.
    //
    if ( (StartOfWord < EndOfWord) && (Space == 0) )
        {
        AUTO FILE_SIZE Size = (EndOfWord - StartOfWord);

        if ( ! NewWord( Context,Size,& Text[ StartOfWord ] ) )
            { return False; }
        }

    //
    //  We count the number of new lines.
    //
    if ( Text[ EndOfWord ] == '\n' )
        { NewLine ++; }

    //
    //  We count any occurances of white space.
    //
    StartOfWord = (++ EndOfWord);
    Space ++;

    return True;
    }

    //
    //  Word.
    //
    //  We track the start of every new word.
    //
void WORDS::Word( void )
    {
    if ( Space > 0 )
        { StartOfWord = EndOfWord; }

    EndOfWord ++;
    }

    //
    //  Zero All.
    //
    //  Zero all the member values.
    //
void WORDS::ZeroAll( void )
    {
    NewLine = 0;

    ParagraphStart = 0;
    ParagraphSize = 0;

    StartOfWord = 0;
    EndOfWord = 0;

    Space = 0;

    TitleStart = 0;
    TitleSize = 0;
    }

    //
    //  Words (Destructor).
    //
WORDS::~WORDS( void )
    { ZeroAll(); }
