#ifndef _WORDS_HPP_
#define _WORDS_HPP_ 1
    //
    //  Header files.
    //
    //  The header files provide access to functionality in the
    //  Rockall-DB library.
    //
#include "RockallTypes.hpp"

    //
    //  Reference.
    //
    //  The entire book is read into the database as a single
    //  large string.  A reference points to a paragraph and
    //  a chapter title within the book.
    //
typedef struct
    {
    FILE_ADDRESS                            ParagraphStart;
    FILE_SIZE                               ParagraphSize;

    FILE_ADDRESS                            TitleStart;
    FILE_SIZE                               TitleSize;
    }
REFERENCE;

    //
    //  Words.
    //
    //  The words class does not use Rockall-DB and exists
    //  purely to process the words of the book into a collection
    //  of references that are stored in the database.
    //
class WORDS
    {
        //
        //  Private constants.
        //
        STATIC CONST int MinParagraphLines  = 2;
        STATIC CONST int MinTitleLines      = 5;

        //
        //  Private data.
        //
        FILE_SIZE                           NewLine;

        FILE_ADDRESS                        ParagraphStart;
        FILE_SIZE                           ParagraphSize;

        FILE_ADDRESS                        StartOfWord;
        FILE_ADDRESS                        EndOfWord;

        FILE_SIZE                           Space;

        FILE_ADDRESS                        TitleStart;
        FILE_SIZE                           TitleSize;

        //
        //  Function prototypes.
        //
        typedef BOOLEAN (*FUNCTION1)( void *Context,REFERENCE *Reference );

        typedef BOOLEAN (*FUNCTION2)( void *Context,FILE_SIZE Size,char *Text );

    public:
        //
        //  Public functions.
        //
        WORDS( void );

        BOOLEAN Analyse
            (
            void                            *Context,
            FUNCTION1                       NewParagraph,
            FUNCTION2                       NewWord,
            FILE_SIZE                       FileSize,
            char                            *Text
            );

        ~WORDS( void );

    private:
        //
        //  Private functions.
        //
        BOOLEAN Paragraph( void *Context,FILE_SIZE FileSize,FUNCTION1 NewParagraph,char *Text );

        void Title( FILE_SIZE FileSize,char *Text );

        BOOLEAN WhiteSpace( void *Context,FILE_SIZE FileSize,FUNCTION2 NewWord,char *Text );

        void Word( void );

        void ZeroAll( void );

        //
        //  Disabled functionality.
        //
        WORDS( const WORDS & NoCopies );

        void operator=( const WORDS & NoCopies );
    };
#endif