#ifdef ROCKALL_DB_EXAMPLES
#define main Example_04_03
#endif

#include "AutomaticHeapScope.hpp"
#include "DatabaseHeap.hpp"
#include "MultiThreadedHeap.hpp"
#include "Failure.hpp"
#include "SingleThreadedHeap.hpp"
#include "Stack.hpp"
#include "TransactionalHeap.hpp"

static void PushAndPop( void )
  {
  AUTOMATIC_HEAP_SCOPE<VIRTUAL_HEAP> Scope;

  if ( Scope.BeginTransaction() )
    {
    STACK<VIRTUAL_HEAP,NO_LOCK,int> Stack( & Scope );
    static const int MaxCycles = 100;
    int Count;
    int Value;

    for
      (
      Count=0;
      (Count < MaxCycles) && (Stack.PushValue( & Scope,Count ));
      Count ++
      );

    if ( Count < MaxCycles )
      { FAILURE( "Unable to push all the values" ); }

    for
      (
      /* void */;
      (Count > 0) && (Stack.PopValue( & Scope,& Value ));
      Count --
      );

    if ( Count > 0 )
      { FAILURE( "Unable to pop all the values" ); }
    }
  }

int main( int Count,char *Argument[] )
  {
  if ( Count == 2 )
    {
    SINGLE_THREADED_HEAP Heap1;
    MULTI_THREADED_HEAP Heap2;
    TRANSACTIONAL_HEAP Heap3;
    DATABASE_HEAP Heap4;
    VIRTUAL_HEAP *Heaps[4] = { & Heap1,& Heap2,& Heap3,& Heap4 };
    int Count;

    for ( Count=0;Count < 4;Count ++ )
      {
      AUTOMATIC_HEAP_SCOPE<VIRTUAL_HEAP> Scope( Heaps[ Count ] );

      if ( Scope.CreateFile( Argument[1] ) )
        {
        PushAndPop();

        Scope.CloseFile();
        }
      }
    }

  return 0;
  }
