#ifdef ROCKALL_DB_EXAMPLES
#define main Example_04_11
#endif

#include <stdio.h>

#include "AutomaticHeapScope.hpp"
#include "AutomaticViewScope.hpp"
#include "Example-04-08.hpp"
#include "Failure.hpp"
#include "Hash.hpp"
#include "RockallNew.hpp"
#include "RockallTypes.hpp"
#include "RowSet.hpp"

typedef DATABASE_HEAP HEAP_TYPE;
typedef FIXED_STRING<HEAP_TYPE,char,MaxName> NAME;
typedef HASH<HEAP_TYPE,NO_LOCK,NAME,FILE_ADDRESS> INDEX;
typedef ROW_SET<HEAP_TYPE,NO_LOCK> ROWS;

static bool NewKey( FILE_ADDRESS Address,NAME & Name,INDEX *Index )
  {
  AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Scope;
  FILE_ADDRESS Duplicates;
  ROWS *Rows;

  if ( Index -> FindKey( & Scope,Name,& Duplicates ) )
    {
    AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> ViewScope( & Scope );

    return
      (
      ViewScope.View( Duplicates,((void**) & Rows) )
        &&
      Rows -> NewValue( Address )
      );
    }
  else
    {
    FILE_ADDRESS NewRows;

    return
      (
      (ROCKALL_NEW<ROWS>::New( & Scope,& NewRows,& Rows ))
        &&
      Rows -> NewValue( Address )
        &&
      Index -> NewKey( & Scope,Name,NewRows )
      );
    }
  }

static bool CreateIndex( INDEX *Index )
  {
  AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Scope;
  int Count;

  for ( Count=0;Count < MaxDetails;Count ++ )
    {
    DETAILS *Current1 = & Details[ Count ];
    NAME Name1 = Current1 -> Forename;
    NAME Name2 = Current1 -> Surname;
    NAME Postcode = Current1 -> Postcode;
    FILE_ADDRESS Address;
    DETAILS *Current2;

    if
        (
        (ROCKALL_NEW<DETAILS>::New( & Scope,& Address,& Current2 ))
          &&
        (NewKey( Address,Name1,Index ))
          &&
        (NewKey( Address,Name2,Index ))
          &&
        (NewKey( Address,Postcode,Index ))
        )
      { (*Current2) = (*Current1); }
    else
      { FAILURE( "Unable to add key" ); }
    }

  return (Count == MaxDetails);
  }

static void FindKey( char *Name,INDEX *Index )
  {
  AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Scope;
  FILE_ADDRESS Address;
  NAME Key = Name;

  if ( Index -> FindKey( & Scope,Key,& Address ) )
    {
    AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> View1( & Scope );
    ROWS *Rows;

    if ( View1.View( Address,((void**) & Rows) ) )
      {
      FILE_ADDRESS *Set;
      FILE_SIZE Size;

      Rows -> Touch( & Scope );

      if ( Rows -> View( & Scope,& Set,& Size ) )
        {
        int Count;

        for ( Count=0;Count < Size;Count ++ )
          {
          AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> View2( & Scope );
          DETAILS *Target;

          if ( View2.View( Set[ Count ],((void**) & Target) ) )
            {
            printf
              (
              "The details are: '%s %s' at '%s'\n",
              Target -> Forename,
              Target -> Surname,
              Target -> Postcode
              );
            }
          }

        Rows -> EndView( & Scope );
        }
      }
    }
  }

int main( int Count,char *Argument[] )
  {
  HEAP_TYPE Heap;
  AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Scope( & Heap );

  if ( (Count==2) && (Scope.CreateFile( Argument[1] )) )
    {
    if ( Scope.BeginTransaction() )
      {
      INDEX Index;

      if ( CreateIndex( & Index ) )
        { FindKey( Details[2].Forename,& Index ); }

      Scope.EndTransaction();
      }

    Scope.CloseFile();
    }

  return 0;
  }
