#ifdef ROCKALL_DB_EXAMPLES
#define main Example_04_12
#endif

#include "AutomaticHeapScope.hpp"
#include "DatabaseHeap.hpp"
#include "FlexibleString.hpp"
#include "Failure.hpp"
#include "Queue.hpp"
#include "RockallTypes.hpp"

typedef FLEXIBLE_STRING<DATABASE_HEAP,CHAR> MESSAGE;

void PushAndPop( void )
  {
  AUTOMATIC_HEAP_SCOPE<DATABASE_HEAP> Scope;
  static const int MaxMessages = 10;
  MESSAGE Messages[ MaxMessages ] =
    {
    "One", "Two", "Three", "Four", "Five",
    "Six", "Seven", "Eight", "Nine", "Ten"
    };

  if ( Scope.BeginTransaction() )
    {
    QUEUE<DATABASE_HEAP,NO_LOCK,MESSAGE> Queue( & Scope );
    MESSAGE Value;
    int Count;

    for
      (
      Count=0;
      (Count < MaxMessages)
        &&
      (Queue.PushValue( & Scope,Messages[ Count ] ));
      Count ++
      );

    if ( Count < MaxMessages )
      { FAILURE( "Unable to push all the values" ); }

    for
      (
      /* void */;
      (Count > 0) && (Queue.PopValue( & Scope,& Value ));
      Count --
      );

    if ( Count > 0 )
      { FAILURE( "Unable to pop all the values" ); }
    }
  }

int main( int Count,char *Argument[] )
  {
  DATABASE_HEAP Heap;
  AUTOMATIC_HEAP_SCOPE<DATABASE_HEAP> Scope( & Heap );

  if ( (Count == 2) && (Scope.CreateFile( Argument[1] )) )
    {
    PushAndPop();

    Scope.CloseFile();
    }

  return 0;
  }
