#ifdef ROCKALL_DB_EXAMPLES
#define main Example_06_30
#endif

#include "AutomaticHeapScope.hpp"
#include "AutomaticTransactionScope.hpp"
#include "AutomaticViewScope.hpp"
#include "DatabaseHeap.hpp"
#include "RockallDelete.hpp"
#include "RockallNew.hpp"
#include "RockallTypes.hpp"

typedef DATABASE_HEAP HEAP_TYPE;

int main( int Count,char *Argument[] )
  {
  HEAP_TYPE Heap;
  AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Scope( & Heap );
  FILE_ADDRESS Address;
  int Result = 0;
  int *Data;

  if ( (Count == 2) && (Scope.CreateFile( Argument[1] )) )
    {
      //  No transaction and no view.
      {
      AUTOMATIC_TRANSACTION_SCOPE<HEAP_TYPE> Transaction( & Scope );

      if ( Transaction.BeginTransaction() )
        {
        // Active transaction but no view.
        if ( ! ROCKALL_NEW<int>::New( & Scope,& Address,& Data ) )
          { Result = 1; }
        }
      }

      //  No transaction and no view.
      {
      AUTOMATIC_TRANSACTION_SCOPE<HEAP_TYPE> Transaction( & Scope );

      if ( Transaction.BeginTransaction() )
        {
          // Active transaction but no view.
          {
          AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> View( & Scope );

          if ( View.ExclusiveView( Address,((void**) & Data) ) )
            {
            //  Active transaction and exclusive view.
            }
          }

          // Active transaction but no view.
          {
          AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> View( & Scope );

          if ( View.View( Address,((void**) & Data) ) )
            {
            //  Active transaction and shared view.
            }
          }

          // Active transaction but no view.
          {
          AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> View( & Heap );

          if ( View.ExclusiveView( Address,((void**) & Data) ) )
            {
            //  Active transaction and exclusive view.
            View.Release();
            // Active transaction but no view.
            }
          }
        }

        //  No transaction and no view.
        {
        AUTOMATIC_TRANSACTION_SCOPE<HEAP_TYPE> Transaction( & Heap );

        if ( Transaction.BeginTransaction() )
          {
          // Active transaction but no view.
          if ( ! ROCKALL_DELETE<int>::Delete( & Heap,Address ) )
            { Result = 1; }

          Transaction.EndTransaction();
          //  No transaction and no view.
          }
        }
      }

    Scope.CloseFile();
    }

  return Result;
  }