//												Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

	/**************************************************************************************************************/
	/*																											  */
	/*	Standard copyright and disclaimer information.															  */
	/*																											  */
	/*	Copyright  Michael Parkes 2015.  All rights reserved.													  */
	/*																											  */
	/*	Reproduction or transmission of all or part of this work, whether by photocopying or storing in			  */
	/*	any medium by electronic means or otherwise, without the written permission of the owner, is			  */
	/*	prohibited. The commission of any unauthorised act in relation to this work may result in civil			  */
	/*	or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to			  */
	/*	this work.																								  */
	/*																											  */
	/*	The contents of this file may to contain errors, mistakes, omissions and other types of issues.			  */
	/*	Consequently, the contents of this file are provided 'as is' and it is the users responsibility to		  */
	/*	verify that the contents of the file are fit for purpose for which it is used.  Additionally,			  */
	/*	it might be shown at some point that portions of this file are subject to patent claims or other		  */
	/*	types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at		  */
	/*	the users own risk.  No responsibility can be accepted for any consequences or any losses unless		  */
	/*	clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,		  */
	/*	any user of this file agrees to take reasonable steps to keep the contents of this file private and		  */
	/*	to indemnify the owners for any claims or losses the owners may find themselves liable for as a			  */
	/*	result of the use of this file.																			  */
	/*																											  */
	/*	Please see the related license for additional terms and conditions.  Any use of this file indicates		  */
	/*	acceptance of all the associated terms and conditions.													  */
	/*																											  */
	/**************************************************************************************************************/

#include "DatabaseHeap.hpp"
#include "MultiLanguage.hpp"
#include "MultiThreadedHeap.hpp"
#include "PlacementNew.hpp"
#include "SingleThreadedHeap.hpp"
#include "TransactionalHeap.hpp"
#include "VirtualHeap.hpp"

		/******************************************************************************************************/
		/*																									  */
		/*	Multi-Language Support.																			  */
		/*																									  */
		/*	We support multiple langauges by providing a simple 'C' style interface.						  */
		/*																									  */
		/******************************************************************************************************/

NAMESPACE ROCKALL
	{
			/**********************************************************************************************/
			/*																							  */
			/*	Class Static Data.																		  */
			/*																							  */
			/*	The class static data needs to be initialised before it's ready for use.				  */
			/*																							  */
			/**********************************************************************************************/

	STATIC MULTI_THREADED_HEAP SystemHeap;

			/**********************************************************************************************/
			/*																							  */
			/*	Create New Database Heap.																  */
			/*																							  */
			/*	We create a new instance of a heap.														  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CreateNewDatabaseHeap( VOID **Heap )
		{
		//
		//  We create space for the new 'Heap'.
		//
		if ( SystemHeap.New( Heap,sizeof(DATABASE_HEAP) ) )
			{
			//
			//  Call the constructor.
			//
			PLACEMENT_NEW( (*Heap),DATABASE_HEAP );

			return True;
			}
		else
			{
			//
			//  We were unable to create a 'Heap'.
			//
			(*Heap) = NULL; 

			return False;
			}
		}

			/**********************************************************************************************/
			/*																							  */
			/*	Create New Multi Threaded Heap.															  */
			/*																							  */
			/*	We create a new instance of a heap.														  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CreateNewMultiThreadedHeap( VOID **Heap )
		{
		//
		//  We create space for the new 'Heap'.
		//
		if ( SystemHeap.New( Heap,sizeof(MULTI_THREADED_HEAP) ) )
			{
			//
			//  Call the constructor.
			//
			PLACEMENT_NEW( (*Heap),MULTI_THREADED_HEAP );

			return True;
			}
		else
			{
			//
			//  We were unable to create a 'Heap'.
			//
			(*Heap) = NULL; 

			return False;
			}
		}

			/**********************************************************************************************/
			/*																							  */
			/*	Create New Single Threaded Heap.														  */
			/*																							  */
			/*	We create a new instance of a heap.														  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CreateNewSingleThreadedHeap( VOID **Heap )
		{
		//
		//  We create space for the new 'Heap'.
		//
		if ( SystemHeap.New( Heap,sizeof(SINGLE_THREADED_HEAP) ) )
			{
			//
			//  Call the constructor.
			//
			PLACEMENT_NEW( (*Heap),SINGLE_THREADED_HEAP );

			return True;
			}
		else
			{
			//
			//  We were unable to create a 'Heap'.
			//
			(*Heap) = NULL; 

			return False;
			}
		}

			/**********************************************************************************************/
			/*																							  */
			/*	Create New Transactional Heap.															  */
			/*																							  */
			/*	We create a new instance of a heap.														  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CreateNewTransactionalHeap( VOID **Heap )
		{
		//
		//  We create space for the new 'Heap'.
		//
		if ( SystemHeap.New( Heap,sizeof(TRANSACTIONAL_HEAP) ) )
			{
			//
			//  Call the constructor.
			//
			PLACEMENT_NEW( (*Heap),TRANSACTIONAL_HEAP );

			return True;
			}
		else
			{
			//
			//  We were unable to create a 'Heap'.
			//
			(*Heap) = NULL; 

			return False;
			}
		}

			/**********************************************************************************************/
			/*																							  */
			/*	Create File.																			  */
			/*																							  */
			/*	Create a new file.																		  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CreateFile( VOID *Heap,CHAR *FileName,VOID *Function,VOID *UserValue )
		{ return (((VIRTUAL_HEAP*) Heap) -> CreateFile( FileName,((ENCRYPTION_FUNCTION) Function),UserValue )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Create Shadow File.																		  */
			/*																							  */
			/*	Create a new shadow file.																  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CreateShadowFile( VOID *Heap,CHAR *FileName,SBIT32 *ID,BOOLEAN Wait )
		{ return (((VIRTUAL_HEAP*) Heap) -> CreateShadowFile( FileName,ID,Wait )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Create Shadow File.																		  */
			/*																							  */
			/*	Create a new shadow file.																  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CreateSnapshotFile( VOID *Heap,CHAR *FileName )
		{ return (((VIRTUAL_HEAP*) Heap) -> CreateSnapshotFile( FileName )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Open File.																				  */
			/*																							  */
			/*	Open an existing file.																	  */
			/*																							  */
			/**********************************************************************************************/

	ROCKALL_DB_LINKAGE BOOLEAN OpenFile
			(
			VOID								*Heap,
			CHAR								*FileName,
			BOOLEAN								ReadOnly,
			VOID								*Function,
			VOID								*UserValue
			)
		{
		return
			(
			((VIRTUAL_HEAP*) Heap) -> OpenFile
				( 
				FileName,
				ReadOnly,
				((ENCRYPTION_FUNCTION) Function),
				UserValue 
				)
			); 
		}

			/**********************************************************************************************/
			/*																							  */
			/*	Create Region.																			  */
			/*																							  */
			/*	Create a new region.																	  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CreateRegion( VOID *Heap,SBIT32 *NewRegionID )
		{ return (((VIRTUAL_HEAP*) Heap) -> CreateRegion( NewRegionID )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Begin Transaction.																		  */
			/*																							  */
			/*	Create a new transaction.																  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN BeginTransaction( VOID *Heap,FILE_ADDRESS *NewTransactionID,VOID *UserLock )
		{ return (((VIRTUAL_HEAP*) Heap) -> BeginTransaction( NewTransactionID,((EXPORT_MEMORY_LOCK*) UserLock) )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Change Region.																			  */
			/*																							  */
			/*	Change the region.																		  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN ChangeRegion( VOID *Heap,SBIT32 NewRegionID,SBIT32 *OldRegionID )
		{ return (((VIRTUAL_HEAP*) Heap) -> ChangeRegion( NewRegionID,OldRegionID )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Commit.																					  */
			/*																							  */
			/*	Commit all transactions.																  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN Commit( VOID *Heap )
		{ return (((VIRTUAL_HEAP*) Heap) -> Commit()); }

			/**********************************************************************************************/
			/*																							  */
			/*	Delete.																					  */
			/*																							  */
			/*	Delete a memory space.																	  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN Delete( VOID *Heap,FILE_ADDRESS Address,BOOLEAN Zero )
		{ return (((VIRTUAL_HEAP*) Heap) -> Delete( Address,False,Zero )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Delete All.																				  */
			/*																							  */
			/*	Delete all memory spaces.																  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN DeleteAll( VOID *Heap,BOOLEAN Zero )
		{ return (((VIRTUAL_HEAP*) Heap) -> DeleteAll( False,Zero )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Details.																				  */
			/*																							  */
			/*	Find the details for a memory address.													  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN Details( VOID *Heap,FILE_ADDRESS Address,FILE_SIZE *Space )
		{ return (((VIRTUAL_HEAP*) Heap) -> Details( Address,NULL,Space )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Exclusive View.																			  */
			/*																							  */
			/*	Claim an exclusive view on a memory space.												  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN ExclusiveView( VOID *Heap,FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space )
		{ return (((VIRTUAL_HEAP*) Heap) -> ExclusiveView( Address,Data,Space )); }

			/**********************************************************************************************/
			/*																							  */
			/*	End Exclusive View.																		  */
			/*																							  */
			/*	End an exclusive view on a memory space.												  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN EndExclusiveView( VOID *Heap,FILE_ADDRESS Address )
		{ return (((VIRTUAL_HEAP*) Heap) -> EndExclusiveView( Address,False )); }

			/**********************************************************************************************/
			/*																							  */
			/*	File Size.																				  */
			/*																							  */
			/*	Calculate the file size.																  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN FileSize( VOID *Heap,FILE_SIZE *Size )
		{ return (((VIRTUAL_HEAP*) Heap) -> FileSize( Size )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Get Error.																				  */
			/*																							  */
			/*	Get the error number.																	  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN GetError( VOID *Heap,SBIT32 *ErrorNumber )
		{ return (((VIRTUAL_HEAP*) Heap) -> GetError( ErrorNumber )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Get User Value.																			  */
			/*																							  */
			/*	Get a user value.																		  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN GetUserValue( VOID *Heap,SBIT32 ID,FILE_ADDRESS *Value )
		{ return (((VIRTUAL_HEAP*) Heap) -> GetUserValue( ID,Value )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Join Transaction.																		  */
			/*																							  */
			/*	Join a transaction.																		  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN JoinTransaction( VOID *Heap,FILE_ADDRESS TransactionID,VOID *UserLock )
		{ return (((VIRTUAL_HEAP*) Heap) -> JoinTransaction( TransactionID,((EXPORT_MEMORY_LOCK*) UserLock ) )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Leave Transaction.																		  */
			/*																							  */
			/*	Leave a transaction.																	  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN LeaveTransaction( VOID *Heap )
		{ return (((VIRTUAL_HEAP*) Heap) -> LeaveTransaction()); }

			/**********************************************************************************************/
			/*																							  */
			/*	New.																					  */
			/*																							  */
			/*	Create a new memory space.																  */
			/*																							  */
			/**********************************************************************************************/

	ROCKALL_DB_LINKAGE BOOLEAN New
			(
			VOID								*Heap,
			FILE_ADDRESS						*Address,
			VOID								**Data,
			FILE_SIZE							Size,
			FILE_SIZE							*Space,
			BOOLEAN								Zero
			)
		{ return (((VIRTUAL_HEAP*) Heap) -> New( Address,Data,Size,Space,False,Zero )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Resize.																					  */
			/*																							  */
			/*	Resize a memory space.																	  */
			/*																							  */
			/**********************************************************************************************/

	ROCKALL_DB_LINKAGE BOOLEAN Resize
			(
			VOID								*Heap,
			FILE_ADDRESS						*Address,
			VOID								**Data,
			FILE_SIZE							Size,
			FILE_SIZE							*Space,
			BOOLEAN								Zero
			)
		{ return (((VIRTUAL_HEAP*) Heap) -> Resize( Address,Data,Size,Space,Zero )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Set Error.																				  */
			/*																							  */
			/*	Set the error number.																	  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN SetError( VOID *Heap,SBIT32 ErrorNumber )
		{ return (((VIRTUAL_HEAP*) Heap) -> SetError( ErrorNumber )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Set User Value.																			  */
			/*																							  */
			/*	Set a user value.																		  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN SetUserValue( VOID *Heap,SBIT32 ID,FILE_ADDRESS Value )
		{ return (((VIRTUAL_HEAP*) Heap) -> SetUserValue( ID,Value )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Size.																					  */
			/*																							  */
			/*	Compute the size of the heap.															  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN Size( VOID *Heap,FILE_SIZE *Space )
		{ return (((VIRTUAL_HEAP*) Heap) -> Size( Space )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Touch.																					  */
			/*																							  */
			/*	Touch a memory space.																	  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN Touch( VOID *Heap,FILE_ADDRESS Address )
		{ return (((VIRTUAL_HEAP*) Heap) -> Touch( Address )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Update.																					  */
			/*																							  */
			/*	Claim an update view on a memory space.													  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN Update( VOID *Heap,FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space )
		{ return (((VIRTUAL_HEAP*) Heap) -> Update( Address,Data,Space )); }

			/**********************************************************************************************/
			/*																							  */
			/*	View.																					  */
			/*																							  */
			/*	Claim a view on a memory space.															  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN View( VOID *Heap,FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space )
		{ return (((VIRTUAL_HEAP*) Heap) -> View( Address,Data,Space )); }

			/**********************************************************************************************/
			/*																							  */
			/*	End View.																				  */
			/*																							  */
			/*	End a view on a memory space.															  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN EndView( VOID *Heap,FILE_ADDRESS Address )
		{ return (((VIRTUAL_HEAP*) Heap) -> EndView( Address,False )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Walk.																					  */
			/*																							  */
			/*	Walk the heap.																			  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN Walk( VOID *Heap,VOID *WalkFunction,VOID *UserValue )
		{ return (((VIRTUAL_HEAP*) Heap) -> Walk( ((HEAP_WALK_FUNCTION) WalkFunction),UserValue )); }

			/**********************************************************************************************/
			/*																							  */
			/*	End Transaction.																		  */
			/*																							  */
			/*	End a transaction.																		  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN EndTransaction( VOID *Heap,BOOLEAN Abort,BOOLEAN Commit )
		{ return (((VIRTUAL_HEAP*) Heap) -> EndTransaction( Abort,Commit )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Delete Region.																			  */
			/*																							  */
			/*	Delete a region.																		  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN DeleteRegion( VOID *Heap,SBIT32 RegionID )
		{ return (((VIRTUAL_HEAP*) Heap) -> DeleteRegion( RegionID )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Close Shadow File.																		  */
			/*																							  */
			/*	Close a shadow file.																	  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CloseShadowFile( VOID *Heap,SBIT32 ID )
		{ return (((VIRTUAL_HEAP*) Heap) -> CloseShadowFile( ID )); }

			/**********************************************************************************************/
			/*																							  */
			/*	Close File.																				  */
			/*																							  */
			/*	Close a file.																			  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN CloseFile( VOID *Heap )
		{ return (((VIRTUAL_HEAP*) Heap) -> CloseFile()); }

			/**********************************************************************************************/
			/*																							  */
			/*	Delete Heap.																			  */
			/*																							  */
			/*	We delete an instance of a heap.														  */
			/*																							  */
			/**********************************************************************************************/

	BOOLEAN DeleteHeap( VOID *Heap )
		{
		PLACEMENT_DELETE( Heap,VIRTUAL_HEAP );

		return (SystemHeap.Delete( Heap ));
		}
	}
