#ifndef _MULTI_LANGUAGE_HPP_
#define _MULTI_LANGUAGE_HPP_ 1
//												Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

	/**************************************************************************************************************/
	/*																											  */
	/*	Standard copyright and disclaimer information.															  */
	/*																											  */
	/*	Copyright  Michael Parkes 2015.  All rights reserved.													  */
	/*																											  */
	/*	Reproduction or transmission of all or part of this work, whether by photocopying or storing in			  */
	/*	any medium by electronic means or otherwise, without the written permission of the owner, is			  */
	/*	prohibited. The commission of any unauthorised act in relation to this work may result in civil			  */
	/*	or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to			  */
	/*	this work.																								  */
	/*																											  */
	/*	The contents of this file may to contain errors, mistakes, omissions and other types of issues.			  */
	/*	Consequently, the contents of this file are provided 'as is' and it is the users responsibility to		  */
	/*	verify that the contents of the file are fit for purpose for which it is used.  Additionally,			  */
	/*	it might be shown at some point that portions of this file are subject to patent claims or other		  */
	/*	types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at		  */
	/*	the users own risk.  No responsibility can be accepted for any consequences or any losses unless		  */
	/*	clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,		  */
	/*	any user of this file agrees to take reasonable steps to keep the contents of this file private and		  */
	/*	to indemnify the owners for any claims or losses the owners may find themselves liable for as a			  */
	/*	result of the use of this file.																			  */
	/*																											  */
	/*	Please see the related license for additional terms and conditions.  Any use of this file indicates		  */
	/*	acceptance of all the associated terms and conditions.													  */
	/*																											  */
	/**************************************************************************************************************/

#include "RockallTypes.hpp"

		/******************************************************************************************************/
		/*																									  */
		/*	Multi-Language Support.																			  */
		/*																									  */
		/*	We support multiple langauges by providing a simple 'C' style interface.						  */
		/*																									  */
		/******************************************************************************************************/

NAMESPACE ROCKALL
	{
			/**********************************************************************************************/
			/*																							  */
			/*	Multi-Language Heap Linkage.															  */
			/*																							  */
			/*	The linkage functions are responsible for creating and deleting heap instances.			  */
			/*																							  */
			/**********************************************************************************************/

	extern "C"
		{
		ROCKALL_DB_LINKAGE BOOLEAN CreateNewDatabaseHeap( VOID **Heap );

		ROCKALL_DB_LINKAGE BOOLEAN CreateNewMultiThreadedHeap( VOID **Heap );

		ROCKALL_DB_LINKAGE BOOLEAN CreateNewSingleThreadedHeap( VOID **Heap );

		ROCKALL_DB_LINKAGE BOOLEAN CreateNewTransactionalHeap( VOID **Heap );

		//
		//  The Rockall-DB core functionality.
		//
		ROCKALL_DB_LINKAGE BOOLEAN Delete( VOID *Heap,FILE_ADDRESS Address,BOOLEAN Zero );

		ROCKALL_DB_LINKAGE BOOLEAN DeleteAll( VOID *Heap,BOOLEAN Zero );

		ROCKALL_DB_LINKAGE BOOLEAN Details( VOID *Heap,FILE_ADDRESS Address,FILE_SIZE *Space );

		ROCKALL_DB_LINKAGE BOOLEAN GetError( VOID *Heap,SBIT32 *ErrorNumber );

		ROCKALL_DB_LINKAGE BOOLEAN New
			(
			VOID								*Heap,
			FILE_ADDRESS						*Address,
			VOID								**Data,
			FILE_SIZE							Size,
			FILE_SIZE							*Space,
			BOOLEAN								Zero
			);

		ROCKALL_DB_LINKAGE BOOLEAN Resize
			(
			VOID								*Heap,
			FILE_ADDRESS						*Address,
			VOID								**Data,
			FILE_SIZE							NewSize,
			FILE_SIZE							*Space,
			BOOLEAN								Zero
			);

		ROCKALL_DB_LINKAGE BOOLEAN SetError( VOID *Heap,SBIT32 ErrorNumber );

		ROCKALL_DB_LINKAGE BOOLEAN Size( VOID *Heap,FILE_SIZE *Space );

		ROCKALL_DB_LINKAGE BOOLEAN Walk( VOID *Heap,VOID *WalkFunction,VOID *UserValue );

		//
		//  The Rockall-DB transactional functionality.
		//
		ROCKALL_DB_LINKAGE BOOLEAN CreateFile( VOID *Heap,CHAR *FileName,VOID *Function,VOID *UserValue );

		ROCKALL_DB_LINKAGE BOOLEAN CreateRegion( VOID *Heap,SBIT32 *NewRegionID );

		ROCKALL_DB_LINKAGE BOOLEAN BeginTransaction( VOID *Heap,FILE_ADDRESS *NewTransactionID,VOID *UserLock ); 

		ROCKALL_DB_LINKAGE BOOLEAN ChangeRegion( VOID *Heap,SBIT32 NewRegionID,SBIT32 *OldRegionID );

		ROCKALL_DB_LINKAGE BOOLEAN ExclusiveView( VOID *Heap,FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space );

		ROCKALL_DB_LINKAGE BOOLEAN EndExclusiveView( VOID *Heap,FILE_ADDRESS Address );

		ROCKALL_DB_LINKAGE BOOLEAN JoinTransaction( VOID *Heap,FILE_ADDRESS TransactionID,VOID *UserLock );

		ROCKALL_DB_LINKAGE BOOLEAN LeaveTransaction( VOID *Heap );

		ROCKALL_DB_LINKAGE BOOLEAN Update( VOID *Heap,FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space );

		ROCKALL_DB_LINKAGE BOOLEAN View( VOID *Heap,FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space );

		ROCKALL_DB_LINKAGE BOOLEAN EndView( VOID *Heap,FILE_ADDRESS Address );

		ROCKALL_DB_LINKAGE BOOLEAN EndTransaction( VOID *Heap,BOOLEAN Abort,BOOLEAN Commit );

		ROCKALL_DB_LINKAGE BOOLEAN DeleteRegion( VOID *Heap,SBIT32 RegionID );

		ROCKALL_DB_LINKAGE BOOLEAN CloseFile( VOID *Heap );

		//
		//  The Rockall-DB database functionality.
		//
		ROCKALL_DB_LINKAGE BOOLEAN CreateShadowFile( VOID *Heap,CHAR *FileName,SBIT32 *ID,BOOLEAN Wait );

		ROCKALL_DB_LINKAGE BOOLEAN CreateSnapshotFile( VOID *Heap,CHAR *FileName );

		ROCKALL_DB_LINKAGE BOOLEAN OpenFile
			(
			VOID								*Heap,
			CHAR								*FileName,
			BOOLEAN								ReadOnly,
			VOID								*Function,
			VOID								*UserValue
			);

		ROCKALL_DB_LINKAGE BOOLEAN Commit( VOID *Heap );

		ROCKALL_DB_LINKAGE BOOLEAN FileSize( VOID *Heap,FILE_SIZE *Size );

		ROCKALL_DB_LINKAGE BOOLEAN GetUserValue( VOID *Heap,SBIT32 ID,FILE_ADDRESS *Value );

		ROCKALL_DB_LINKAGE BOOLEAN SetUserValue( VOID *Heap,SBIT32 ID,FILE_ADDRESS Value );

		ROCKALL_DB_LINKAGE BOOLEAN Touch( VOID *Heap,FILE_ADDRESS Address );

		ROCKALL_DB_LINKAGE BOOLEAN CloseShadowFile( VOID *Heap,SBIT32 ID );

		ROCKALL_DB_LINKAGE BOOLEAN DeleteHeap
			(
			VOID								*Heap
			);
		}
	}
#endif