#ifndef _MULTI_THREADED_HEAP_HPP_
#define _MULTI_THREADED_HEAP_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "ExportHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Multi Threaded Heap.                                                                              */
        /*                                                                                                    */
        /*  A multi threaded heap is optimized for use by multiple threads.  A multi threaded heap            */
        /*  contains full locking code and is thread safe in a full multi-threaded environment (i.e.          */
        /*  multiple threads using the same instance of a heap at the same time).  All instances of           */
        /*  a multi threaded heap are entirely seperate and align the data on cache lines where               */
        /*  possible to minimize false sharing.                                                               */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Multi Threaded Heap Linkage.                                                              */
            /*                                                                                            */
            /*  The linkage functions are responsible for creating and deleting heap instances.           */
            /*                                                                                            */
            /**********************************************************************************************/

    extern "C"
        {
        ROCKALL_DB_LINKAGE VOID CreateMultiThreadedHeap
            (
            BOOLEAN                             *Active,
            VIRTUAL_HEAP                        **Heap,
            VOID                                **UserValue
            );

        ROCKALL_DB_LINKAGE VOID DeleteMultiThreadedHeap
            (
            BOOLEAN                             *Active,
            VIRTUAL_HEAP                        **Heap,
            VOID                                **UserValue
            );
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Multi Threaded Heap.                                                                      */
            /*                                                                                            */
            /*  We create and destroy instances of the heap using the above functions.                    */
            /*                                                                                            */
            /**********************************************************************************************/

    class MULTI_THREADED_HEAP : public EXPORT_HEAP
        {
        public:
            //
            //  Public virtual functions.
            //
            MULTI_THREADED_HEAP( VOID )
                { CreateMultiThreadedHeap( & Active,& Heap,& UserValue ); }

            VIRTUAL SBIT32 Feature( VOID )
                { return (RockallMultiThreadedHeapFlags); }

            INLINE BOOLEAN Feature( SBIT32 Features )
                { return ((Features & RockallMultiThreadedHeapFlags) == Features); }

            ~MULTI_THREADED_HEAP( VOID )
                { DeleteMultiThreadedHeap( & Active,& Heap,& UserValue ); }

        private:
            //
            //  Disabled functionality.
            //
            MULTI_THREADED_HEAP( CONST MULTI_THREADED_HEAP & NoCopies );

            VOID operator=( CONST MULTI_THREADED_HEAP & NoCopies );
        };
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the namespace to be ommitted in some cases but may         */
            /*  be optional disabled if needed.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
typedef ROCKALL::MULTI_THREADED_HEAP MULTI_THREADED_HEAP;
#endif
#endif