#ifndef _VIRTUAL_HEAP_HPP_
#define _VIRTUAL_HEAP_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "PublicConstants.hpp"
#include "VirtualDestructor.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Virtual Heap Interface.                                                                           */
        /*                                                                                                    */
        /*  The virtual heap interface is the standard interface for the 'Rockall-DB' suite of heaps          */
        /*  and databases.                                                                                    */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  A Class Forward Reference.                                                                */
            /*                                                                                            */
            /*  A class reference is sometimes made before a class is fully specified.  If this is        */
            /*  the case a forward reference to the class is made here.                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    class EXPORT_MEMORY_LOCK;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Public types.                                                                             */
            /*                                                                                            */
            /*  The public types specify structures and function prototypes that are useful to            */
            /*  external code.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    STATIC CONST SBIT32 NoRegion                = NoValue;

    typedef VOID (*ENCRYPTION_FUNCTION)
        (
        FILE_ADDRESS                            Address,
        VOID                                    *Data,
        BOOLEAN                                 Decrypt,
        BOOLEAN                                 Encrypt,
        FILE_SIZE                               Size,
        VOID                                    *UserValue
        );

    typedef BOOLEAN (*HEAP_WALK_FUNCTION)
        (
        BOOLEAN                                 Active,
        FILE_ADDRESS                            Address,
        VOID                                    *Data,
        BOOLEAN                                 Destroy,
        FILE_SIZE                               Space,
        VOID                                    *UserValue
        );

    typedef enum
        {
        RockallDatabaseHeap                     = 0x0100,
        RockallMultiThreadedHeap                = 0x0200,
        RockallSingleThreadedHeap               = 0x0400,
        RockallTransactionalHeap                = 0x0800,

        RockallFileSupport                      = 0x0010,
        RockallLockSupport                      = 0x0020,
        RockallMemorySupport                    = 0x0040,
        RockallTransactionalSupport             = 0x0080,

        RockallNoFileSupport                    = 0x0001,
        RockallNoLockSupport                    = 0x0002,
        RockallNoMemorySupport                  = 0x0004,
        RockallNoTransactionalSupport           = 0x0008,

        RockallDatabaseHeapFlags                = 0x01f0,
        RockallMultiThreadedHeapFlags           = 0x0269,
        RockallSingleThreadedHeapFlags          = 0x044b,
        RockallTransactionalHeapFlags           = 0x08e1,

        RockallFileLockFlags                    = 0x0071,
        RockallMemoryLockFlags                  = 0x0061,
        RockallNoLockFlags                      = 0x0017,

        RockallFileLockRequirements             = 0x0040,
        RockallMemoryLockRequirements           = 0x0041,
        RockallNoLockRequirements               = 0x0000
        }
    ROCKALL_FEATURE_FLAGS;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Virtual Heap.                                                                             */
            /*                                                                                            */
            /*  The virtual heap class provides prototypes for all the publically available               */
            /*  functionality within 'RockallDB.dll'.                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    class VIRTUAL_HEAP : public VIRTUAL_DESTRUCTOR
        {
        public:
            //
            //  The Rockall-DB core functionality.
            //
            VIRTUAL_HEAP( VOID ) { /* void */ };

            VIRTUAL BOOLEAN Delete( FILE_ADDRESS Address,BOOLEAN NoDestroy = False,BOOLEAN Zero = False ) = 0;

            VIRTUAL BOOLEAN Delete( VOID *Data,BOOLEAN NoDestroy = False,BOOLEAN Zero = False ) = 0;

            VIRTUAL BOOLEAN DeleteAll( BOOLEAN NoDestroy = False,BOOLEAN Zero = False ) = 0;

            VIRTUAL BOOLEAN Details( FILE_ADDRESS Address,BOOLEAN *Destroy = NULL,FILE_SIZE *Space = NULL ) = 0;

            VIRTUAL BOOLEAN Details( VOID *Data,BOOLEAN *Destroy = NULL,FILE_SIZE *Space = NULL ) = 0;

            VIRTUAL SBIT32 Feature( VOID ) = 0;

            VIRTUAL BOOLEAN GetError( SBIT32 *ErrorNumber ) = 0;

            VIRTUAL BOOLEAN New
                (
                FILE_ADDRESS                    *Address,
                VOID                            **Data,
                FILE_SIZE                       Size,
                FILE_SIZE                       *Space = NULL,
                BOOLEAN                         Destroy = False,
                BOOLEAN                         Zero = False
                ) = 0;

            VIRTUAL BOOLEAN Resize
                (
                FILE_ADDRESS                    *Address,
                VOID                            **Data,
                FILE_SIZE                       NewSize,
                FILE_SIZE                       *Space = NULL,
                BOOLEAN                         Zero = False
                ) = 0;

            VIRTUAL BOOLEAN SetError( SBIT32 ErrorNumber ) = 0;

            VIRTUAL BOOLEAN Size( FILE_SIZE *Space ) = 0;

            VIRTUAL BOOLEAN Walk( HEAP_WALK_FUNCTION WalkFunction,VOID *UserValue = NULL ) = 0;

            //
            //  The Rockall-DB transactional functionality.
            //
            VIRTUAL BOOLEAN CreateFile
                (
                CHAR                            *FileName,
                ENCRYPTION_FUNCTION             Function = NULL,
                VOID                            *UserValue = NULL
                ) = 0;

            VIRTUAL BOOLEAN CreateFile
                (
                WCHAR                           *FileName = NULL,
                ENCRYPTION_FUNCTION             Function = NULL,
                VOID                            *UserValue = NULL
                ) = 0;

            VIRTUAL BOOLEAN CreateRegion( SBIT32 *NewRegionID ) = 0;

            VIRTUAL BOOLEAN BeginTransaction
                (
                FILE_ADDRESS                    *NewTransactionID = NULL,
                EXPORT_MEMORY_LOCK              *UserLock = NULL
                ) = 0;

            VIRTUAL BOOLEAN ChangeRegion( SBIT32 NewRegionID,SBIT32 *OldRegionID = NULL ) = 0;

            VIRTUAL BOOLEAN ExclusiveView( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL ) = 0;

            VIRTUAL BOOLEAN EndExclusiveView( FILE_ADDRESS Address,BOOLEAN Reset = True ) = 0;

            VIRTUAL BOOLEAN JoinTransaction( FILE_ADDRESS TransactionID,EXPORT_MEMORY_LOCK *UserLock = NULL ) = 0;

            VIRTUAL BOOLEAN LeaveTransaction( VOID ) = 0;

            VIRTUAL BOOLEAN Update( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL ) = 0;

            VIRTUAL BOOLEAN View( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL ) = 0;

            VIRTUAL BOOLEAN EndView( FILE_ADDRESS Address,BOOLEAN Reset = True ) = 0;

            VIRTUAL BOOLEAN EndTransaction( BOOLEAN Abort = False,BOOLEAN Commit = False ) = 0;

            VIRTUAL BOOLEAN DeleteRegion( SBIT32 RegionID ) = 0;

            VIRTUAL BOOLEAN CloseFile( VOID ) = 0;

            //
            //  The Rockall-DB database functionality.
            //
            VIRTUAL BOOLEAN CreateShadowFile( CHAR *FileName,SBIT32 *ID = NULL,BOOLEAN Wait = True ) = 0;

            VIRTUAL BOOLEAN CreateShadowFile( WCHAR *FileName,SBIT32 *ID = NULL,BOOLEAN Wait = True ) = 0;

            VIRTUAL BOOLEAN CreateSnapshotFile( CHAR *FileName ) = 0;

            VIRTUAL BOOLEAN CreateSnapshotFile( WCHAR *FileName ) = 0;

            VIRTUAL BOOLEAN OpenFile
                (
                CHAR                            *FileName,
                BOOLEAN                         ReadOnly = False,
                ENCRYPTION_FUNCTION             Function = NULL,
                VOID                            *UserValue = NULL
                ) = 0;

            VIRTUAL BOOLEAN OpenFile
                (
                WCHAR                           *FileName,
                BOOLEAN                         ReadOnly = False,
                ENCRYPTION_FUNCTION             Function = NULL,
                VOID                            *UserValue = NULL
                ) = 0;

            VIRTUAL BOOLEAN Commit( VOID ) = 0;

            VIRTUAL BOOLEAN FileSize( FILE_SIZE *Size ) = 0;

            VIRTUAL BOOLEAN GetUserValue( SBIT32 ID,FILE_ADDRESS *Value ) = 0;

            VIRTUAL BOOLEAN SetUserValue( SBIT32 ID,FILE_ADDRESS Value ) = 0;

            VIRTUAL BOOLEAN Touch( FILE_ADDRESS Address ) = 0;

            VIRTUAL BOOLEAN CloseShadowFile( SBIT32 ID = 0 ) = 0;

            VIRTUAL ~VIRTUAL_HEAP( VOID ) { /* void */ };

            //
            //  Public inline functions.
            //
            BOOLEAN Feature( SBIT32 Features )
                { return ((this != NULL) ? ((Features & Feature()) == Features) : False); }

            //
            //  Public virtual functions.
            //
            VIRTUAL BOOLEAN New
                    (
                    VOID                        **Data,
                    FILE_SIZE                   Size,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( NULL,Data,Size,Space,Destroy,Zero )); }

            VIRTUAL BOOLEAN Resize
                    (
                    VOID                        **Data,
                    FILE_SIZE                   NewSize,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Zero = False
                    )
                { return (Resize( NULL,Data,NewSize,Space,Zero )); }

        private:
            //
            //  Disabled functionality.
            //
            VIRTUAL_HEAP( CONST VIRTUAL_HEAP & NoCopies );

            VOID operator=( CONST VIRTUAL_HEAP & NoCopies );
        };
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
typedef ROCKALL::VIRTUAL_HEAP VIRTUAL_HEAP;
#endif
#endif