#ifndef _MEMORY_LOCK_HPP_
#define _MEMORY_LOCK_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "ExportMemoryLock.hpp"
#include "PublicConstants.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Memory Lock.                                                                                      */
        /*                                                                                                    */
        /*  A memory lock is lock that is held in memory and is not suitable for use in files.                */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Memory Lock.                                                                              */
            /*                                                                                            */
            /*  A memory lock is lock that is held in memory and is not suitable for use in files.        */
            /*                                                                                            */
            /**********************************************************************************************/

    class MEMORY_LOCK
        {
            //
            //  Private data.
            //
            EXPORT_MEMORY_LOCK                  ExportedLock;

        public:
            //
            //  Public inline functions.
            //
            MEMORY_LOCK( VOID )
                { /* void */ }

            INLINE BOOLEAN ClaimExclusiveLock( BOOLEAN Wait = True )
                { return (ExportedLock.ClaimExclusiveLock( Wait )); }

            INLINE BOOLEAN ClaimSharedLock( BOOLEAN Wait = True )
                { return (ExportedLock.ClaimSharedLock( Wait )); }

            INLINE VOID ReleaseExclusiveLock( VOID )
                { ExportedLock.ReleaseExclusiveLock(); }

            INLINE VOID ReleaseSharedLock( VOID )
                { ExportedLock.ReleaseSharedLock(); }

            ~MEMORY_LOCK( VOID )
                { /* void */ }

            //
            //  Public inline functions.
            //
            STATIC INLINE SBIT32 Feature( VOID )
                { return (RockallMemoryLockRequirements); }

            INLINE BOOLEAN Feature( SBIT32 Features )
                { return ((Features & RockallMemoryLockFlags) == Features); }

        private:
            //
            //  Disabled functionality.
            //
            MEMORY_LOCK( CONST MEMORY_LOCK & NoCopies );

            VOID operator=( CONST MEMORY_LOCK & NoCopies );
        };
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
typedef ROCKALL::MEMORY_LOCK MEMORY_LOCK;
#endif
#endif