#ifndef _FIXED_STRING_HPP_
#define _FIXED_STRING_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "AutomaticViewScope.hpp"
#include "BaseTypes.hpp"
#include "Failure.hpp"
#include "FlexibleString.hpp"
#include "ReservedWords.hpp"
#include "String.hpp"
#include "SubString.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Fixed String.                                                                                     */
        /*                                                                                                    */
        /*  A fixed string has a known fixed size and so does not need to make use of the heap.               */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Fixed String.                                                                             */
            /*                                                                                            */
            /*  A fixed string has a known fixed size and so does not need to make use of the heap.       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> class FIXED_STRING : public SUB_STRING<HEAP,TYPE>
        {
            //
            //  Private data.
            //
            STRING<TYPE>                        FixedString;
            TYPE                                DoNotTouch[ MAX_SIZE ];

        public:
            //
            //  Public functions.
            //
            FIXED_STRING( VOID );
            FIXED_STRING( CONST FIXED_STRING<HEAP,TYPE,MAX_SIZE> & Value );
            FIXED_STRING( CONST SUB_STRING<HEAP,TYPE> & Value );
            FIXED_STRING( CONST TYPE *Value );

            BOOLEAN EndView( VOID );
            BOOLEAN EndView( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN Size( FILE_SIZE *Value );
            BOOLEAN Size( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_SIZE *Value );

            BOOLEAN View( TYPE **Text,FILE_SIZE *Size = NULL );
            BOOLEAN View( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,TYPE **Text,FILE_SIZE *Size = NULL );

            ~FIXED_STRING( VOID );

            //
            //  Public operators.
            //
            VOID operator=( CONST FIXED_STRING<HEAP,CHAR,MAX_SIZE> & String );
            VOID operator=( CONST FIXED_STRING<HEAP,WCHAR,MAX_SIZE> & String );
            VOID operator=( CONST SUB_STRING<HEAP,CHAR> & String );
            VOID operator=( CONST SUB_STRING<HEAP,WCHAR> & String );
            VOID operator=( CONST CHAR *Text );
            VOID operator=( CONST WCHAR *Text );

            VOID operator+=( CONST SUB_STRING<HEAP,TYPE> & String );
            VOID operator+=( CONST TYPE *Text );

            bool operator==( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator!=( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator< ( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator<=( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator> ( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator>=( CONST SUB_STRING<HEAP,TYPE> & String );

            bool operator==( CONST TYPE *Text );
            bool operator!=( CONST TYPE *Text );
            bool operator< ( CONST TYPE *Text );
            bool operator<=( CONST TYPE *Text );
            bool operator> ( CONST TYPE *Text );
            bool operator>=( CONST TYPE *Text );

            //
            //  Public internal functions.
            //
            BOOLEAN Resize( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,STRING<TYPE> **String,FILE_SIZE Size );

            BOOLEAN Update
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
                STRING<TYPE>                        **String,
                FILE_SIZE                           *Size = NULL
                );

            BOOLEAN View
                (
                AUTOMATIC_VIEW_SCOPE<HEAP>          *AutomaticViewScope,
                STRING<TYPE>                        **String,
                FILE_SIZE                           *Size = NULL
                );

        private:
            //
            //  Private static inline functions.
            //
            STATIC SNATIVE StartOfText( VOID )
                { return ((SNATIVE) & (((FIXED_STRING<HEAP,TYPE,MAX_SIZE>*) 0) -> FixedString.Text)); }
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'FIXED_STRING'.                                                           */
            /*                                                                                            */
            /*  The constructor for 'FIXED_STRING' prepares the class for use.                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> FIXED_STRING<HEAP,TYPE,MAX_SIZE>::FIXED_STRING( VOID )
        {
        AUTO SNATIVE MaxSize = ((sizeof(FIXED_STRING<HEAP,TYPE,MAX_SIZE>) - StartOfText()) / sizeof(TYPE));

        //
        //  Setup the parent class members.
        //
        Type = FixedStringType;

        //
        //  Setup the class members.
        //
        FixedString.MaxSize = MaxSize;
        FixedString.Used = 0;

        //
        //  Zero the 'Text'.
        //
        MemoryFunctions.Zero( & FixedString.Text,((SNATIVE) (FixedString.MaxSize * sizeof(TYPE))) );
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'FIXED_STRING'.                                                           */
            /*                                                                                            */
            /*  The constructor for 'FIXED_STRING' prepares the class for use.                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> FIXED_STRING<HEAP,TYPE,MAX_SIZE>::FIXED_STRING
            (
            CONST FIXED_STRING<HEAP,TYPE,MAX_SIZE>  & Value
            )
        {
        AUTO SNATIVE MaxSize = ((sizeof(FIXED_STRING<HEAP,TYPE,MAX_SIZE>) - StartOfText()) / sizeof(TYPE));

        //
        //  Setup the parent class members.
        //
        Type = FixedStringType;

        //
        //  Setup the class members.
        //
        FixedString.MaxSize = MaxSize;
        FixedString.Used = MaxSize;

        //
        //  We can now assign the new value to the string.
        //
        if ( ! Assignment( Value ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'FIXED_STRING'.                                                           */
            /*                                                                                            */
            /*  The constructor for 'FIXED_STRING' prepares the class for use.                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> FIXED_STRING<HEAP,TYPE,MAX_SIZE>::FIXED_STRING
            (
            CONST SUB_STRING<HEAP,TYPE>             & Value
            )
        {
        AUTO SNATIVE MaxSize = ((sizeof(FIXED_STRING<HEAP,TYPE,MAX_SIZE>) - StartOfText()) / sizeof(TYPE));

        //
        //  Setup the parent class members.
        //
        Type = FixedStringType;

        //
        //  Setup the class members.
        //
        FixedString.MaxSize = MaxSize;
        FixedString.Used = MaxSize;

        //
        //  We can now assign the new value to the string.
        //
        if ( ! Assignment( Value ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'FIXED_STRING'.                                                           */
            /*                                                                                            */
            /*  The constructor for 'FIXED_STRING' prepares the class for use.                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> FIXED_STRING<HEAP,TYPE,MAX_SIZE>::FIXED_STRING
            (
            CONST TYPE                          *Value
            )
        {
        AUTO SNATIVE MaxSize = ((sizeof(FIXED_STRING<HEAP,TYPE,MAX_SIZE>) - StartOfText()) / sizeof(TYPE));

        //
        //  Setup the parent class members.
        //
        Type = FixedStringType;

        //
        //  Setup the class members.
        //
        FixedString.MaxSize = MaxSize;
        FixedString.Used = MaxSize;

        //
        //  We can now assign the new value to the string.
        //
        if ( ! Assignment( Value ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End View.                                                                                 */
            /*                                                                                            */
            /*  End a view.                                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::EndView( VOID )
        { return True; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End View.                                                                                 */
            /*                                                                                            */
            /*  End a view.                                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::EndView
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        { return True; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> VOID FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator=
            (
            CONST FIXED_STRING<HEAP,CHAR,MAX_SIZE>  & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> VOID FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator=
            (
            CONST FIXED_STRING<HEAP,WCHAR,MAX_SIZE> & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> VOID FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator=
            (
            CONST SUB_STRING<HEAP,CHAR>         & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> VOID FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator=
            (
            CONST SUB_STRING<HEAP,WCHAR>        & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> VOID FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator=
            (
            CONST CHAR                          *Text
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = Text; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> VOID FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator=
            (
            CONST WCHAR                         *Text
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = Text; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '+='.                                                                            */
            /*                                                                                            */
            /*  The operator '+='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> VOID FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator+=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) += String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '+='.                                                                            */
            /*                                                                                            */
            /*  The operator '+='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> VOID FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator+=
            (
            CONST TYPE                          *Text
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) += Text; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '=='.                                                                            */
            /*                                                                                            */
            /*  The operator '=='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator==
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) == String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '!='.                                                                            */
            /*                                                                                            */
            /*  The operator '!='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator!=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) != String); }


            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<'.                                                                             */
            /*                                                                                            */
            /*  The operator '<'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator<
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) < String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<='.                                                                            */
            /*                                                                                            */
            /*  The operator '<='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator<=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) <= String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>'.                                                                             */
            /*                                                                                            */
            /*  The operator '>'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator>
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) > String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>='.                                                                            */
            /*                                                                                            */
            /*  The operator '>='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator>=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) >= String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '=='.                                                                            */
            /*                                                                                            */
            /*  The operator '=='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator==
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) == Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '!='.                                                                            */
            /*                                                                                            */
            /*  The operator '!='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator!=
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) != Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<'.                                                                             */
            /*                                                                                            */
            /*  The operator '<'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator<
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) < Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<='.                                                                            */
            /*                                                                                            */
            /*  The operator '<='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator<=
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) <= Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>'.                                                                             */
            /*                                                                                            */
            /*  The operator '>'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator>
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) > Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>='.                                                                            */
            /*                                                                                            */
            /*  The operator '>='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> bool FIXED_STRING<HEAP,TYPE,MAX_SIZE>::operator>=
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) >= Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Resize.                                                                                   */
            /*                                                                                            */
            /*  Resize a string.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::Resize
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            STRING<TYPE>                        **String,
            FILE_SIZE                           Size
            )
        {
        //
        //  We can't resize a 'FIXED_STRING' but if it is large enough still return 'True'.
        //
        (*String) = & FixedString;

        return (FixedString.MaxSize >= Size);
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the string.                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::Size
            (
            FILE_SIZE                           *Value
            )
        { return (Size( NULL,Value )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the string.                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::Size
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           *Value
            )
        {
        //
        //  We return the 'Size' of the 'FixedString'.
        //
        (*Value) = (FixedString.Used);

        return True;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update.                                                                                   */
            /*                                                                                            */
            /*  Update a string.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::Update
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            STRING<TYPE>                        **String,
            FILE_SIZE                           *Size
            )
        {
        //
        //  We simply return the 'FixedString' and optionally the 'Size'.
        //
        if ( Size == NULL )
            { (*String) = & FixedString; }
        else
            {
            (*String) = & FixedString;
            (*Size) = sizeof(FIXED_STRING<HEAP,TYPE,MAX_SIZE>);
            }

        return True;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a string.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::View
            (
            AUTOMATIC_VIEW_SCOPE<HEAP>          *AutomaticViewScope,
            STRING<TYPE>                        **String,
            FILE_SIZE                           *Size
            )
        {
        //
        //  We simply return the 'FixedString' and optionally the 'Size'.
        //
        if ( Size == NULL )
            { (*String) = & FixedString; }
        else
            {
            (*String) = & FixedString;
            (*Size) = sizeof(FIXED_STRING<HEAP,TYPE,MAX_SIZE>);
            }

        return True;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a string.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::View
            (
            TYPE                                **Text,
            FILE_SIZE                           *Size
            )
        { return (View( NULL,Text,Size )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a string.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> BOOLEAN FIXED_STRING<HEAP,TYPE,MAX_SIZE>::View
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TYPE                                **Text,
            FILE_SIZE                           *Size
            )
        {
        //
        //  The 'Text' and 'Size' are returned (if needed).
        //
        if ( Size == NULL )
            { (*Text) = FixedString.Text; }
        else
            {
            (*Text) = FixedString.Text;
            (*Size) = FixedString.Used;
            }

        return True;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'FIXED_STRING'.                                                            */
            /*                                                                                            */
            /*  The destructor for 'FIXED_STRING' cleans up the class.                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> FIXED_STRING<HEAP,TYPE,MAX_SIZE>::~FIXED_STRING( VOID )
        {
        //
        //  Zero the 'Text'.
        //
        MemoryFunctions.Zero( & FixedString.Text,((SNATIVE) (FixedString.Used * sizeof(TYPE))) );

        //
        //  Zero the class members.
        //
        FixedString.Used = 0;
        FixedString.MaxSize = 0;

        //
        //  Zero the parent class members.
        //
        Type = NoStringType;
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define FIXED_STRING ROCKALL::FIXED_STRING
#endif
#endif
