#ifndef _FLEXIBLE_STRING_HPP_
#define _FLEXIBLE_STRING_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "AutomaticViewScope.hpp"
#include "BaseTypes.hpp"
#include "Failure.hpp"
#include "FixedString.hpp"
#include "PlacementNew.hpp"
#include "ReservedWords.hpp"
#include "String.hpp"
#include "SubString.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Flexible String.                                                                                  */
        /*                                                                                                    */
        /*  A flexible string can vary in size and is stored on the heap.                                     */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Flexible String.                                                                          */
            /*                                                                                            */
            /*  A flexible string can vary in size and is stored on the heap.                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> class FLEXIBLE_STRING : public SUB_STRING<HEAP,TYPE>
        {
            //
            //  Private data.
            //
            FILE_ADDRESS                        Address;

        public:
            //
            //  Public functions.
            //
            FLEXIBLE_STRING( VOID );
            FLEXIBLE_STRING( CONST FLEXIBLE_STRING<HEAP,TYPE> & Value );
            FLEXIBLE_STRING( CONST SUB_STRING<HEAP,TYPE> & Value );
            FLEXIBLE_STRING( CONST TYPE *Value );

            BOOLEAN EndView( VOID );
            BOOLEAN EndView( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN Size( FILE_SIZE *Value );
            BOOLEAN Size( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_SIZE *Value );

            BOOLEAN View( TYPE **Text,FILE_SIZE *Size = NULL );
            BOOLEAN View( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,TYPE **Text,FILE_SIZE *Size = NULL );

            ~FLEXIBLE_STRING( VOID );

            //
            //  Public operators.
            //
            VOID operator=( CONST FLEXIBLE_STRING<HEAP,CHAR> & String );
            VOID operator=( CONST FLEXIBLE_STRING<HEAP,WCHAR> & String );
            VOID operator=( CONST SUB_STRING<HEAP,CHAR> & String );
            VOID operator=( CONST SUB_STRING<HEAP,WCHAR> & String );
            VOID operator=( CONST CHAR *Text );
            VOID operator=( CONST WCHAR *Text );

            VOID operator+=( CONST SUB_STRING<HEAP,TYPE> & String );
            VOID operator+=( CONST TYPE *Text );

            bool operator==( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator!=( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator< ( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator<=( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator> ( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator>=( CONST SUB_STRING<HEAP,TYPE> & String );

            bool operator==( CONST TYPE *Text );
            bool operator!=( CONST TYPE *Text );
            bool operator< ( CONST TYPE *Text );
            bool operator<=( CONST TYPE *Text );
            bool operator> ( CONST TYPE *Text );
            bool operator>=( CONST TYPE *Text );

            //
            //  Public internal functions.
            //
            BOOLEAN Resize( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,STRING<TYPE> **String,FILE_SIZE Size );

            BOOLEAN Update
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
                STRING<TYPE>                        **String,
                FILE_SIZE                           *Size = NULL
                );

            BOOLEAN View
                (
                AUTOMATIC_VIEW_SCOPE<HEAP>          *AutomaticViewScope,
                STRING<TYPE>                        **String,
                FILE_SIZE                           *Size = NULL
                );

        private:
            //
            //  Private static inline functions.
            //
            STATIC SNATIVE StartOfText( VOID )
                { return ((SNATIVE) & (((STRING<TYPE>*) 0) -> Text)); }
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'FLEXIBLE_STRING'.                                                        */
            /*                                                                                            */
            /*  The constructor for 'FLEXIBLE_STRING' prepares the class for use.                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> FLEXIBLE_STRING<HEAP,TYPE>::FLEXIBLE_STRING( VOID )
        {
        //
        //  Setup the parent class members.
        //
        Type = FlexibleStringType;

        //
        //  Setup the class members.
        //
        Address = NULL;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'FLEXIBLE_STRING'.                                                        */
            /*                                                                                            */
            /*  The constructor for 'FLEXIBLE_STRING' prepares the class for use.                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> FLEXIBLE_STRING<HEAP,TYPE>::FLEXIBLE_STRING
            (
            CONST FLEXIBLE_STRING<HEAP,TYPE>        & Value
            )
        {
        //
        //  Setup the parent class members.
        //
        Type = FlexibleStringType;

        //
        //  Setup the class members.
        //
        Address = NULL;

        //
        //  We can now assign the new value to the string.
        //
        if ( ! Assignment( Value ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'FLEXIBLE_STRING'.                                                        */
            /*                                                                                            */
            /*  The constructor for 'FLEXIBLE_STRING' prepares the class for use.                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> FLEXIBLE_STRING<HEAP,TYPE>::FLEXIBLE_STRING
            (
            CONST SUB_STRING<HEAP,TYPE>             & Value
            )
        {
        //
        //  Setup the parent class members.
        //
        Type = FlexibleStringType;

        //
        //  Setup the class members.
        //
        Address = NULL;

        //
        //  We can now assign the new value to the string.
        //
        if ( ! Assignment( Value ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'FLEXIBLE_STRING'.                                                        */
            /*                                                                                            */
            /*  The constructor for 'FLEXIBLE_STRING' prepares the class for use.                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> FLEXIBLE_STRING<HEAP,TYPE>::FLEXIBLE_STRING
            (
            CONST TYPE                          *Value
            )
        {
        //
        //  Setup the parent class members.
        //
        Type = FlexibleStringType;

        //
        //  Setup the class members.
        //
        Address = NULL;

        //
        //  We can now assign the new value to the string.
        //
        if ( ! Assignment( Value ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End View.                                                                                 */
            /*                                                                                            */
            /*  End a view.                                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::EndView( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (EndView( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End View.                                                                                 */
            /*                                                                                            */
            /*  End a view.                                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::EndView
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        { return (AutomaticHeapScope -> EndView( Address )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID FLEXIBLE_STRING<HEAP,TYPE>::operator=
            (
            CONST FLEXIBLE_STRING<HEAP,CHAR>    & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID FLEXIBLE_STRING<HEAP,TYPE>::operator=
            (
            CONST FLEXIBLE_STRING<HEAP,WCHAR>   & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID FLEXIBLE_STRING<HEAP,TYPE>::operator=
            (
            CONST SUB_STRING<HEAP,CHAR>         & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID FLEXIBLE_STRING<HEAP,TYPE>::operator=
            (
            CONST SUB_STRING<HEAP,WCHAR>        & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID FLEXIBLE_STRING<HEAP,TYPE>::operator=
            (
            CONST CHAR                          *Text
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = Text; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID FLEXIBLE_STRING<HEAP,TYPE>::operator=
            (
            CONST WCHAR                         *Text
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) = Text; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '+='.                                                                            */
            /*                                                                                            */
            /*  The operator '+='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID FLEXIBLE_STRING<HEAP,TYPE>::operator+=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) += String; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '+='.                                                                            */
            /*                                                                                            */
            /*  The operator '+='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID FLEXIBLE_STRING<HEAP,TYPE>::operator+=
            (
            CONST TYPE                          *Text
            )
        { (*((SUB_STRING<HEAP,TYPE>*) this)) += Text; }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '=='.                                                                            */
            /*                                                                                            */
            /*  The operator '=='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator==
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) == String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '!='.                                                                            */
            /*                                                                                            */
            /*  The operator '!='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator!=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) != String); }


            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<'.                                                                             */
            /*                                                                                            */
            /*  The operator '<'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator<
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) < String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<='.                                                                            */
            /*                                                                                            */
            /*  The operator '<='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator<=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) <= String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>'.                                                                             */
            /*                                                                                            */
            /*  The operator '>'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator>
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) > String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>='.                                                                            */
            /*                                                                                            */
            /*  The operator '>='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator>=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) >= String); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '=='.                                                                            */
            /*                                                                                            */
            /*  The operator '=='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator==
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) == Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '!='.                                                                            */
            /*                                                                                            */
            /*  The operator '!='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator!=
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) != Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<'.                                                                             */
            /*                                                                                            */
            /*  The operator '<'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator<
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) < Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<='.                                                                            */
            /*                                                                                            */
            /*  The operator '<='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator<=
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) <= Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>'.                                                                             */
            /*                                                                                            */
            /*  The operator '>'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator>
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) > Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>='.                                                                            */
            /*                                                                                            */
            /*  The operator '>='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool FLEXIBLE_STRING<HEAP,TYPE>::operator>=
            (
            CONST TYPE                          *Text
            )
        { return ((*((SUB_STRING<HEAP,TYPE>*) this)) >= Text); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Resize.                                                                                   */
            /*                                                                                            */
            /*  Resize a string.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::Resize
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            STRING<TYPE>                        **String,
            FILE_SIZE                           Size
            )
        {
        AUTO FILE_SIZE Space;

        //
        //  When a string is empty we set the 'Address' to 'NULL' to save space.  We may need to
        //  to create some new space now if the string is no longer empty.
        //
        if ( Address != NULL )
            {
            //
            //  When the new 'Size' is zero we can free any space associated with the string and
            //  set the 'Address' to 'NULL' to save space.
            //
            if ( Size > 0 )
                {
                //
                //  Let's see of the new value requires extra space.
                //
                if ( Size > (*String) -> MaxSize )
                    {
                    AUTO FILE_SIZE NewSize = ((*String) -> ComputeSize( Size ));

                    //
                    //  We now we 'Resize()' the string to make it large enough and update the
                    //  'MaxSize' to include the extra space.
                    //
                    if ( AutomaticHeapScope -> Resize( & Address,((VOID**) String),NewSize,& Space,True ) )
                        {
                        (*String) -> MaxSize = ((Space - StartOfText()) / sizeof(TYPE));

                        return True;
                        }
                    }
                else
                    { return True; }
                }
            else
                {
                //
                //  When the 'Size' is zero we can delete the string.
                //
                if ( AutomaticHeapScope -> Delete( Address ) )
                    {
                    //
                    //  Set the 'Address' to 'NULL' to save space and return 'NoString'.
                    //
                    Address = NULL;

                    return (((*String) = ((STRING<TYPE>*) & (*String) -> NoString)), True);
                    }
                else
                    { return (FAILURE( "Unable to delete string" )); }
                }
            }
        else
            {
            //
            //  When we have an empty string we will need to create some new space for it.
            //  However, if the new value is empty then we don't need to do anything and we
            //  can just exit.
            //
            if ( Size > 0 )
                {
                AUTO FILE_SIZE NewSize = ((*String) -> ComputeSize( Size ));

                //
                //  We now need we allocate some space for the new string.
                //
                if ( AutomaticHeapScope -> New( & Address,((VOID**) String),NewSize,& Space,False,True ) )
                    {
                    //
                    //  Call the constructor.
                    //
                    PLACEMENT_NEW( String,STRING<TYPE> );

                    //
                    //  Setup the string.
                    //
                    (*String) -> MaxSize = ((Space - StartOfText()) / sizeof(TYPE));
                    (*String) -> Used = 0;

                    return True;
                    }
                else
                    { return (((*String) = ((STRING<TYPE>*) & (*String) -> NoString)), False); }
                }
            else
                { return True; }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the string.                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::Size
            (
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Size( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the string.                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::Size
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           *Value
            )
        {
        //
        //  We only need to 'View()' when we have a non-empty string.
        //
        if ( Address != NULL )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO STRING<TYPE> *String;

            //
            //  We need to 'View()' the 'String'.
            //
            if ( AutomaticViewScope.View( Address,((VOID**) & String) ) )
                { return (((*Value) = String -> Used), True); }

            return False;
            }
        else
            { return (((*Value) = 0), True); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update.                                                                                   */
            /*                                                                                            */
            /*  Update a string.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::Update
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            STRING<TYPE>                        **String,
            FILE_SIZE                           *Size
            )
        {
        //
        //  We only need to 'Update()' when we have a non-empty string.
        //
        if ( Address != NULL )
            { return (AutomaticHeapScope -> Update( Address,((VOID**) String),Size )); }
        else
            { return (((*String) = ((STRING<TYPE>*) & (*String) -> NoString)), True); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a string.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::View
            (
            AUTOMATIC_VIEW_SCOPE<HEAP>          *AutomaticViewScope,
            STRING<TYPE>                        **String,
            FILE_SIZE                           *Size
            )
        {
        //
        //  We only need to 'View()' when we have a non-empty string.
        //
        if ( Address != NULL )
            { return (AutomaticViewScope -> View( Address,((VOID**) String),Size )); }
        else
            { return (((*String) = ((STRING<TYPE>*) & (*String) -> NoString)), True); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a string.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::View
            (
            TYPE                                **Text,
            FILE_SIZE                           *Size
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (View( & AutomaticHeapScope,Text,Size ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a string.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN FLEXIBLE_STRING<HEAP,TYPE>::View
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TYPE                                **Text,
            FILE_SIZE                           *Size
            )
        {
        //
        //  We only need to 'View()' when we have a non-empty string.
        //
        if ( Address != NULL )
            {
            AUTO STRING<TYPE> *String;

            //
            //  We need to 'View()' the 'String'.
            //
            if ( AutomaticHeapScope -> View( Address,((VOID**) & String) ) )
                {
                //
                //  The 'Text' and 'Size' are returned (if needed).
                //
                if ( Size == NULL )
                    { (*Text) = String -> Text; }
                else
                    {
                    (*Text) = String -> Text;
                    (*Size) = String -> Used;
                    }

                return True;
                }

            return False;
            }
        else
            {
            //
            //  The 'Text' and 'Size' are returned (if needed).
            //
            if ( Size == NULL )
                { (*Text) = ((TYPE*) (STRING<TYPE>::NoString.Text)); }
            else
                {
                (*Text) = ((TYPE*) (STRING<TYPE>::NoString.Text));
                (*Size) = (STRING<TYPE>::NoString.Used);
                }

            return True;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'FLEXIBLE_STRING'.                                                         */
            /*                                                                                            */
            /*  The destructor for 'FLEXIBLE_STRING' cleans up the class.                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> FLEXIBLE_STRING<HEAP,TYPE>::~FLEXIBLE_STRING( VOID )
        {
        //
        //  We do not need to do anything if the string is empty.
        //
        if ( Address != NULL )
            {
            AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;
            AUTO STRING<TYPE> *String;

            //
            //  Delete the string.
            //
            if
                    (
                    (AutomaticHeapScope.BeginTransaction())
                        &&
                    (AutomaticHeapScope.Update( Address,((VOID**) & String) ))
                    )
                {
                //
                //  Call the destructor.
                //
                PLACEMENT_DELETE( String,STRING<TYPE> );

                if ( ! AutomaticHeapScope.Delete( Address ) )
                    { FAILURE( "Unable to delete string" ); }
                }
            else
                { FAILURE( "Unable to update string" ); }
            }

        //
        //  Zero all class members.
        //
        Type = NoStringType;
        Address = NULL;
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define FLEXIBLE_STRING ROCKALL::FLEXIBLE_STRING
#endif
#endif