#ifndef _HASH_HPP_
#define _HASH_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "AutomaticLockScope.hpp"
#include "AutomaticTransactionScope.hpp"
#include "AutomaticViewScope.hpp"
#include "Failure.hpp"
#include "FileLock.hpp"
#include "FlexibleString.hpp"
#include "MemoryFunctions.hpp"
#include "MemoryLock.hpp"
#include "NoLock.hpp"
#include "PlacementNew.hpp"
#include "RockallDelete.hpp"
#include "RockallNew.hpp"
#include "TransactionLock.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Hash.                                                                                             */
        /*                                                                                                    */
        /*  A hash is a simple table that directly maps keys to values.                                       */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Hash.                                                                                     */
            /*                                                                                            */
            /*  A hash is a simple table that directly maps keys to values.                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> class HASH
        {
            //
            //  Private constants.
            //
            STATIC CONST SBIT32 BranchSize      = 64;
            STATIC CONST SBIT32 DefaultSize     = (BranchSize * 1024);
            STATIC CONST SBIT32 TwigSize        = 1;
            STATIC CONST SBIT32 ViewSize        = 8192;

            //
            //  The root signature.
            //
            STATIC CONST SBIT64 Signature1      = ((sizeof(KEY) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature2      = (((Signature1 + sizeof(LOCK)) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature3      = (((Signature2 + sizeof(VALUE)) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature4      = (((Signature3 + BranchSize) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature5      = (((Signature4 + TwigSize) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature       = (0x0000100000000000 + Signature5);

            //
            //  Private types.
            //
            typedef TRANSACTION_LOCK<HEAP> TRANSACTIONAL_LOCK;

            typedef struct
                {
                KEY                             Key;
                VALUE                           Value;
                }
            LEAF;

            typedef struct
                {
                SBIT32                          MaxSize;
                SBIT32                          Used;

                LEAF                            Leaf[ TwigSize ];
                }
            TWIG;

            typedef struct
                {
                //
                //  Data locks.
                //
                LOCK                            BranchLock;

                //
                //  Protected data.
                //
                FILE_ADDRESS                    Address[ BranchSize ];
                SBIT32                          Used;
                }
            BRANCH;

            typedef struct
                {
                FILE_ADDRESS                    Address;
                FILE_SIZE                       Mask;
                FILE_SIZE                       Size;
                }
            TRUNK;

            typedef struct
                {
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope;
                HASH<HEAP,LOCK,KEY,VALUE>       *Class;
                FILE_ADDRESS                    *Hash;
                TRUNK                           *Trunk;
                }
            CONTEXT;

            typedef struct
                {
                //
                //  Data locks.
                //
                LOCK                            ClassLock;

                //
                //  Protected data.
                //
                FILE_SIZE                       DefaultSize;
                FILE_SIZE                       Signature;
                TRUNK                           Trunk;
                }
            ROOT;

            //
            //  Private data.
            //
            BOOLEAN                             Active;
            FILE_ADDRESS                        Address;

            //
            //  Private static data.
            //
            STATIC MEMORY_FUNCTIONS<SNATIVE>    MemoryFunctions;

        public:
            //
            //  public types.
            //
            typedef BOOLEAN (*HASH_WALK_FUNCTION)
                (
                KEY                             & Key,
                VALUE                           & Value,
                VOID                            *UserValue
                );

            //
            //  Public functions.
            //
            HASH( FILE_SIZE Size = DefaultSize );
            HASH( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_SIZE Size = DefaultSize );

            BOOLEAN DeleteAll( VOID );
            BOOLEAN DeleteAll( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN DeleteKey( CONST KEY & Key );
            BOOLEAN DeleteKey( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST KEY & Key );

            BOOLEAN FindKey( CONST KEY & Key,VALUE *Value );
            BOOLEAN FindKey( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST KEY & Key,VALUE *Value );

            BOOLEAN NewKey( CONST KEY & Key,CONST VALUE & Value,BOOLEAN Overwrite = False );

            BOOLEAN NewKey
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                CONST KEY                       & Key,
                CONST VALUE                     & Value,
                BOOLEAN                         Overwrite = False
                );

            BOOLEAN Size( FILE_SIZE *Value );
            BOOLEAN Size( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_SIZE *Value );

            BOOLEAN UpdateKey( CONST KEY & Key,CONST VALUE & Value );
            BOOLEAN UpdateKey( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST KEY & Key,CONST VALUE & Value );

            BOOLEAN Walk( HASH_WALK_FUNCTION WalkFunction,VOID *UserValue = NULL );

            BOOLEAN Walk
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                HASH_WALK_FUNCTION              WalkFunction,
                VOID                            *UserValue = NULL
                );

            ~HASH( VOID );

        private:
            //
            //  Private functions.
            //
            BOOLEAN AdjustTrunkSize( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,TRUNK *Trunk );

            FILE_ADDRESS CalculateHashValue
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                VOID                            *Key,
                FILE_SIZE                       Size
                );

            FILE_ADDRESS CalculateHashValue
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                FLEXIBLE_STRING<HEAP,CHAR>      *Key,
                FILE_SIZE                       Size
                );

            FILE_ADDRESS CalculateHashValue
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                FLEXIBLE_STRING<HEAP,WCHAR>     *Key,
                FILE_SIZE                       Size
                );

            BOOLEAN DeleteKey
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                FILE_ADDRESS                    *Address,
                SBIT32                          Offset,
                TWIG                            *Twig
                );

            BOOLEAN DeleteTrunk( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,TRUNK *Trunk );

            BOOLEAN DeleteTwig( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_ADDRESS *Address );

            BOOLEAN FindBranchAddress
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                FILE_ADDRESS                    *Address,
                KEY                             *Key,
                FILE_ADDRESS                    *Hash,
                FILE_ADDRESS                    *HashKey,
                TRUNK                           *Trunk
                );

            BOOLEAN FindTwigKey( KEY *Key,SBIT32 *Offset,TWIG *Twig );

            BOOLEAN NewBranch
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                BOOLEAN                         *Full,
                FILE_ADDRESS                    *Hash,
                KEY                             *Key,
                BOOLEAN                         Overwrite,
                TRUNK                           *Trunk,
                VALUE                           *Value
                );

            BOOLEAN CreateTrunk( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,TRUNK *Trunk,FILE_SIZE Size );

            BOOLEAN NewTwig
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                FILE_ADDRESS                    *Address,
                KEY                             *Key,
                TWIG                            **Twig,
                VALUE                           *Value
                );

            BOOLEAN WalkBranches
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                TRUNK                           *Trunk,
                HASH_WALK_FUNCTION              WalkFunction,
                VOID                            *UserValue = NULL
                );

            BOOLEAN WalkTwigs
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                BRANCH                          *Branch,
                HASH_WALK_FUNCTION              WalkFunction,
                VOID                            *UserValue
                );

            //
            //  Private inline functions.
            //
            INLINE FILE_ADDRESS ComputeHashValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,KEY *Key )
                { return (CalculateHashValue( AutomaticHeapScope,Key,sizeof(KEY) )); }

            INLINE BOOLEAN MatchingKeys( KEY *Key1,KEY *Key2 )
                { return ((*Key1) == (*Key2)); }

            //
            //  Private static functions.
            //
            STATIC BOOLEAN CopyTrunk( KEY & Key,VALUE & Value,VOID *UserValue );

            //
            //  Disabled functionality.
            //
            HASH( CONST HASH & NoCopies );

            VOID operator=( CONST HASH & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Class Static Data.                                                                        */
            /*                                                                                            */
            /*  The class static data needs to be initialised before it's ready for use.                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> MEMORY_FUNCTIONS<SNATIVE>
        HASH<HEAP,LOCK,KEY,VALUE>::MemoryFunctions;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'HASH'.                                                                   */
            /*                                                                                            */
            /*  The constructor for 'HASH' prepares the class for use.                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> HASH<HEAP,LOCK,KEY,VALUE>::HASH( FILE_SIZE Size )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        //
        //  When we create a hash we ensure that the 'DefaultSize' is reasonable.
        //
        if ( (Size >= Kilobyte) && ((Size & (Size-1)) == 0)  )
            {
            //
            //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
            //  we fail to highlight the issue.
            //
            if ( AutomaticHeapScope.Feature( (LOCK::Feature()) ) )
                {
                //
                //  Setup class members.
                //
                Active = False;
                Address = NULL;

                //
                //  Where necessary we automatically wrap all the changes within a new transaction.
                //
                if ( AutomaticHeapScope.BeginTransaction() )
                    {
                    AUTO ROOT *Root;

                    //
                    //  We create a new 'Root' on the heap and store the 'Address'.
                    //
                    if ( ROCKALL_NEW<ROOT>::New( & AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                        {
                        //
                        //  We now setup the 'Root' structure.
                        //
                        Root -> DefaultSize = Size;
                        Root -> Signature = Signature;

                        //
                        //  Zero the 'Trunk'.
                        //
                        MemoryFunctions.Zero( & Root -> Trunk,sizeof(TRUNK) );

                        Active = True;
                        }
                    }
                }
            else
                { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
            }
        else
            { FAILURE( "The size must be a power of 2 and greater or equal to than 1K" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'HASH'.                                                                   */
            /*                                                                                            */
            /*  The constructor for 'HASH' prepares the class for use.                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> HASH<HEAP,LOCK,KEY,VALUE>::HASH
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           Size
            )
        {
        //
        //  When we create a hash we ensure that the 'DefaultSize' is reasonable.
        //
        if ( (Size >= Kilobyte) && ((Size & (Size-1)) == 0)  )
            {
            //
            //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
            //  we fail to highlight the issue.
            //
            if ( AutomaticHeapScope -> Feature( (LOCK::Feature()) ) )
                {
                AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

                //
                //  Setup class members.
                //
                Active = False;
                Address = NULL;

                //
                //  Where necessary we automatically wrap all the changes within a new transaction.
                //
                if ( AutomaticTransactionScope.BeginTransaction() )
                    {
                    AUTO ROOT *Root;

                    //
                    //  We create a new 'Root' on the heap and store the 'Address'.
                    //
                    if ( ROCKALL_NEW<ROOT>::New( AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                        {
                        //
                        //  We now setup the 'Root' structure.
                        //
                        Root -> DefaultSize = Size;
                        Root -> Signature = Signature;

                        //
                        //  Zero the 'Trunk'.
                        //
                        MemoryFunctions.Zero( & Root -> Trunk,sizeof(TRUNK) );

                        Active = True;
                        }
                    }
                }
            else
                { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
            }
        else
            { FAILURE( "The size must be a power of 2 and greater or equal to than 1K" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Adjust Trunk Size.                                                                        */
            /*                                                                                            */
            /*  Adjust the size of the trunk.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::AdjustTrunkSize
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TRUNK                               *Trunk1
            )
        {
        AUTO TRUNK Trunk2 = { 0 };

        //
        //  Create a new empty 'Trunk2' that is larger than 'Trunk1'.
        //
        if ( CreateTrunk( AutomaticHeapScope,& Trunk2,(Trunk1 -> Size * FastGrowthFactor) ) )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO FILE_ADDRESS *Hash2;

            //
            //  We will need to 'View()' the 'Hash'.
            //
            if ( AutomaticViewScope.View( Trunk2.Address,((VOID**) & Hash2) ) )
                {
                AUTO CONTEXT Context = { AutomaticHeapScope,this,Hash2,& Trunk2 };

                //
                //  We walk 'Trunk1' and insert all the entries into 'Trunk2'.
                //
                if ( WalkBranches( AutomaticHeapScope,Trunk1,CopyTrunk,((VOID*) & Context )) )
                    {
                    //
                    //  Now we have sucessfully created a new larger 'Trunk2' and copied
                    //  across all of the contents of 'Trunk1'.  We can now delete the
                    //  original 'Trunk1' as it is no longer needed.
                    //
                    if ( DeleteTrunk( AutomaticHeapScope,Trunk1 ) )
                        { return (((*Trunk1) = Trunk2), True); }
                    else
                        { FAILURE( "Unable to delete trunk" ); }
                    }
                }

            //
            //  We will need to release the view before we delete 'Trunk2'.
            //
            AutomaticViewScope.Release();

            //
            //  Delete 'Trunk2' if we have any problems.
            //
            if ( ! DeleteTrunk( AutomaticHeapScope,& Trunk2 ) )
                { FAILURE( "Unable to delete trunk" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Calculate Hash Value.                                                                     */
            /*                                                                                            */
            /*  Calculate the hash value.                                                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> FILE_ADDRESS HASH<HEAP,LOCK,KEY,VALUE>::CalculateHashValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            VOID                                *Key,
            FILE_SIZE                           Size
            )
        {
        AUTO UBIT8 *KeyByte = ((UBIT8*) Key);
        AUTO FILE_ADDRESS HashValue = 0;
        AUTO FILE_ADDRESS Count;

        //
        //  We compute the 'HashValue' using every 'KeyByte' to ensure a smooth distribution even
        //  when the values are very similar.
        //
        for ( Count=0;Count < Size;Count ++ )
            {
            //
            //  Compute the 'HashValue' for the current 'KeyByte' and add it the the 'HashValue'.
            //
            HashValue = (((HashValue | 1) + KeyByte[ Count ]) * PrimeSBit64);
            HashValue = ((HashValue >> 32) + (HashValue << 32));
            HashValue &= HashMaskSBit64;
            }

        return HashValue;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Calculate Hash Value.                                                                     */
            /*                                                                                            */
            /*  Calculate the hash value.                                                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> FILE_ADDRESS HASH<HEAP,LOCK,KEY,VALUE>::CalculateHashValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FLEXIBLE_STRING<HEAP,CHAR>          *Key,
            FILE_SIZE                           Size
            )
        {
        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
        AUTO FILE_ADDRESS HashValue = 0;
        AUTO UBIT8 *KeyByte;

        //
        //  We have to be careful with a 'FLEXIBLE_STRING' to get the value of the string and not
        //  the pointer to the string.  If not, we can end up hashing the pointer to the string which
        //  can produce different hash values for matching strings.
        //
        if ( Key -> View( & AutomaticViewScope,((STRING<CHAR>**) & KeyByte),& Size ) )
            {
            AUTO FILE_ADDRESS Count;

            //
            //  We compute the 'HashValue' using every 'KeyByte' to ensure a smooth distribution even
            //  when the values are very similar.
            //
            for ( Count=0;Count < Size;Count ++ )
                {
                //
                //  Compute the 'HashValue' for the current 'KeyByte' and add it the the 'HashValue'.
                //
                HashValue = (((HashValue | 1) + KeyByte[ Count ]) * PrimeSBit64);
                HashValue = ((HashValue >> 32) + (HashValue << 32));
                HashValue &= HashMaskSBit64;
                }
            }
        else
            { FAILURE( "Unable to view string" ); }

        return HashValue;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Calculate Hash Value.                                                                     */
            /*                                                                                            */
            /*  Calculate the hash value.                                                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> FILE_ADDRESS HASH<HEAP,LOCK,KEY,VALUE>::CalculateHashValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FLEXIBLE_STRING<HEAP,WCHAR>         *Key,
            FILE_SIZE                           Size
            )
        {
        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
        AUTO FILE_ADDRESS HashValue = 0;
        AUTO UBIT8 *KeyByte;

        //
        //  We have to be careful with a 'FLEXIBLE_STRING' to get the value of the string and not
        //  the pointer to the string.  If not, we can end up hashing the pointer to the string which
        //  can produce different hash values for matching strings.
        //
        if ( Key -> View( & AutomaticViewScope,((STRING<WCHAR>**) & KeyByte),& Size ) )
            {
            AUTO FILE_ADDRESS Count;

            //
            //  We compute the 'HashValue' using every 'KeyByte' to ensure a smooth distribution even
            //  when the values are very similar.
            //
            for ( Count=0;Count < Size;Count ++ )
                {
                //
                //  Compute the 'HashValue' for the current 'KeyByte' and add it the the 'HashValue'.
                //
                HashValue = (((HashValue | 1) + KeyByte[ Count ]) * PrimeSBit64);
                HashValue = ((HashValue >> 32) + (HashValue << 32));
                HashValue &= HashMaskSBit64;
                }
            }
        else
            { FAILURE( "Unable to view string" ); }

        return HashValue;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy Trunk.                                                                               */
            /*                                                                                            */
            /*  Copy an existing trunk to a new trunk.                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::CopyTrunk
            (
            KEY                                 & Key,
            VALUE                               & Value,
            VOID                                *UserValue
            )
        {
        AUTO CONTEXT *Context = ((CONTEXT*) UserValue);

        //
        //  We copy every 'Key' and 'Value' into the new 'Trunk'.
        //
        return
            (
            Context -> Class -> NewBranch
                (
                Context -> AutomaticHeapScope,
                NULL,
                Context -> Hash,
                ((KEY*) & Key),
                False,
                Context -> Trunk,
                ((VALUE*) & Value)
                )
            );
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Create Trunk.                                                                             */
            /*                                                                                            */
            /*  Create a new trunk.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::CreateTrunk
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TRUNK                               *Trunk,
            FILE_SIZE                           Size
            )
        {
        AUTO FILE_SIZE Branches = (Size / BranchSize);
        AUTO FILE_ADDRESS Address;
        AUTO FILE_ADDRESS *Hash;

        //
        //  Create a new 'Trunk' and all the related 'Branches'.
        //
        if ( ROCKALL_NEW<FILE_ADDRESS>::New( AutomaticHeapScope,& Address,& Hash,Branches,NULL,False,True ) )
            {
            AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
            AUTO FILE_SIZE Count;

            //
            //  The 'Trunk' points to a collection of 'Branches' that in turn point to a
            //  number of 'Twigs'.  We create all of the 'Branches' now to minimize the
            //  number of exclusive locks we will need later.
            //
            for ( Count=0;Count < Branches;Count ++ )
                {
                AUTO BRANCH *Branch;

                //
                //  We will delete the 'Trunk' if we can't create all the 'Branches'.
                //
                if ( ! ROCKALL_NEW<BRANCH>::New( AutomaticHeapScope,& Hash[ Count ],& Branch,1,NULL,False,True ) )
                    {
                    //
                    //  Delete any existing 'Branches'.
                    //
                    for ( Count--;Count >= 0;Count -- )
                        {
                        if ( ! ROCKALL_DELETE<BRANCH>::Delete( AutomaticHeapScope,Hash[ Count ] ) )
                            { FAILURE( "Unable to delete a branch" ); }
                        }

                    //
                    //  Delete the 'Trunk'.
                    //
                    if ( ! ROCKALL_DELETE<FILE_ADDRESS>::Delete( AutomaticHeapScope,Address,Branches ) )
                        { FAILURE( "Unable to delete a trunk" ); }

                    return False;
                    }
                }

            //
            //  We have now created the 'Trunk' and all the related 'Branches' so
            //  update the related structure.
            //
            Trunk -> Address = Address;
            Trunk -> Mask = ((Branches * BranchSize)-1);
            Trunk -> Size = (Branches * BranchSize);

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the hash values.                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::DeleteAll( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (DeleteAll( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the hash values.                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::DeleteAll
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the hash.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Delete the 'Trunk'.
                //
                return (DeleteTrunk( AutomaticHeapScope,& Root -> Trunk ));
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Key.                                                                               */
            /*                                                                                            */
            /*  Delete a key in the hash table.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::DeleteKey
            (
            CONST KEY                           & Key
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (DeleteKey( & AutomaticHeapScope,Key ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Key.                                                                               */
            /*                                                                                            */
            /*  Delete a Key in the hash table.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::DeleteKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Key
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the hash.
            //
            if
                    (
                    (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  Let's ensure there is a 'Trunk'.
                //
                if ( Root -> Trunk.Address != NULL )
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                    AUTO FILE_ADDRESS *Hash;

                    //
                    //  We will need to 'View()' the 'Hash'.
                    //
                    if ( AutomaticViewScope2.View( Root -> Trunk.Address,((VOID**) & Hash) ) )
                        {
                        AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
                        AUTO FILE_ADDRESS Address1;
                        AUTO FILE_ADDRESS HashKey;

                        //
                        //  Find the 'Branch' address in the 'Hash'.
                        //
                        if
                                (
                                FindBranchAddress
                                    (
                                    AutomaticHeapScope,
                                    & Address1,
                                    ((KEY*) & Key),
                                    Hash,
                                    & HashKey,
                                    & Root -> Trunk
                                    )
                                )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope );
                            AUTO BRANCH *Branch;

                            //
                            //  We will need to 'ExclusiveView()' the 'Branch'.
                            //
                            if ( AutomaticViewScope3.ExclusiveView( Address1,((VOID**) & Branch) ) )
                                {
                                AUTO LOCK *BranchLock = & Branch -> BranchLock;
                                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( BranchLock );
                                AUTO FILE_ADDRESS *Address2 = & Branch -> Address[ (HashKey % BranchSize) ];

                                //
                                //  Let's ensure there is a 'Twig'.
                                //
                                if ( (*Address2) != NULL )
                                    {
                                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope4( AutomaticHeapScope );
                                    AUTO TWIG *Twig;

                                    //
                                    //  We will need to 'ExclusiveView()' the 'Twig'.
                                    //
                                    if ( AutomaticViewScope4.ExclusiveView( (*Address2),((VOID**) & Twig) ) )
                                        {
                                        AUTO SBIT32 Offset;

                                        //
                                        //  Find the 'Offset' of the 'Key' in the 'Twig'.
                                        //
                                        if ( FindTwigKey( ((KEY*) & Key),& Offset,Twig ) )
                                            {
                                            //
                                            //  We will need to 'Update()' the 'Branch' and 'Twig'.
                                            //
                                            if
                                                    (
                                                    (AutomaticHeapScope -> Update( Address1,((VOID**) & Branch) ))
                                                        &&
                                                    (AutomaticHeapScope -> Update( (*Address2),((VOID**) & Twig) ))
                                                    )
                                                {
                                                //
                                                //  We need to 'Release()' any views on the 'Twig' just
                                                //  in case we delete it.
                                                //
                                                AutomaticViewScope4.Release();

                                                //
                                                //  Delete the 'Key' in the 'Twig' or if it is the last
                                                //  'Key' in the 'Twig' then delete the entire 'Twig'.
                                                //
                                                if ( DeleteKey( AutomaticHeapScope,Address2,Offset,Twig ) )
                                                    {
                                                    //
                                                    //  Update the number of 'Key' values in the 'Twig'.
                                                    //
                                                    if ( (-- Branch -> Used) >= 0 )
                                                        { return True; }
                                                    else
                                                        { FAILURE( "Negative use count" ); }
                                                    }
                                                }
                                            }
                                        else
                                            { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                                        }
                                    }
                                else
                                    { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                                }
                            }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                        }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Key.                                                                               */
            /*                                                                                            */
            /*  Delete a key from the hash table.                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::DeleteKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_ADDRESS                        *Address,
            SBIT32                              Offset,
            TWIG                                *Twig
            )
        {
        AUTO LEAF (*View)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);

        //
        //  We ensure that we have a valid 'Offset'.
        //
        if ( (Offset >= 0) && (Twig -> Used > 0) && (Offset < Twig -> Used) )
            {
            //
            //  If this 'Key' is the last entry we can delete the entire 'Twig'.
            //
            if ( Twig -> Used <= 1 )
                {
                AUTO SBIT32 Count;

                //
                //  Call any extra destructors.
                //
                for ( Count=(Twig -> MaxSize-1);Count >= TwigSize;Count -- )
                    { PLACEMENT_DELETE( & Twig -> Leaf[ Count ],LEAF ); }

                //
                //  Delete the 'Twig'.
                //
                if ( ROCKALL_DELETE<TWIG>::Delete( AutomaticHeapScope,Twig ) )
                    {
                    //
                    //  Zero the reference to the 'Twig' and exit.
                    //
                    (*Address) = NULL;

                    return True;
                    }
                else
                    { FAILURE( "Unable to delete twig" ); }
                }
            else
                {
                AUTO LEAF *Zero = & Twig -> Leaf[ (-- Twig -> Used) ];
                AUTO SBIT32 Count;

                //
                //  Slide any other entries down the array to overwrite the deleted 'Key'.
                //
                for ( Count=Offset;Count < Twig -> Used;Count ++ )
                    { Twig -> Leaf[ Count ] = Twig -> Leaf[ (Count+1) ]; }

                //
                //  A 'Leaf' contains user types so we have to be careful how we zero it.
                //
                PLACEMENT_DELETE( Zero,LEAF );

                MemoryFunctions.Zero( Zero,sizeof(LEAF) );

                PLACEMENT_NEW( Zero,LEAF );

                return True;
                }
            }
        else
            { FAILURE( "Unable to delete twig" ); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Trunk.                                                                             */
            /*                                                                                            */
            /*  Delete the trunk.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::DeleteTrunk
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TRUNK                               *Trunk
            )
        {
        //
        //  Let's ensure there is a 'Trunk'.
        //
        if ( Trunk -> Address != NULL  )
            {
            AUTO FILE_ADDRESS *Hash;

            //
            //  We will need to 'Update()' the 'Hash'.
            //
            if ( AutomaticHeapScope -> Update( Trunk -> Address,((VOID**) & Hash) ) )
                {
                AUTO FILE_ADDRESS (*View1)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
                AUTO FILE_SIZE Branches = (Trunk -> Size / BranchSize);
                AUTO FILE_SIZE Count1;

                //
                //  We walk the 'Hash' and delete all the 'Branches'.
                //
                for ( Count1=0;Count1 < Branches;Count1 ++ )
                    {
                    AUTO BRANCH *Branch;

                    //
                    //  We will need to 'Update()' every 'Branch'.
                    //
                    if ( AutomaticHeapScope -> Update( Hash[ Count1 ],((VOID**) & Branch) ) )
                        {
                        AUTO FILE_SIZE Count2;

                        //
                        //  Walk the 'Branch' and delete every 'Twig'.
                        //
                        for ( Count2=0;Count2 < BranchSize;Count2 ++ )
                            {
                            AUTO FILE_ADDRESS *Address = & Branch -> Address[ Count2 ];

                            //
                            //  Let's ensure there is a 'Twig'.
                            //
                            if ( (*Address) != NULL )
                                {
                                if ( ! DeleteTwig( AutomaticHeapScope,Address ) )
                                    { FAILURE( "Unable to delete a twig" ); }
                                }
                            }

                        //
                        //  Delete the 'Branch'.
                        //
                        if ( ! ROCKALL_DELETE<BRANCH>::Delete( AutomaticHeapScope,Branch ) )
                            { FAILURE( "Unable to delete a branch" ); }
                        }
                    else
                        { FAILURE( "Unable to update a branch" ); }
                    }

                //
                //  Delete the 'Hash'.
                //
                if ( ROCKALL_DELETE<FILE_ADDRESS>::Delete( AutomaticHeapScope,Hash,Branches ) )
                    {
                    //
                    //  Zero the 'Trunk'.
                    //
                    MemoryFunctions.Zero( Trunk,sizeof(TRUNK) );

                    return True;
                    }
                else
                    { FAILURE( "Unable to delete the hash table" ); }
                }
            else
                { FAILURE( "Unable to update the hash table" ); }
            }
        else
            { return True; }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Twig.                                                                              */
            /*                                                                                            */
            /*  Delete a twig.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::DeleteTwig
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_ADDRESS                        *Address
            )
        {
        //
        //  Let's ensure that we have a 'Twig'.
        //
        if ( (*Address) != NULL  )
            {
            AUTO TWIG *Twig;

            //
            //  We will need to 'Update()' the 'Twig'.
            //
            if ( AutomaticHeapScope -> Update( (*Address),((VOID**) & Twig) ) )
                {
                AUTO LEAF (*View)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
                AUTO SBIT32 Count;

                //
                //  Call any extra destructors.
                //
                for ( Count=(Twig -> MaxSize-1);Count >= TwigSize;Count -- )
                    { PLACEMENT_DELETE( & Twig -> Leaf[ Count ],LEAF ); }

                //
                //  Delete the 'Twig'.
                //
                if ( ROCKALL_DELETE<TWIG>::Delete( AutomaticHeapScope,Twig ) )
                    {
                    //
                    //  Zero the refrence to the 'Twig'.
                    //
                    (*Address) = NULL;

                    return True;
                    }
                else
                    { FAILURE( "Unable to delete a twig" ); }
                }
            }
        else
            { return True; }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Branch Address.                                                                      */
            /*                                                                                            */
            /*  Find the address of the branch in the hash table.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::FindBranchAddress
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_ADDRESS                        *Address,
            KEY                                 *Key,
            FILE_ADDRESS                        *Hash,
            FILE_ADDRESS                        *HashKey,
            TRUNK                               *Trunk
            )
        {
        AUTO FILE_ADDRESS (*View1)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
        AUTO FILE_ADDRESS HashValue = ((ComputeHashValue( AutomaticHeapScope,Key ) + 1) & Trunk -> Mask);

        //
        //  We always return the 'HashKey' to allow the 'Hash' table to be updated (if needed).
        //
        (*Address) = Hash[ (HashValue / BranchSize) ];
        (*HashKey) = HashValue;

        return ((*Address) != NULL);
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Key.                                                                                 */
            /*                                                                                            */
            /*  Find a key value.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::FindKey
            (
            CONST KEY                           & Key,
            VALUE                               *Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (FindKey( & AutomaticHeapScope,Key,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Key.                                                                                 */
            /*                                                                                            */
            /*  Find a key value.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::FindKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Key,
            VALUE                               *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the hash.
            //
            if
                    (
                    (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  Let's ensure there is a 'Trunk'.
                //
                if ( Root -> Trunk.Address != NULL )
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                    AUTO FILE_ADDRESS *Hash;

                    //
                    //  We will need to 'View()' the 'Hash'.
                    //
                    if ( AutomaticViewScope2.View( Root -> Trunk.Address,((VOID**) & Hash) ) )
                        {
                        AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
                        AUTO FILE_ADDRESS Address1;
                        AUTO FILE_ADDRESS HashKey;

                        //
                        //  Find the 'Branch' address in the 'Hash'.
                        //
                        if
                                (
                                FindBranchAddress
                                    (
                                    AutomaticHeapScope,
                                    & Address1,
                                    ((KEY*) & Key),
                                    Hash,
                                    & HashKey,
                                    & Root -> Trunk
                                    )
                                )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope );
                            AUTO BRANCH *Branch;

                            //
                            //  We will need to 'View()' the 'Branch'.
                            //
                            if ( AutomaticViewScope3.View( Address1,((VOID**) & Branch) ) )
                                {
                                AUTO LOCK *BranchLock = & Branch -> BranchLock;
                                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock3( BranchLock );
                                AUTO FILE_ADDRESS Address2 = Branch -> Address[ (HashKey % BranchSize) ];

                                //
                                //  Let's ensure there is a 'Twig'.
                                //
                                if ( Address2 != NULL )
                                    {
                                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope4( AutomaticHeapScope );
                                    AUTO TWIG *Twig;

                                    //
                                    //  We will need to 'View()' the 'Twig'.
                                    //
                                    if ( AutomaticViewScope4.View( Address2,((VOID**) & Twig) ) )
                                        {
                                        AUTO SBIT32 Offset;

                                        //
                                        //  Find the 'Offset' of the 'Key' in the 'Twig'.
                                        //
                                        if ( FindTwigKey( ((KEY*) & Key),& Offset,Twig ) )
                                            {
                                            //
                                            //  We have now found the 'Key' so return the 'Value'.
                                            //
                                            (*Value) = Twig -> Leaf[ Offset ].Value;

                                            return True;
                                            }
                                        else
                                            { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                                        }
                                    }
                                else
                                    { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                                }
                            }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                        }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Twig Key.                                                                            */
            /*                                                                                            */
            /*  Find a key in a twig.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::FindTwigKey
            (
            KEY                                 *Key,
            SBIT32                              *Offset,
            TWIG                                *Twig
            )
        {
        AUTO LEAF (*View)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
        AUTO SBIT32 Count;

        //
        //  Find the 'Key' in the 'Twig'.
        //
        for ( Count=0;Count < Twig -> Used;Count ++ )
            {
            AUTO LEAF *Current = & Twig -> Leaf[ Count ];

            //
            //  We compare the all the keys.
            //
            if ( MatchingKeys( Key,& Current -> Key ) )
                {
                //
                //  We have found a match so return the offset in the current 'Leaf'.
                //
                (*Offset) = Count;

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Branch.                                                                               */
            /*                                                                                            */
            /*  A new key and value are inserted into a branch.                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::NewBranch
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            BOOLEAN                             *Full,
            FILE_ADDRESS                        *Hash,
            KEY                                 *Key,
            BOOLEAN                             Overwrite,
            TRUNK                               *Trunk,
            VALUE                               *Value
            )
        {
        AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
        AUTO FILE_ADDRESS Address1;
        AUTO FILE_ADDRESS HashKey;

        //
        //  Find the 'Branch' address in the 'Hash'.
        //
        if ( FindBranchAddress( AutomaticHeapScope,& Address1,Key,Hash,& HashKey,Trunk ) )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO BRANCH *Branch;

            //
            //  We need to 'ExclusiveView()' the 'Branch'.  We can't 'Update()' it because
            //  we would deadlock if it was full when we tried to grow the 'Hash'.
            //
            if ( AutomaticViewScope.ExclusiveView( Address1,((VOID**) & Branch) ) )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Branch -> BranchLock );

                //
                //  Let's make sure the 'Branch' is not 'Full'.
                //
                if ( (Full == NULL) || (! ((*Full) = (Branch -> Used > BranchSize))) )
                    {
                    //
                    //  Now we can 'Update()' the 'Branch'.
                    //
                    if ( AutomaticHeapScope -> Update( Address1,((VOID**) & Branch) ) )
                        {
                        AUTO FILE_ADDRESS *Address2 = & Branch -> Address[ (HashKey % BranchSize) ];
                        AUTO TWIG *Twig = NULL;

                        //
                        //  Let's ensure we don't alreday have a 'Twig'.
                        //
                        if ( (*Address2) == NULL )
                            {
                            //
                            //  We need to create a new 'Twig'.
                            //
                            if ( NewTwig( AutomaticHeapScope,Address2,Key,& Twig,Value ) )
                                {
                                //
                                //  Update the statistics.
                                //
                                Branch -> Used ++;

                                return True;
                                }
                            }
                        else
                            {
                            //
                            //  We will need to 'Update()' the 'Twig'.
                            //
                            if ( AutomaticHeapScope -> Update( (*Address2),((VOID**) & Twig) ) )
                                {
                                AUTO SBIT32 Offset;

                                //
                                //  Find the 'Offset' of the 'Key' in the 'Twig'.
                                //
                                if ( ! FindTwigKey( Key,& Offset,Twig ) )
                                    {
                                    //
                                    //  We need to add a new 'Leaf' into the 'Twig'.
                                    //
                                    if ( NewTwig( AutomaticHeapScope,Address2,Key,& Twig,Value ) )
                                        {
                                        //
                                        //  Update the statistics.
                                        //
                                        Branch -> Used ++;

                                        return True;
                                        }
                                    }
                                else
                                    {
                                    //
                                    //  We will update the 'Value' if 'Overwrite' is set.
                                    //
                                    if ( Overwrite )
                                        {
                                        //
                                        //  Update the 'Value' and exit.
                                        //
                                        Twig -> Leaf[ Offset ].Value = (*Value);

                                        return True;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Key.                                                                                  */
            /*                                                                                            */
            /*  A new value is inserted into the hash table.                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::NewKey
            (
            CONST KEY                           & Key,
            CONST VALUE                         & Value,
            BOOLEAN                             Overwrite
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (NewKey( & AutomaticHeapScope,Key,Value,Overwrite ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Key.                                                                                  */
            /*                                                                                            */
            /*  A new value is inserted into the hash table.                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::NewKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Key,
            CONST VALUE                         & Value,
            BOOLEAN                             Overwrite
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()', 'ExclusiveView()',
        //  'Update()' and 'View()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO BOOLEAN SharedNavigation = True;
            AUTO BOOLEAN NoTrunk = False;
            AUTO FILE_SIZE NewSize = 0;

            //
            //  Clearly, it is much better to use 'SharedNavigation' where we can to avoid
            //  exclusive locks.
            //
            if ( SharedNavigation )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
                AUTO ROOT *Root;

                //
                //  We 'View()' the 'Root' as we need to examine it.  We claimed the
                //  'TransactionLock' above to prevent threads in the same transaction
                //  from changing the hash.
                //
                if
                        (
                        (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                    //
                    //  Now that we have all the locks let's see if we have a 'Trunk'. If not,
                    //  we will try to create it later.
                    //
                    if ( Root -> Trunk.Address != NULL )
                        {
                        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                        AUTO FILE_ADDRESS *Hash;

                        //
                        //  We will need to 'View()' the 'Hash'.
                        //
                        if ( AutomaticViewScope2.View( Root -> Trunk.Address,((VOID**) & Hash) ) )
                            {
                            AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
                            AUTO BOOLEAN Full = False;

                            //
                            //  We now try to add the 'Key' and 'Value' to the 'Hash'.  We will
                            //  fail if we are 'Full' and will need to adjust the size of the 'Hash'.
                            //
                            if
                                    (
                                    NewBranch
                                        (
                                        AutomaticHeapScope,
                                        & Full,
                                        Hash,
                                        ((KEY*) & Key),
                                        Overwrite,
                                        & Root -> Trunk,
                                        ((VALUE*) & Value)
                                        )
                                    )
                                { return True; }
                            else
                                {
                                //
                                //  If we are 'Full' we can not use 'SharedNavigation'.
                                //
                                if ( Full )
                                    { NewSize = (Root -> Trunk.Size * FastGrowthFactor); SharedNavigation = False; }
                                else
                                    { return (AutomaticHeapScope -> SetError( ErrorDuplicateKey )); }
                                }
                            }
                        else
                            { return False; }
                        }
                    else
                        { NewSize = 0; NoTrunk = True; SharedNavigation = False; }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
                }

            //
            //  If we have 'NoTrunk' then we try to create a 'Trunk'.  We can then add the
            //  new 'Key' and 'Value' later.
            //
            if ( NoTrunk )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock3( & TransactionLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope );
                AUTO ROOT *Root;

                //
                //  We 'ExclusiveView()' the 'Root' as we need to examine it.  We claimed
                //  the 'TransactionLock' above to prevent threads in the same transaction
                //  from changing the hash.
                //
                if
                        (
                        (AutomaticViewScope3.ExclusiveView( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock4( & Root -> ClassLock );

                    //
                    //  When we have multiple threads in a single transaction we need to claim a
                    //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                    //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                    //  and then claim an exclusive lock to prevent them from returning.  Once we
                    //  have claimed the exclusive lock we can release the 'TransactionLock' as
                    //  we no longer need it.
                    //
                    AutomaticExclusiveLock3.Release();

                    //
                    //  We recheck that we still have 'NoTrunk' as it might have been
                    //  created while we were claiming the locks.  If so, then we will
                    //  call 'CreateTrunk()' we can then add the new 'Key' and 'Value'
                    //  later.
                    //
                    if
                            (
                                (Root -> Trunk.Address != NULL)
                                    ||
                                (
                                (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                                    &&
                                (CreateTrunk( AutomaticHeapScope,& Root -> Trunk,Root -> DefaultSize ))
                                )
                            )
                        { NewSize = (Root -> Trunk.Size); NoTrunk = False; }
                    else
                        { return False; }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
                }

            //
            //  We will claim exclusive locks and tehn add the new 'Key' and 'Value' as
            //  long as we still have a 'Trunk'.
            //
            if ( (! NoTrunk) && (! SharedNavigation) )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock5( & TransactionLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope4( AutomaticHeapScope );
                AUTO ROOT *Root;

                //
                //  We 'ExclusiveViewView()' the 'Root' as we need to examine it.  We
                //  claimed the 'TransactionLock' above to prevent threads in the same
                //  transaction from changing the hash.
                //
                if
                        (
                        (AutomaticViewScope4.ExclusiveView( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock6( & Root -> ClassLock );

                    //
                    //  When we have multiple threads in a single transaction we need to claim a
                    //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                    //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                    //  and then claim an exclusive lock to prevent them from returning.  Once we
                    //  have claimed the exclusive lock we can release the 'TransactionLock' as
                    //  we no longer need it.
                    //
                    AutomaticExclusiveLock5.Release();

                    //
                    //  Let's ensure we have a 'Trunk'.
                    //
                    if ( Root -> Trunk.Address != NULL )
                        {
                        //
                        //  We will grow the 'Trunk' if necessary.
                        //
                        if
                                (
                                    (Root -> Trunk.Size >= NewSize)
                                        ||
                                    (
                                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                                        &&
                                    (AdjustTrunkSize( AutomaticHeapScope,& Root -> Trunk ))
                                    )
                                )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope5( AutomaticHeapScope );
                            AUTO FILE_ADDRESS *Hash;

                            //
                            //  We will need to 'View()' the 'Hash'.
                            //
                            if ( AutomaticViewScope5.View( Root -> Trunk.Address,((VOID**) & Hash) ) )
                                {
                                AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);

                                //
                                //  We now try to add the 'Key' and 'Value' to the 'Hash'.  We
                                //  do not pass 'Full' and so we force the addition of the 'Key'
                                //  and 'Value' even if 'Branch' is full.
                                //
                                if
                                        (
                                        NewBranch
                                            (
                                            AutomaticHeapScope,
                                            NULL,
                                            Hash,
                                            ((KEY*) & Key),
                                            Overwrite,
                                            & Root -> Trunk,
                                            ((VALUE*) & Value)
                                            )
                                        )
                                    { return True; }
                                else
                                    { return (AutomaticHeapScope -> SetError( ErrorDuplicateKey )); }
                                }
                            else
                                { return False; }
                            }
                        else
                            { return False; }
                        }
                    else
                        { return False; }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Twig.                                                                                 */
            /*                                                                                            */
            /*  A new key and value are inserted into a twig.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::NewTwig
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_ADDRESS                        *Address,
            KEY                                 *Key,
            TWIG                                **Twig,
            VALUE                               *Value
            )
        {
        //
        //  Let's see if we have an existing 'Twig'.
        //
        if ( (*Twig) == NULL )
            {
            //
            //  Create a new 'Twig'.
            //
            if ( ROCKALL_NEW<TWIG>::New( AutomaticHeapScope,Address,Twig,1,NULL,False,True ) )
                {
                //
                //  Set up the 'Twig' and the 'Leaf'.
                //
                (*Twig) -> MaxSize = TwigSize;
                (*Twig) -> Used = 1;

                (*Twig) -> Leaf[0].Key = (*Key);
                (*Twig) -> Leaf[0].Value = (*Value);

                return True;
                }
            }
        else
            {
            //
            //  Let's see if the 'Twig' already has enough space.
            //
            if ( (*Twig) -> Used >= (*Twig) -> MaxSize )
                {
                AUTO SBIT32 NewMaxSize = ((*Twig) -> MaxSize * GrowthFactor);
                AUTO SBIT32 NewSize = (sizeof(TWIG) + ((NewMaxSize - TwigSize) * sizeof(LEAF)));
                AUTO FILE_SIZE Space;

                //
                //  Expand the size of the 'Twig'.
                //
                if ( AutomaticHeapScope -> Resize( Address,((VOID**) Twig),NewSize,& Space,True ) )
                    {
                    AUTO LEAF (*View)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) (*Twig) -> Leaf);
                    AUTO SBIT32 MaxSize = ((*Twig) -> MaxSize);
                    AUTO SBIT32 Count;

                    //
                    //  Compute the actual 'MaxSize' from the 'Space'.
                    //
                    (*Twig) -> MaxSize = ((SBIT32) (((Space - sizeof(TWIG)) / sizeof(LEAF)) + TwigSize));

                    //
                    //  Call any extra constructors.
                    //
                    for ( Count=MaxSize;Count < (*Twig) -> MaxSize;Count ++ )
                        { PLACEMENT_NEW( & (*Twig) -> Leaf[ Count ],LEAF ); }
                    }
                else
                    { return False; }
                }

            //
            //  Set up the 'Leaf'.
            //
            (*Twig) -> Leaf[ ((*Twig) -> Used) ].Key = (*Key);
            (*Twig) -> Leaf[ ((*Twig) -> Used) ].Value = (*Value);

            (*Twig) -> Used ++;

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the hash.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::Size
            (
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Size( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the hash.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::Size
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the hash.
            //
            if
                    (
                    (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  We ensure there is a 'Hash' table.
                //
                if ( Root -> Trunk.Address != NULL )
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                    AUTO FILE_ADDRESS *Hash;

                    //
                    //  We will need to 'View()' the 'Hash'.
                    //
                    if ( AutomaticViewScope2.View( Root -> Trunk.Address,((VOID**) & Hash) ) )
                        {
                        AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
                        AUTO FILE_SIZE Branches = (Root -> Trunk.Size / BranchSize);
                        AUTO FILE_SIZE TotalSize = 0;
                        AUTO FILE_SIZE Count;

                        //
                        //  We walk every 'Branch' in the 'Hash'.
                        //
                        for ( Count=0;Count < Branches;Count ++ )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope );
                            AUTO BRANCH *Branch;

                            //
                            //  We will need to 'View()' the 'Branch'.
                            //
                            if ( AutomaticViewScope3.View( Hash[ Count ],((VOID**) & Branch) ) )
                                {
                                AUTO LOCK *BranchLock = & Branch -> BranchLock;
                                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock3( BranchLock );

                                //
                                //  Compute the total.
                                //
                                TotalSize += Branch -> Used;
                                }
                            else
                                { return False; }
                            }

                        //
                        //  Update 'Value' and exit.
                        //
                        (*Value) = TotalSize;

                        return True;
                        }
                    }
                else
                    {
                    //
                    //  Update 'Value' and exit.
                    //
                    (*Value) = 0;

                    return True;
                    }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }


            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update Key.                                                                               */
            /*                                                                                            */
            /*  Update a key value.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::UpdateKey
            (
            CONST KEY                           & Key,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (UpdateKey( & AutomaticHeapScope,Key,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update Key.                                                                               */
            /*                                                                                            */
            /*  Update a key value.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::UpdateKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Key,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the hash.
            //
            if
                    (
                    (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  Let's ensure there is a 'Trunk'.
                //
                if ( Root -> Trunk.Address != NULL )
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                    AUTO FILE_ADDRESS *Hash;

                    //
                    //  We will need to 'View()' the 'Hash'.
                    //
                    if ( AutomaticViewScope2.View( Root -> Trunk.Address,((VOID**) & Hash) ) )
                        {
                        AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
                        AUTO FILE_ADDRESS Address1;
                        AUTO FILE_ADDRESS HashKey;

                        //
                        //  Find the 'Branch' address in the 'Hash'.
                        //
                        if
                                (
                                FindBranchAddress
                                    (
                                    AutomaticHeapScope,
                                    & Address1,
                                    ((KEY*) & Key),
                                    Hash,
                                    & HashKey,
                                    & Root -> Trunk
                                    )
                                )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope );
                            AUTO BRANCH *Branch;

                            //
                            //  We will need to 'View()' the 'Branch'.
                            //
                            if ( AutomaticViewScope3.View( Address1,((VOID**) & Branch) ) )
                                {
                                AUTO LOCK *BranchLock = & Branch -> BranchLock;
                                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock3( BranchLock );
                                AUTO FILE_ADDRESS Address2 = Branch -> Address[ (HashKey % BranchSize) ];

                                //
                                //  Let's ensure there is a 'Twig'.
                                //
                                if ( Address2 != NULL )
                                    {
                                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope4( AutomaticHeapScope );
                                    AUTO TWIG *Twig;

                                    //
                                    //  We will need to 'ExclusiveView()' the 'Twig'.
                                    //
                                    if ( AutomaticViewScope4.ExclusiveView( Address2,((VOID**) & Twig) ) )
                                        {
                                        AUTO SBIT32 Offset;

                                        //
                                        //  Find the 'Offset' of the 'Key' in the 'Twig'.
                                        //
                                        if ( FindTwigKey( ((KEY*) & Key),& Offset,Twig ) )
                                            {
                                            //
                                            //  We will need to 'Update()' the 'Twig'.
                                            //
                                            if ( AutomaticHeapScope -> Update( Address2,((VOID**) & Twig) ) )
                                                {
                                                //
                                                //  We have now found the 'Key' so update the 'Value'.
                                                //
                                                Twig -> Leaf[ Offset ].Value = Value;

                                                return True;
                                                }
                                            }
                                        else
                                            { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                                        }
                                    }
                                else
                                    { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                                }
                            }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                        }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk all the keys and values in the hash table.                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::Walk
            (
            HASH_WALK_FUNCTION                  WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Walk( & AutomaticHeapScope,WalkFunction,UserValue ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk all the keys and values in the hash table.                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::Walk
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            HASH_WALK_FUNCTION                  WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the hash.
            //
            if
                    (
                    (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  We ensure there is a 'Hash' table.
                //
                if ( Root -> Trunk.Address != NULL )
                    { return (WalkBranches( AutomaticHeapScope,& Root -> Trunk,WalkFunction,UserValue )); }
                else
                    { return True; }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk Branches.                                                                            */
            /*                                                                                            */
            /*  Walk all the keys and values in the branches.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::WalkBranches
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TRUNK                               *Trunk,
            HASH_WALK_FUNCTION                  WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
        AUTO FILE_ADDRESS *Hash;

        //
        //  We will need to 'View()' the 'Hash'.
        //
        if ( AutomaticViewScope1.View( Trunk -> Address,((VOID**) & Hash) ) )
            {
            AUTO FILE_ADDRESS (*View)[ ViewSize ] = ((FILE_ADDRESS(*)[ ViewSize ]) Hash);
            AUTO FILE_SIZE Branches = (Trunk -> Size / BranchSize);
            AUTO FILE_SIZE Count;

            //
            //  We walk every 'Branch' in the 'Hash'.
            //
            for ( Count=0;Count < Branches;Count ++ )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  We will need to 'View()' the 'Branch'.
                //
                if ( AutomaticViewScope2.View( Hash[ Count ],((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  We walk every 'Twig' in a 'Branch'.
                    //
                    if ( ! WalkTwigs( AutomaticHeapScope,Branch,WalkFunction,UserValue ) )
                        { return False; }
                    }
                else
                    { return False; }
                }

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk Twigs.                                                                               */
            /*                                                                                            */
            /*  Walk all the keys and values in the twigs.                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> BOOLEAN HASH<HEAP,LOCK,KEY,VALUE>::WalkTwigs
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            BRANCH                              *Branch,
            HASH_WALK_FUNCTION                  WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO FILE_ADDRESS Count1;

        //
        //  We walk every 'Twig' in a 'Branch'.
        //
        for ( Count1=0;Count1 < BranchSize;Count1 ++ )
            {
            AUTO FILE_ADDRESS Address = (Branch -> Address[ Count1 ]);

            //
            //  Let's ensure we have a 'Twig'.
            //
            if ( Address != NULL )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope ) ;
                AUTO TWIG *Twig;

                //
                //  We will need to 'View()' every 'Twig'.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Twig) ) )
                    {
                    AUTO LEAF (*View)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
                    AUTO SBIT32 Count2;

                    //
                    //  We walk all the 'KeyValues'.
                    //
                    for ( Count2=0;Count2 < Twig -> Used;Count2 ++ )
                        {
                        AUTO LEAF *Leaf = & Twig -> Leaf[ Count2 ];

                        //
                        //  We simply don't trust user supplied code so we prepare for trouble.
                        //
#ifndef DISABLE_CRT_FEATURES
                        try
                            {
#endif
                            //
                            //  Execute the user supplied walk function.
                            //
                            if ( ! WalkFunction( Leaf -> Key,Leaf -> Value,UserValue ) )
                                { return False; }
#ifndef DISABLE_CRT_FEATURES
                            }
                        catch ( ... )
                            {
                            //
                            //  We will catch and skip all errors detected in the user supplied
                            //  walk function.
                            //
                            return (AutomaticHeapScope -> SetError( ErrorExceptionCaught ));
                            }
#endif
                        }
                    }
                else
                    { return False; }
                }
            }

        return True;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'HASH'.                                                                    */
            /*                                                                                            */
            /*  The destructor for 'HASH' cleans up the class.                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE> HASH<HEAP,LOCK,KEY,VALUE>::~HASH( VOID )
        {
        //
        //  We need to ensure that the class is 'Active'.  If not, then it is safe to exit.
        //
        if ( Active )
            {
            AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;
            AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( & AutomaticHeapScope );
            AUTO TRANSACTIONAL_LOCK TransactionLock( & AutomaticHeapScope );

            //
            //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
            //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
            //  should be automatically optimized away for heaps that do not support this
            //  functionality by any decent compiler, due to the way the code has been written.
            //
            if
                    (
                    (Active)
                        &&
                    (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
                AUTO ROOT *Root;

                //
                //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
                //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
                //  above to prevent threads in the same transaction from accessing the hash.
                //
                if
                        (
                        (AutomaticHeapScope.Update( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                    //
                    //  Delete the 'Trunk'.
                    //
                    if ( DeleteTrunk( & AutomaticHeapScope,& Root -> Trunk ) )
                        {
                        //
                        //  Zero the 'Root' structure.
                        //
                        Root -> Signature = 0;
                        Root -> DefaultSize = 0;

                        //
                        //  We are no longer 'Active'.
                        //
                        Active = False;

                        //
                        //  Release the lock on the 'Root'.
                        //
                        AutomaticExclusiveLock2.Release();

                        //
                        //  Delete the 'Root'.
                        //
                        if ( ROCKALL_DELETE<ROOT>::Delete( & AutomaticHeapScope,Address ) )
                            {
                            //
                            //  Zero class members.
                            //
                            Address = NULL;
                            Active = False;
                            }
                        else
                            { FAILURE( "Unable to delete the root" ); }
                        }
                    else
                        { FAILURE( "Unable to delete the trunk" ); }
                    }
                else
                    { FAILURE( "Unable to update the root" ); }
                }
            else
                { FAILURE( "Unable to update the root" ); }
            }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define HASH ROCKALL::HASH
#endif
#endif