#ifndef _ROW_SET_HPP_
#define _ROW_SET_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "Set.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Row Set.                                                                                          */
        /*                                                                                                    */
        /*  The Rockall-DB library does not in general support duplicate records.  This short-comming         */
        /*  is overcome by 'ROW_SET' which permits a single key to be mapped to multiple values.  The         */
        /*  class does not store the values itself but rather the address of each value.  This allows         */
        /*  different sized rows to be dynamically added and removed simply and quickly.                      */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Row Set.                                                                                  */
            /*                                                                                            */
            /*  The Rockall-DB library does not in general support duplicate records.  This               */
            /*  short-comming is overcome by this class which permits a single key to be mapped to        */
            /*  multiple values.  The class does not store the values itself but rather the address       */
            /*  of each value.  This allows different sized rows to be dynamically added and removed      */
            /*  simply and quickly.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK> class ROW_SET : public SET<HEAP,LOCK,FILE_ADDRESS>
        {
            //
            //  Private constants.
            //
            CONST STATIC SBIT32 HeapSegmentSize = 65536;
            CONST STATIC SBIT32 HeapSegmentMask = (HeapSegmentSize-1);

        public:
            //
            //  Public functions.
            //
            ROW_SET( VOID ) { /* void */ }
            ROW_SET( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN Touch( VOID );
            BOOLEAN Touch( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            ~ROW_SET( VOID ) { /* void */ }

        private:
            //
            //  Disabled functionality.
            //
            ROW_SET( CONST ROW_SET & NoCopies );

            VOID operator=( CONST ROW_SET & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'ROW_SET'.                                                                */
            /*                                                                                            */
            /*  The constructor for 'ROW_SET' prepares the class for use.                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK> ROW_SET<HEAP,LOCK>::ROW_SET( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope )
            : SET<HEAP,LOCK,FILE_ADDRESS>( AutomaticHeapScope )
        { /* void */ }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Touch.                                                                                    */
            /*                                                                                            */
            /*  Touch all the rows.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK> BOOLEAN ROW_SET<HEAP,LOCK>::Touch( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Touch( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Touch.                                                                                    */
            /*                                                                                            */
            /*  Touch all the rows.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK> BOOLEAN ROW_SET<HEAP,LOCK>::Touch
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO BOOLEAN Result = True;

        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( AutomaticHeapScope -> Feature( RockallFileSupport ) )
            {
            AUTO FILE_ADDRESS *Rows;
            AUTO FILE_SIZE Size;

            //
            //  We now 'View()' the 'Rows' in the 'Set'.
            //
            if ( View( AutomaticHeapScope,& Rows,& Size ) )
                {
                AUTO SBIT32 Count;

                //
                //  We 'Touch()' all the 'Rows'.
                //
                for ( Count=0;Count < Size;/* void */ )
                    {
                    AUTO FILE_ADDRESS Current = Rows[ (Count ++) ];

                    //
                    //  We 'Touch()' the 'Current' row.
                    //
                    if ( ! AutomaticHeapScope -> Touch( Current ) )
                        { Result = False; }

                    //
                    //  We skip past any 'Rows' in the same segment as we do not need
                    //  to 'Touch()' them.
                    //
                    while
                            (
                            (Count < Size)
                                &&
                            ((Current & ~HeapSegmentMask) == (Rows[ Count ] & ~HeapSegmentMask))
                            )
                        { Count ++; }
                    }

                //
                //  We can now 'EndView()'.
                //
                if ( ! EndView( AutomaticHeapScope ) )
                    { FAILURE( "Unable to end view" ); }
                }
            }

        return Result;
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define ROW_SET ROCKALL::ROW_SET
#endif
#endif