#ifndef _SET_HPP_
#define _SET_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "AutomaticLockScope.hpp"
#include "AutomaticTransactionScope.hpp"
#include "AutomaticViewScope.hpp"
#include "Failure.hpp"
#include "FileLock.hpp"
#include "MemoryLock.hpp"
#include "NoLock.hpp"
#include "PlacementNew.hpp"
#include "RockallDelete.hpp"
#include "RockallNew.hpp"
#include "SubSet.hpp"
#include "TransactionLock.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Set.                                                                                              */
        /*                                                                                                    */
        /*  A set is a collection of values stored in sorted order.  Although sets can be used in a           */
        /*  number of ways its main role is to allow Rockall-DB to support duplicate records.  The            */
        /*  libary classes such as 'Hash' and 'Tree' do not support duplicate keys.  This short-comming       */
        /*  is overcome by sets which permit a single key to be mapped to multiple values.  The class         */
        /*  can either store the values itself or the address of each value.  The latter allows different     */
        /*  sized rows or values to be dynamically added and removed simply and quickly.                      */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Set.                                                                                      */
            /*                                                                                            */
            /*  A set is a collection of values stored in sorted order.  Although sets can be used in     */
            /*  a number of ways its main role is to allow Rockall-DB to support duplicate records.       */
            /*  The libary classes such as 'Hash' and 'Tree' do not support duplicate keys.  This         */
            /*  short-comming is overcome by sets which permit a single key to be mapped to multiple      */
            /*  values.  The class can either store the values itself or the address of each value.       */
            /*  The latter allows different sized rows or values to be dynamically added and removed      */
            /*  simply and quickly.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> class SET
        {
            //
            //  Private constants.
            //
            STATIC CONST SBIT32 CacheSize       = 32;
            STATIC CONST SBIT32 DefaultSize     = 1;
            STATIC CONST SBIT32 MaxExtendedSize = (CacheSize * 4);
            STATIC CONST SBIT32 MinExtendedSize = 16;
            STATIC CONST SBIT32 ViewSize        = 8192;

            //
            //  The root signature.
            //
            STATIC CONST SBIT64 Signature1      = ((sizeof(LOCK) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature2      = (((Signature1 + sizeof(VALUE)) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature3      = (((Signature2 + CacheSize) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature       = (0x0000300000000000 + Signature3);

            //
            //  Private types.
            //
            typedef SUB_SET<CacheSize,VALUE>    CACHE;
            typedef TRANSACTION_LOCK<HEAP>      TRANSACTIONAL_LOCK;

            typedef enum
                {
                SimpleRowSet                    = 0x0001,
                ExtendedRowSet                  = 0x0002,
                ComplexRowSet                   = 0x0003
                }
            SET_TYPE;

            typedef struct
                {
                CACHE                           DeleteCache;
                CACHE                           NewCache;
                }
            CACHES;

            typedef struct
                {
                //
                //  Data locks.
                //
                LOCK                            ClassLock;

                //
                //  Protected data.
                //
                FILE_ADDRESS                    Caches;
                FILE_SIZE                       MaxSize;
                FILE_ADDRESS                    Set;
                FILE_SIZE                       Signature;
                SET_TYPE                        Type;
                FILE_SIZE                       Used;
                VALUE                           Value;
                }
            ROOT;

            //
            //  Private data.
            //
            BOOLEAN                             Active;
            FILE_ADDRESS                        Address;

            //
            //  Private static data.
            //
            STATIC MEMORY_FUNCTIONS<SNATIVE>    MemoryFunctions;

        public:
            //
            //  Public types.
            //
            typedef BOOLEAN (*SET_WALK_FUNCTION)
                (
                VALUE                           & Value,
                VOID                            *UserValue
                );

            //
            //  Public functions.
            //
            SET( VOID );
            SET( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN Copy( SET<HEAP,LOCK,VALUE> *Source );
            BOOLEAN Copy( SET<HEAP,LOCK,VALUE> *Destination,SET<HEAP,LOCK,VALUE> *Source );
            BOOLEAN Copy( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,SET<HEAP,LOCK,VALUE> *Source );

            BOOLEAN Copy
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                SET<HEAP,LOCK,VALUE>            *Destination,
                SET<HEAP,LOCK,VALUE>            *Source
                );

            BOOLEAN Copy
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope1,
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope2,
                SET<HEAP,LOCK,VALUE>            *Destination,
                SET<HEAP,LOCK,VALUE>            *Source
                );

            BOOLEAN DeleteAll( VOID );
            BOOLEAN DeleteAll( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN DeleteValue( CONST VALUE & Value );
            BOOLEAN DeleteValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST VALUE & Value );

            BOOLEAN Difference( SET<HEAP,LOCK,VALUE> *Source );
            BOOLEAN Difference( SET<HEAP,LOCK,VALUE> *Destination,SET<HEAP,LOCK,VALUE> *Source );
            BOOLEAN Difference( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,SET<HEAP,LOCK,VALUE> *Source );

            BOOLEAN Difference
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                SET<HEAP,LOCK,VALUE>            *Destination,
                SET<HEAP,LOCK,VALUE>            *Source
                );

            BOOLEAN Difference
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope1,
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope2,
                SET<HEAP,LOCK,VALUE>            *Destination,
                SET<HEAP,LOCK,VALUE>            *Source
                );

            BOOLEAN EndView( VOID );
            BOOLEAN EndView( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN FindValue( CONST VALUE & Value );
            BOOLEAN FindValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST VALUE & Value );

            BOOLEAN Intersect( SET<HEAP,LOCK,VALUE> *Source );
            BOOLEAN Intersect( SET<HEAP,LOCK,VALUE> *Destination,SET<HEAP,LOCK,VALUE> *Source );
            BOOLEAN Intersect( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,SET<HEAP,LOCK,VALUE> *Source );

            BOOLEAN Intersect
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                SET<HEAP,LOCK,VALUE>            *Destination,
                SET<HEAP,LOCK,VALUE>            *Source
                );

            BOOLEAN Intersect
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope1,
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope2,
                SET<HEAP,LOCK,VALUE>            *Destination,
                SET<HEAP,LOCK,VALUE>            *Source
                );

            BOOLEAN Join( SET<HEAP,LOCK,VALUE> *Source );
            BOOLEAN Join( SET<HEAP,LOCK,VALUE> *Destination,SET<HEAP,LOCK,VALUE> *Source );
            BOOLEAN Join( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,SET<HEAP,LOCK,VALUE> *Source );

            BOOLEAN Join
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                SET<HEAP,LOCK,VALUE>            *Destination,
                SET<HEAP,LOCK,VALUE>            *Source
                );

            BOOLEAN Join
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope1,
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope2,
                SET<HEAP,LOCK,VALUE>            *Destination,
                SET<HEAP,LOCK,VALUE>            *Source
                );

            BOOLEAN NewValue( CONST VALUE & Value );
            BOOLEAN NewValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST VALUE & Value );

            BOOLEAN Size( FILE_SIZE *Value );
            BOOLEAN Size( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_SIZE *Value );

            BOOLEAN Value( FILE_SIZE MaxSize,FILE_SIZE *Size,VALUE *Values );

            BOOLEAN Value
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                FILE_SIZE                       MaxSize,
                FILE_SIZE                       *Size,
                VALUE                           *Values
                );

            BOOLEAN View( VALUE **Set,FILE_SIZE *Size = NULL );
            BOOLEAN View( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,VALUE **Set,FILE_SIZE *Size = NULL );

            BOOLEAN Walk( SET_WALK_FUNCTION WalkFunction,VOID *UserValue = NULL );

            BOOLEAN Walk
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                SET_WALK_FUNCTION               WalkFunction,
                VOID                            *UserValue = NULL
                );

            ~SET( VOID );

        private:
            //
            //  Private functions.
            //
            BOOLEAN DeleteAll( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,ROOT *Root );

            BOOLEAN DeleteSetValue( VALUE *Set,FILE_SIZE *Used,VALUE *Value );

            VOID DifferenceSets
                (
                FILE_SIZE                       MaxSize,
                VALUE                           *Set1,
                VALUE                           *Set2,
                FILE_SIZE                       *Used1,
                FILE_SIZE                       Used2
                );

            BOOLEAN FindSetValue( FILE_ADDRESS *Offset,VALUE *Set,FILE_SIZE Used,VALUE *Value );

            VOID IntersectSets
                (
                FILE_SIZE                       MaxSize,
                VALUE                           *Set1,
                VALUE                           *Set2,
                FILE_SIZE                       *Used1,
                FILE_SIZE                       Used2
                );

            VOID JoinSets
                (
                FILE_SIZE                       MaxSize,
                VALUE                           *Set1,
                VALUE                           *Set2,
                FILE_SIZE                       *Used1,
                FILE_SIZE                       Used2
                );

            BOOLEAN MergeCaches( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CACHES *Caches,ROOT *Root );

            BOOLEAN NewSetValue( VALUE *Set,FILE_SIZE *Used,VALUE *Value );

            BOOLEAN NormalizeSets
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope1,
                AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope2,
                AUTOMATIC_VIEW_SCOPE<HEAP>          *AutomaticViewScope1,
                ROOT                                *Root1,
                ROOT                                **Root2,
                VALUE                               **Set1,
                VALUE                               **Set2,
                SET<HEAP,LOCK,VALUE>                *Source
                );

            BOOLEAN Resize( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,ROOT *Root,FILE_SIZE Size );

            VOID ZeroValues( VALUE *Values,FILE_SIZE Size );

            //
            //  Disabled functionality.
            //
            SET( CONST SET & NoCopies );

            VOID operator=( CONST SET & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Class Static Data.                                                                        */
            /*                                                                                            */
            /*  The class static data needs to be initialised before it's ready for use.                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> MEMORY_FUNCTIONS<SNATIVE> SET<HEAP,LOCK,VALUE>::MemoryFunctions;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'SET'.                                                                    */
            /*                                                                                            */
            /*  The constructor for 'SET' prepares the class for use.                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> SET<HEAP,LOCK,VALUE>::SET( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        //
        //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
        //  we fail to highlight the issue.
        //
        if ( AutomaticHeapScope.Feature( (LOCK::Feature()) ) )
            {
            //
            //  Setup class members.
            //
            Active = False;
            Address = NULL;

            //
            //  Where necessary we automatically wrap all the changes within a new transaction.
            //
            if ( AutomaticHeapScope.BeginTransaction() )
                {
                AUTO ROOT *Root;

                //
                //  We create a new 'Root' on the heap and store the 'Address'.
                //
                if ( ROCKALL_NEW<ROOT>::New( & AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                    {
                    //
                    //  We now setup the 'Root' of the new 'Set'.
                    //
                    Root -> Caches = NULL;
                    Root -> MaxSize = DefaultSize;
                    Root -> Set = NULL;
                    Root -> Signature = Signature;
                    Root -> Type = SimpleRowSet;
                    Root -> Used = 0;

                    //
                    //  We zero 'Value' just to be tidy.
                    //
                    ZeroValues( & Root -> Value,1 );

                    Active = True;
                    }
                }
            }
        else
            { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'SET'.                                                                    */
            /*                                                                                            */
            /*  The constructor for 'SET' prepares the class for use.                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> SET<HEAP,LOCK,VALUE>::SET
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        //
        //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
        //  we fail to highlight the issue.
        //
        if ( AutomaticHeapScope -> Feature( (LOCK::Feature()) ) )
            {
            AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

            //
            //  Setup class members.
            //
            Active = False;
            Address = NULL;

            //
            //  Where necessary we automatically wrap all the changes within a new transaction.
            //
            if ( AutomaticTransactionScope.BeginTransaction() )
                {
                AUTO ROOT *Root;

                //
                //  We create a new 'Root' on the heap and store the 'Address'.
                //
                if ( ROCKALL_NEW<ROOT>::New( AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                    {
                    //
                    //  We now setup the 'Root' of the new 'Set'.
                    //
                    Root -> Caches = NULL;
                    Root -> MaxSize = DefaultSize;
                    Root -> Set = NULL;
                    Root -> Signature = Signature;
                    Root -> Type = SimpleRowSet;
                    Root -> Used = 0;

                    //
                    //  We zero 'Value' just to be tidy.
                    //
                    ZeroValues( & Root -> Value,1 );

                    Active = True;
                    }
                }
            }
        else
            { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy.                                                                                     */
            /*                                                                                            */
            /*  A set is copied to another set.  If the 'HEAP' type differs between the two sets then     */
            /*  'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for    */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Copy
            (
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Copy( & AutomaticHeapScope,& AutomaticHeapScope,this,Source ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy.                                                                                     */
            /*                                                                                            */
            /*  A set is copied to another set.  If the 'HEAP' type differs between the two sets then     */
            /*  'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for    */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Copy
            (
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Copy( & AutomaticHeapScope,& AutomaticHeapScope,Destination,Source ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy.                                                                                     */
            /*                                                                                            */
            /*  A set is copied to another set.  If the 'HEAP' type differs between the two sets then     */
            /*  'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for    */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Copy
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        { return (Copy( AutomaticHeapScope,AutomaticHeapScope,this,Source )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy.                                                                                     */
            /*                                                                                            */
            /*  A set is copied to another set.  If the 'HEAP' type differs between the two sets then     */
            /*  'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for    */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Copy
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        { return (Copy( AutomaticHeapScope,AutomaticHeapScope,Destination,Source )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy.                                                                                     */
            /*                                                                                            */
            /*  A set is copied to another set.  If the 'HEAP' type differs between the two sets then     */
            /*  'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for    */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Copy
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope1,
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope2,
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO BOOLEAN SingleTransaction = ((AutomaticHeapScope1 -> GetHeap()) == (AutomaticHeapScope2 -> GetHeap()));
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope1( AutomaticHeapScope1 );
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope2( AutomaticHeapScope2 );
        AUTO TRANSACTIONAL_LOCK TransactionLock1( AutomaticHeapScope1 );
        AUTO TRANSACTIONAL_LOCK TransactionLock2( AutomaticHeapScope2 );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' 'ExclsiveView()',
        //  'Update()' and 'View()' should be automatically optimized away for heaps that do
        //  not support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if
                (
                ((Destination -> Active) && (Source -> Active))
                    &&
                (AutomaticTransactionScope1.BeginTransaction( NULL,& TransactionLock1 ))
                    &&
                ((SingleTransaction) || (AutomaticTransactionScope2.BeginTransaction( NULL,& TransactionLock2 )))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock1 );
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock2( & TransactionLock2 );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope2 );
            AUTO ROOT *Root1;
            AUTO ROOT *Root2;

            //
            //  We 'Update()' the 'Destination' and 'View()' the 'Source'.  Finally, we check the
            //  'Signature' ensure each 'Root' is of the correct type. We claimed a 'TransactionLock'
            //  above to prevent threads in the same transactions from accessing the sets.
            //
            if
                    (
                    (AutomaticHeapScope1 -> Update( Destination -> Address,((VOID**) & Root1) ))
                        &&
                    (AutomaticViewScope1.ExclusiveView( Source -> Address,((VOID**) & Root2) ))
                        &&
                    (Root1 -> Signature == Signature)
                        &&
                    (Root2 -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock3( & Root1 -> ClassLock );
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock4( & Root2 -> ClassLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope2 );

                //
                //  A 'Copy()' is similar to a 'Join()'.  The key difference is that we 'DeleteAll()'
                //  on 'Set1' and then 'Join()' to 'Set2'.
                //
                if ( DeleteAll( AutomaticHeapScope1,Root1 ) )
                    {
                    //
                    //  We must have enough space before we are willing to do anything.  So we
                    //  'Resize()' 'Set1' to make sure we have enough space.
                    //
                    if ( Resize( AutomaticHeapScope1,Root1,(Root1 -> Used + Root2 -> Used) ) )
                        {
                        AUTO VALUE *Set1;
                        AUTO VALUE *Set2;

                        //
                        //  We need to put the sets into a standard normalized format before we
                        //  can work on them.
                        //
                        if
                                (
                                NormalizeSets
                                    (
                                    AutomaticHeapScope1,
                                    AutomaticHeapScope2,
                                    & AutomaticViewScope2,
                                    Root1,
                                    & Root2,
                                    & Set1,
                                    & Set2,
                                    Source
                                    )
                                )
                            {
                            //
                            //  Compute the set 'Join'.
                            //
                            JoinSets( Root1 -> MaxSize,Set1,Set2,& Root1 -> Used,Root2 -> Used );

                            return True;
                            }
                        }
                    }
                }
            else
                { return (AutomaticHeapScope1 -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope1 -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the values in a set.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::DeleteAll( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (DeleteAll( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the values in a set.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::DeleteAll
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the set.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  Delete all the values.
                //
                return (DeleteAll( AutomaticHeapScope,Root ));
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete an entire set.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::DeleteAll
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            ROOT                                *Root
            )
        {
        //
        //  The way we delete a set depends on its 'Type'.
        //
        switch ( Root -> Type )
            {
            case SimpleRowSet:
                {
                //
                //  It is not very hard to delete a simple row set.
                //
                Root -> Caches = NULL;
                Root -> MaxSize = DefaultSize;
                Root -> Set = NULL;
                Root -> Type = SimpleRowSet;
                Root -> Used = 0;

                //
                //  We zero 'Value' just to be tidy.
                //
                ZeroValues( & Root -> Value,1 );

                return True;
                }

            case ExtendedRowSet:
                {
                AUTO VALUE *Set;

                //
                //  We will need to 'Update()' the associated 'Set'.
                //
                if ( AutomaticHeapScope -> Update( Root -> Set,((VOID**) & Set) ) )
                    {
                    AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);
                    AUTO FILE_SIZE Count;

                    //
                    //  Call the extra destructors.
                    //
                    for ( Count=(Root -> MaxSize-1);Count >= MinExtendedSize;Count -- )
                        { PLACEMENT_DELETE( & Set[ Count ],VALUE ); }

                    //
                    //  We will need to 'Delete()' the associated 'Set'.
                    //
                    if ( ROCKALL_DELETE<VALUE>::Delete( AutomaticHeapScope,Root -> Set,MinExtendedSize ) )
                        {
                        //
                        //  We now zero the 'Root' of the 'Set'.
                        //
                        Root -> Caches = NULL;
                        Root -> MaxSize = DefaultSize;
                        Root -> Set = NULL;
                        Root -> Type = SimpleRowSet;
                        Root -> Used = 0;

                        //
                        //  We zero 'Value' just to be tidy.
                        //
                        ZeroValues( & Root -> Value,1 );

                        return True;
                        }
                    }

                break;
                }

            case ComplexRowSet:
                {
                AUTO CACHES *Caches;

                //
                //  We will need to 'Update()' the associated 'Set'.
                //
                if ( AutomaticHeapScope -> Update( Root -> Caches,((VOID**) & Caches) ) )
                    {
                    //
                    //  We will need to 'Delete()' the associated 'Caches'.
                    //
                    if ( ROCKALL_DELETE<CACHES>::Delete( AutomaticHeapScope,Root -> Caches ) )
                        {
                        AUTO VALUE *Set;

                        //
                        //  Regardless, the 'Caches' have now gone.
                        //
                        Root -> Caches = NULL;
                        Root -> Type = ExtendedRowSet;

                        //
                        //  We will need to 'Update()' the associated 'Set'.
                        //
                        if ( AutomaticHeapScope -> Update( Root -> Set,((VOID**) & Set) ) )
                            {
                            AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);
                            AUTO FILE_SIZE Count;

                            //
                            //  Call the extra destructors.
                            //
                            for ( Count=(Root -> MaxSize-1);Count >= MinExtendedSize;Count -- )
                                { PLACEMENT_DELETE( & Set[ Count ],VALUE ); }

                            //
                            //  We will need to 'Delete()' the associated 'Set'.
                            //
                            if ( ROCKALL_DELETE<VALUE>::Delete( AutomaticHeapScope,Root -> Set,MinExtendedSize ) )
                                {
                                //
                                //  We now zero the 'Root' of the 'Set'.
                                //
                                Root -> Caches = NULL;
                                Root -> MaxSize = DefaultSize;
                                Root -> Set = NULL;
                                Root -> Type = SimpleRowSet;
                                Root -> Used = 0;

                                //
                                //  We zero 'Value' just to be tidy.
                                //
                                ZeroValues( & Root -> Value,1 );

                                return True;
                                }
                            }
                        }
                    }

                break;
                }

            default:
                { FAILURE( "Unknown type" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Set Value.                                                                         */
            /*                                                                                            */
            /*  Delete a value from a set.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::DeleteSetValue
            (
            VALUE                               *Set,
            FILE_SIZE                           *Used,
            VALUE                               *Value
            )
        {
        AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);

        //
        //  We way we delete the 'Value' depends on the size of the 'Set'.
        //
        if ( (*Used) == 1 )
            {
            //
            //  When the 'Set' has just one value a delete is trivial.
            //
            if ( (*Value) == Set[0] )
                {
                //
                //  Delete the value.
                //
                ZeroValues( & Set[0],1 );

                (*Used) = 0;

                return True;
                }
            }
        else
            {
            AUTO FILE_SIZE Offset;

            //
            //  When the 'Set' is larger we need to find the value to delete.
            //
            if ( FindSetValue( & Offset,Set,(*Used),Value ) )
                {
                AUTO FILE_SIZE Count;

                //
                //  We now move any greater values up the 'Set'.
                //
                for ( Count=Offset,(*Used) --;Count < (*Used);Count ++ )
                    { Set[ Count ] = Set[ (Count+1) ]; }

                //
                //  Zero the value just to be tidy.
                //
                ZeroValues( & Set[ (*Used) ],1 );

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Value.                                                                             */
            /*                                                                                            */
            /*  Delete a value from a set.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::DeleteValue
            (
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (DeleteValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Value.                                                                             */
            /*                                                                                            */
            /*  Delete a value from a set.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::DeleteValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the set.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  The way we delete a 'Value' depends on the 'Type' of the set.
                //
                switch ( Root -> Type )
                    {
                    case SimpleRowSet:
                        {
                        //
                        //  It is not very hard to delete the value in a 'SimpleRowSet'.
                        //
                        if ( (Root -> Used == 1) && (Root -> Value == Value) )
                            {
                            Root -> Used --;

                            ZeroValues( & Root -> Value,1 );

                            return True;
                            }

                        break;
                        }

                    case ExtendedRowSet:
                        {
                        AUTO VALUE *Set;

                        //
                        //  We will need to 'Update()' the associated 'Set'.
                        //
                        if ( AutomaticHeapScope -> Update( Root -> Set,((VOID**) & Set) ) )
                            {
                            if ( DeleteSetValue( Set,& Root -> Used,((VALUE*) & Value) ) )
                                { return True; }
                            else
                                { return (AutomaticHeapScope -> SetError( ErrorValueNotFound )); }
                            }

                        break;
                        }

                    case ComplexRowSet:
                        {
                        AUTO CACHES *Caches;

                        //
                        //  We will need to 'Update()' the associated 'Caches'.
                        //
                        if ( AutomaticHeapScope -> Update( Root -> Caches,((VOID**) & Caches) ) )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                            AUTO CACHE *DeleteCache = & Caches -> DeleteCache;
                            AUTO CACHE *NewCache = & Caches -> NewCache;
                            AUTO FILE_ADDRESS Offset;
                            AUTO VALUE *Set;

                            //
                            //  We will need to 'MergeCaches()' if there is not enough space
                            //  to delete a value.
                            //
                            if ( Caches -> DeleteCache.Full() )
                                { MergeCaches( AutomaticHeapScope,Caches,Root ); }

                            //
                            //  We will need to 'View()' the associated 'Set'.
                            //
                            if ( AutomaticViewScope.View( Root -> Set,((VOID**) & Set) ) )
                                {
                                AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);

                                //
                                //  We search the 'Set' for the 'Value'.  If it is present
                                //  then it must not be in the 'DeleteCache' or it is an error.
                                //  If it is not present in the 'Set' it must be in the
                                //  'NewCache' or it is an error.
                                //
                                if ( FindSetValue( & Offset,Set,Root -> Used,((VALUE*) & Value) ) )
                                    {
                                    if ( DeleteCache -> NewValue( Value ) )
                                        { return True; }
                                    else
                                        { return (AutomaticHeapScope -> SetError( ErrorValueNotFound )); }
                                    }
                                else
                                    {
                                    if ( NewCache -> DeleteValue( Value ) )
                                        { return True; }
                                    else
                                        { return (AutomaticHeapScope -> SetError( ErrorValueNotFound )); }
                                    }
                                }
                            }

                        break;
                        }

                    default:
                        { FAILURE( "Unknown type" ); }
                    }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Difference.                                                                               */
            /*                                                                                            */
            /*  We compute the difference between two sets.  If the 'HEAP' type differs between the       */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Difference
            (
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Difference( & AutomaticHeapScope,& AutomaticHeapScope,this,Source ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Difference.                                                                               */
            /*                                                                                            */
            /*  We compute the difference between two sets.  If the 'HEAP' type differs between the       */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Difference
            (
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Difference( & AutomaticHeapScope,& AutomaticHeapScope,Destination,Source ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Difference.                                                                               */
            /*                                                                                            */
            /*  We compute the difference between two sets.  If the 'HEAP' type differs between the       */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Difference
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        { return (Difference( AutomaticHeapScope,AutomaticHeapScope,this,Source )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Difference.                                                                               */
            /*                                                                                            */
            /*  We compute the difference between two sets.  If the 'HEAP' type differs between the       */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Difference
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        { return (Difference( AutomaticHeapScope,AutomaticHeapScope,Destination,Source )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Difference.                                                                               */
            /*                                                                                            */
            /*  We compute the difference between two sets.  If the 'HEAP' type differs between the       */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Difference
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope1,
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope2,
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO BOOLEAN SingleTransaction = ((AutomaticHeapScope1 -> GetHeap()) == (AutomaticHeapScope2 -> GetHeap()));
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope1( AutomaticHeapScope1 );
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope2( AutomaticHeapScope2 );
        AUTO TRANSACTIONAL_LOCK TransactionLock1( AutomaticHeapScope1 );
        AUTO TRANSACTIONAL_LOCK TransactionLock2( AutomaticHeapScope2 );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' 'ExclsiveView()',
        //  'Update()' and 'View()' should be automatically optimized away for heaps that do
        //  not support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if
                (
                ((Destination -> Active) && (Source -> Active))
                    &&
                (AutomaticTransactionScope1.BeginTransaction( NULL,& TransactionLock1 ))
                    &&
                ((SingleTransaction) || (AutomaticTransactionScope2.BeginTransaction( NULL,& TransactionLock2 )))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock1 );
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock2( & TransactionLock2 );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope2 );
            AUTO ROOT *Root1;
            AUTO ROOT *Root2;

            //
            //  We 'Update()' the 'Destination' and 'View()' the 'Source'.  Finally, we check the
            //  'Signature' ensure each 'Root' is of the correct type. We claimed a 'TransactionLock'
            //  above to prevent threads in the same transactions from accessing the sets.
            //
            if
                    (
                    (AutomaticHeapScope1 -> Update( Destination -> Address,((VOID**) & Root1) ))
                        &&
                    (AutomaticViewScope1.ExclusiveView( Source -> Address,((VOID**) & Root2) ))
                        &&
                    (Root1 -> Signature == Signature)
                        &&
                    (Root2 -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock3( & Root1 -> ClassLock );
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock4( & Root2 -> ClassLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope2 );

                //
                //  We must have enough space before we are willing to do anything.  So we
                //  'Resize()' 'Set1' to make sure we have enough space.
                //
                if ( Resize( AutomaticHeapScope1,Root1,(Root1 -> Used + Root2 -> Used) ) )
                    {
                    AUTO VALUE *Set1;
                    AUTO VALUE *Set2;

                    //
                    //  We need to put the sets into a standard normalized format before we
                    //  can work on them.
                    //
                    if
                            (
                            NormalizeSets
                                (
                                AutomaticHeapScope1,
                                AutomaticHeapScope2,
                                & AutomaticViewScope2,
                                Root1,
                                & Root2,
                                & Set1,
                                & Set2,
                                Source
                                )
                            )
                        {
                        //
                        //  Compute the set 'Difference'.
                        //
                        DifferenceSets( Root1 -> MaxSize,Set1,Set2,& Root1 -> Used,Root2 -> Used );

                        return True;
                        }
                    }
                }
            else
                { return (AutomaticHeapScope1 -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope1 -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Difference Sets.                                                                          */
            /*                                                                                            */
            /*  A difference in an implementation of the difference set operator.                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> VOID SET<HEAP,LOCK,VALUE>::DifferenceSets
            (
            FILE_SIZE                           MaxSize,
            VALUE                               *Set1,
            VALUE                               *Set2,
            FILE_SIZE                           *Used1,
            FILE_SIZE                           Used2
            )
        {
        AUTO VALUE (*View1)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set1);
        AUTO VALUE (*View2)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set2);
        AUTO FILE_SIZE Count1 = ((*Used1)-1);
        AUTO FILE_SIZE Count2 = (Used2-1);
        AUTO FILE_SIZE Count3 = (MaxSize-1);

        //
        //  When we difference the two arrays we put the results at the end of 'Set1'.  We
        //  do this  to avoid allocating any new memory.  We do the diffrence in reverse
        //  order so that we do not accidentially overwrite unmerged values.
        //
        while ( (Count1 >= 0) && (Count2 >= 0) )
            {
            //
            //  When 'Set1' is greater than 'Set2' we copy it to the end of 'Set1'.
            //
            if ( Set1[ Count1 ] > Set2[ Count2 ] )
                { Set1[ (Count3 --) ] = Set1[ (Count1 --) ]; }
            else
                {
                //
                //  When 'Set2' is greater than 'Set1' we copy it to the end of
                //  'Set1'.
                //
                if ( Set1[ Count1 ] < Set2[ Count2 ] )
                    { Set1[ (Count3 --) ] = Set2[ (Count2 --) ]; }
                else
                    { Count1 --; Count2 --; }
                }
            }

        //
        //  Now if we run out of values in 'Set1' we copy 'Set2' to the start of 'Set1' so the
        //  values are in their final resting place.  If we run out of values in 'Set2' we just
        //  increment 'Count1' which the leaves the values in 'Set1' in the right place for the
        //  next step.
        //
        if ( (Count1 < 0) && (Count2 >= 0) )
            {
            for ( Count1=0;Count1 <= Count2;Count1 ++ )
                { Set1[ Count1 ] = Set2[ Count1 ]; }
            }
        else
            { Count1 ++; }

        //
        //  Now the values at the start of 'Set1' are already in the correct
        //  place.  So just copy the merged values back into place from the
        //  end of 'Set1'.
        //
        for ( Count3 ++;Count3 < MaxSize;Count1 ++,Count3 ++ )
            { Set1[ Count1 ] = Set1[ Count3 ]; }

        //
        //  Finally, we zero any junk we may have left at the end of 'Set1'.
        //
        ZeroValues( & Set1[ Count1 ],(MaxSize - Count1) );

        //
        //  Update the 'Used1' value.
        //
        (*Used1) = Count1;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End View.                                                                                 */
            /*                                                                                            */
            /*  End a set view.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::EndView( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (EndView( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End View.                                                                                 */
            /*                                                                                            */
            /*  End a set view.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::EndView
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO ROOT *Root;

        //
        //  We ensure that the class is 'Active'.  Next we 'View()' the 'Root' as we need to
        //  examine it.  Finally, we check the 'Signature' to ensure the 'Root' is of the correct
        //  type.  The function calls to 'View()' should be automatically optimized away for heaps
        //  that do not support this functionality by any decent compiler, due to the way the
        //  code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticHeapScope -> ExclusiveView( Address,((VOID**) & Root) ))
                    &&
                (Root -> Signature == Signature)
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Root -> ClassLock );

            //
            //  The way we view the data depends on the 'Type' of the set.
            //
            switch ( Root -> Type )
                {
                case SimpleRowSet:
                    {
                    AUTO FILE_ADDRESS EndAddress = Address;

                    //
                    //  Release the lock on the 'Set'.
                    //
                    AutomaticExclusiveLock.Release();

                    //
                    //  End the all the outstanding views and the transaction.
                    //
                    return
                        (
                        (AutomaticHeapScope -> EndExclusiveView( EndAddress ))
                            &&
                        (AutomaticHeapScope -> EndExclusiveView( EndAddress ))
                            &&
                        (AutomaticHeapScope -> EndTransaction())
                        );
                    }

                case ExtendedRowSet:
                case ComplexRowSet:
                    {
                    AUTO FILE_ADDRESS EndAddress = Address;
                    AUTO FILE_ADDRESS EndSet = Root -> Set;

                    //
                    //  Release the lock on the 'Set'.
                    //
                    AutomaticExclusiveLock.Release();

                    //
                    //  End the all the outstanding views and the transaction.
                    //
                    return
                        (
                        (AutomaticHeapScope -> EndView( EndSet ))
                            &&
                        (AutomaticHeapScope -> EndExclusiveView( EndAddress ))
                            &&
                        (AutomaticHeapScope -> EndExclusiveView( EndAddress ))
                            &&
                        (AutomaticHeapScope -> EndTransaction())
                        );
                    }

                default:
                    { FAILURE( "Unknown type" ); }
                }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Set Value.                                                                           */
            /*                                                                                            */
            /*  Find a value in a set.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::FindSetValue
            (
            FILE_ADDRESS                        *Offset,
            VALUE                               *Set,
            FILE_SIZE                           Used,
            VALUE                               *Value
            )
        {
        AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);

        //
        //  We way we find a 'Value' depends on the size of the 'Set'.
        //
        switch ( Used )
            {
            case 0:
                {
                //
                //  When the 'Set' is empty we can immediately exit.
                //
                (*Offset) = 0;

                break;
                }

            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 7:
                {
                AUTO SBIT32 Count;

                //
                //  When the 'Set' is small we use a simple loop.
                //
                for ( Count=0;Count < Used;Count ++ )
                    {
                    AUTO VALUE *Current = & Set[ Count ];

                    //
                    //  When 'Current' is greater or equal to the target 'Value' we exit.
                    //
                    if ( (*Current) >= (*Value) )
                        {
                        (*Offset) = Count;

                        return ((*Current) == (*Value));
                        }
                    }

                //
                //  When the 'Value' is the largest value the 'Offset' points to the end
                //  of the 'Set'
                //
                (*Offset) = Used;

                break;
                }

            default:
                {
                AUTO FILE_SIZE Maximum = (Used-1);
                AUTO FILE_SIZE Minimum = 0;

                //
                //  We find the 'Value' using a binary search as the 'Set' is stored in
                //  ascending sorted order.
                //
                while ( Maximum >= Minimum )
                    {
                    AUTO FILE_ADDRESS MidPoint = ((Maximum + Minimum) / 2);
                    AUTO VALUE *MidValue = & Set[ MidPoint ];

                    //
                    //  Let's see if the 'MidValue' is greater than the 'Value'.
                    //
                    if ( (*MidValue) < (*Value) )
                        { Minimum = (MidPoint + 1); }
                    else
                        {
                        //
                        //  Let's see if the 'MidValue' value is less than the 'Value'.
                        //
                        if ( (*MidValue) > (*Value) )
                            { Maximum = (MidPoint - 1); }
                        else
                            {
                            //
                            //  We have found the 'Value'.
                            //
                            (*Offset) = MidPoint;

                            return True;
                            }
                        }
                    }

                //
                //  Update the 'Offset' to point to the nearest greater value.
                //
                (*Offset) = Minimum;

                break;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Value.                                                                               */
            /*                                                                                            */
            /*  Find a value in a set.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::FindValue
            (
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (FindValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Value.                                                                               */
            /*                                                                                            */
            /*  Find a value in a set.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::FindValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
        AUTO ROOT *Root;

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  Next we 'View()' the 'Root' as we need to examine it.
        //  Finally, we check the 'Signature' to ensure the 'Root' is of the correct type.
        //  The function calls to 'BeginTransaction()' and 'View()' should be automatically
        //  optimized away for heaps that do not support this functionality by any decent
        //  compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction())
                    &&
                (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                    &&
                (Root -> Signature == Signature)
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Root -> ClassLock );

            //
            //  The way we find a 'Value' depends on the 'Type' of the set.
            //
            switch ( Root -> Type )
                {
                case SimpleRowSet:
                    {
                    //
                    //  It is not very hard to find the value in a 'SimpleRowSet'.
                    //
                    return ((Root -> Used == 1) && (Root -> Value == Value));
                    }

                case ExtendedRowSet:
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                    AUTO FILE_ADDRESS Offset;
                    AUTO VALUE *Set;

                    //
                    //  We will need to 'View()' the associated 'Set'.
                    //
                    if ( AutomaticViewScope2.View( Root -> Set,((VOID**) & Set) ) )
                        {
                        if ( FindSetValue( & Offset,Set,Root -> Used,((VALUE*) & Value) ) )
                            { return True; }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorValueNotFound )); }
                        }

                    break;
                    }

                case ComplexRowSet:
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                    AUTO VALUE *Set;

                    //
                    //  We will need to 'View()' the associated 'Set'.
                    //
                    if ( AutomaticViewScope2.View( Root -> Set,((VOID**) & Set) ) )
                        {
                        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope );
                        AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);
                        AUTO CACHES *Caches;

                        //
                        //  We will need to 'View()' the associated 'Caches'.
                        //
                        if ( AutomaticViewScope3.View( Root -> Caches,((VOID**) & Caches) ) )
                            {
                            AUTO FILE_ADDRESS Offset;

                            //
                            //  Find the 'Value' in the 'Set'.
                            //
                            if ( FindSetValue( & Offset,Set,Root -> Used,((VALUE*) & Value) ) )
                                {
                                if ( ! Caches -> DeleteCache.FindValue( Value ) )
                                    { return True; }
                                else
                                    { return (AutomaticHeapScope -> SetError( ErrorValueNotFound )); }
                                }
                            else
                                {
                                if ( Caches -> NewCache.FindValue( Value ) )
                                    { return True; }
                                else
                                    { return (AutomaticHeapScope -> SetError( ErrorValueNotFound )); }
                                }
                            }
                        }

                    break;
                    }

                default:
                    { FAILURE( "Unknown type" ); }
                }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Intersect.                                                                                */
            /*                                                                                            */
            /*  We compute the intersect between two sets.  If the 'HEAP' type differs between the        */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Intersect
            (
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Intersect( & AutomaticHeapScope,& AutomaticHeapScope,this,Source ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Intersect.                                                                                */
            /*                                                                                            */
            /*  We compute the intersect between two sets.  If the 'HEAP' type differs between the        */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Intersect
            (
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Intersect( & AutomaticHeapScope,& AutomaticHeapScope,Destination,Source ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Intersect.                                                                                */
            /*                                                                                            */
            /*  We compute the intersect between two sets.  If the 'HEAP' type differs between the        */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Intersect
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        { return (Intersect( AutomaticHeapScope,AutomaticHeapScope,this,Source )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Intersect.                                                                                */
            /*                                                                                            */
            /*  We compute the intersect between two sets.  If the 'HEAP' type differs between the        */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Intersect
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        { return (Intersect( AutomaticHeapScope,AutomaticHeapScope,Destination,Source )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Intersect.                                                                                */
            /*                                                                                            */
            /*  We compute the intersect between two sets.  If the 'HEAP' type differs between the        */
            /*  two sets then 'HEAP' type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual      */
            /*  'HEAP' type for better performance.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Intersect
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope1,
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope2,
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO BOOLEAN SingleTransaction = ((AutomaticHeapScope1 -> GetHeap()) == (AutomaticHeapScope2 -> GetHeap()));
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope1( AutomaticHeapScope1 );
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope2( AutomaticHeapScope2 );
        AUTO TRANSACTIONAL_LOCK TransactionLock1( AutomaticHeapScope1 );
        AUTO TRANSACTIONAL_LOCK TransactionLock2( AutomaticHeapScope2 );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' 'ExclsiveView()',
        //  'Update()' and 'View()' should be automatically optimized away for heaps that do
        //  not support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if
                (
                ((Destination -> Active) && (Source -> Active))
                    &&
                (AutomaticTransactionScope1.BeginTransaction( NULL,& TransactionLock1 ))
                    &&
                ((SingleTransaction) || (AutomaticTransactionScope2.BeginTransaction( NULL,& TransactionLock2 )))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock1 );
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock2( & TransactionLock2 );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope2 );
            AUTO ROOT *Root1;
            AUTO ROOT *Root2;

            //
            //  We 'Update()' the 'Destination' and 'View()' the 'Source'.  Finally, we check the
            //  'Signature' ensure each 'Root' is of the correct type. We claimed a 'TransactionLock'
            //  above to prevent threads in the same transactions from accessing the sets.
            //
            if
                    (
                    (AutomaticHeapScope1 -> Update( Destination -> Address,((VOID**) & Root1) ))
                        &&
                    (AutomaticViewScope1.ExclusiveView( Source -> Address,((VOID**) & Root2) ))
                        &&
                    (Root1 -> Signature == Signature)
                        &&
                    (Root2 -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock3( & Root1 -> ClassLock );
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock4( & Root2 -> ClassLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope2 );
                AUTO VALUE *Set1;
                AUTO VALUE *Set2;

                //
                //  We need to put the sets into a standard normalized format before we
                //  can work on them.
                //
                if
                        (
                        NormalizeSets
                            (
                            AutomaticHeapScope1,
                            AutomaticHeapScope2,
                            & AutomaticViewScope2,
                            Root1,
                            & Root2,
                            & Set1,
                            & Set2,
                            Source
                            )
                        )
                    {
                    //
                    //  Compute the set 'Intersect'.
                    //
                    IntersectSets( Root1 -> MaxSize,Set1,Set2,& Root1 -> Used,Root2 -> Used );

                    return True;
                    }
                }
            else
                { return (AutomaticHeapScope1 -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope1 -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Intersect Sets.                                                                           */
            /*                                                                                            */
            /*  An intersect in an implementation of the intersect set operator.                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> VOID SET<HEAP,LOCK,VALUE>::IntersectSets
            (
            FILE_SIZE                           MaxSize,
            VALUE                               *Set1,
            VALUE                               *Set2,
            FILE_SIZE                           *Used1,
            FILE_SIZE                           Used2
            )
        {
        AUTO VALUE (*View1)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set1);
        AUTO VALUE (*View2)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set2);
        AUTO FILE_SIZE Count1 = ((*Used1)-1);
        AUTO FILE_SIZE Count2 = (Used2-1);
        AUTO FILE_SIZE Count3 = (MaxSize-1);

        //
        //  When we intersect the two arrays we put the results at the end of 'Set1'.  We
        //  do this to avoid allocating any new memory.  We do the intersect in reverse
        //  order so that we do not accidentially overwrite unmerged values.
        //
        while ( (Count1 >= 0) && (Count2 >= 0) )
            {
            //
            //  When 'Set1' is greater than 'Set2' skip the next value in 'Set1'.
            //
            if ( Set1[ Count1 ] > Set2[ Count2 ] )
                { Count1 --; }
            else
                {
                //
                //  When 'Set2' is greater than 'Set1' skip the next value in 'Set2'.
                //  If the values are equal we put a copy at the end of 'Set1'.
                //
                if ( Set1[ Count1 ] < Set2[ Count2 ] )
                    { Count2 --; }
                else
                    { Set1[ (Count3 --) ] = Set2[ (Count1 --, Count2 --) ]; }
                }
            }

        //
        //  Now we copy the intersected values back into place.
        //
        for ( Count1=0,Count3 ++;Count3 < MaxSize;Count1 ++,Count3 ++ )
            { Set1[ Count1 ] = Set1[ Count3 ]; }

        //
        //  Finally, we zero any junk we may have left at the end of 'Set1'.
        //
        ZeroValues( & Set1[ Count1 ],(MaxSize - Count1) );

        //
        //  Update the 'Used1' value.
        //
        (*Used1) = Count1;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  We join two sets.  If the 'HEAP' type differs between the two sets then 'HEAP'            */
            /*  type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for           */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Join
            (
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Join( & AutomaticHeapScope,& AutomaticHeapScope,this,Source ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  We join two sets.  If the 'HEAP' type differs between the two sets then 'HEAP'            */
            /*  type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for           */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Join
            (
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Join( & AutomaticHeapScope,& AutomaticHeapScope,Destination,Source ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  We join two sets.  If the 'HEAP' type differs between the two sets then 'HEAP'            */
            /*  type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for           */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Join
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        { return (Join( AutomaticHeapScope,AutomaticHeapScope,this,Source )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  We join two sets.  If the 'HEAP' type differs between the two sets then 'HEAP'            */
            /*  type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for           */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Join
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        { return (Join( AutomaticHeapScope,AutomaticHeapScope,Destination,Source )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  We join two sets.  If the 'HEAP' type differs between the two sets then 'HEAP'            */
            /*  type needs to be 'VIRTUAL_HEAP'.  If not, we can use the actual 'HEAP' type for           */
            /*  better performance.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Join
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope1,
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope2,
            SET<HEAP,LOCK,VALUE>                *Destination,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        AUTO BOOLEAN SingleTransaction = ((AutomaticHeapScope1 -> GetHeap()) == (AutomaticHeapScope2 -> GetHeap()));
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope1( AutomaticHeapScope1 );
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope2( AutomaticHeapScope2 );
        AUTO TRANSACTIONAL_LOCK TransactionLock1( AutomaticHeapScope1 );
        AUTO TRANSACTIONAL_LOCK TransactionLock2( AutomaticHeapScope2 );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' 'ExclsiveView()',
        //  'Update()' and 'View()' should be automatically optimized away for heaps that do
        //  not support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if
                (
                ((Destination -> Active) && (Source -> Active))
                    &&
                (AutomaticTransactionScope1.BeginTransaction( NULL,& TransactionLock1 ))
                    &&
                ((SingleTransaction) || (AutomaticTransactionScope2.BeginTransaction( NULL,& TransactionLock2 )))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock1 );
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock2( & TransactionLock2 );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope2 );
            AUTO ROOT *Root1;
            AUTO ROOT *Root2;

            //
            //  We 'Update()' the 'Destination' and 'View()' the 'Source'.  Finally, we check the
            //  'Signature' ensure each 'Root' is of the correct type. We claimed a 'TransactionLock'
            //  above to prevent threads in the same transactions from accessing the sets.
            //
            if
                    (
                    (AutomaticHeapScope1 -> Update( Destination -> Address,((VOID**) & Root1) ))
                        &&
                    (AutomaticViewScope1.ExclusiveView( Source -> Address,((VOID**) & Root2) ))
                        &&
                    (Root1 -> Signature == Signature)
                        &&
                    (Root2 -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock3( & Root1 -> ClassLock );
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock4( & Root2 -> ClassLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope2 );

                //
                //  We must have enough space before we are willing to do anything.  So we
                //  'Resize()' 'Set1' to make sure we have enough space.
                //
                if ( Resize( AutomaticHeapScope1,Root1,(Root1 -> Used + Root2 -> Used) ) )
                    {
                    AUTO VALUE *Set1;
                    AUTO VALUE *Set2;

                    //
                    //  We need to put the sets into a standard normalized format before we
                    //  can work on them.
                    //
                    if
                            (
                            NormalizeSets
                                (
                                AutomaticHeapScope1,
                                AutomaticHeapScope2,
                                & AutomaticViewScope2,
                                Root1,
                                & Root2,
                                & Set1,
                                & Set2,
                                Source
                                )
                            )
                        {
                        //
                        //  Compute the set 'Difference'.
                        //
                        JoinSets( Root1 -> MaxSize,Set1,Set2,& Root1 -> Used,Root2 -> Used );

                        return True;
                        }
                    }
                }
            else
                { return (AutomaticHeapScope1 -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope1 -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join Sets.                                                                                */
            /*                                                                                            */
            /*  A join in an implementation of the set join operator.                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> VOID SET<HEAP,LOCK,VALUE>::JoinSets
            (
            FILE_SIZE                           MaxSize,
            VALUE                               *Set1,
            VALUE                               *Set2,
            FILE_SIZE                           *Used1,
            FILE_SIZE                           Used2
            )
        {
        AUTO VALUE (*View1)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set1);
        AUTO VALUE (*View2)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set2);
        AUTO FILE_SIZE Count1 = ((*Used1)-1);
        AUTO FILE_SIZE Count2 = (Used2-1);
        AUTO FILE_SIZE Count3 = (MaxSize-1);

        //
        //  When we join the two arrays we put the results at the end of 'Set1'.  We
        //  do this to avoid allocating any new memory.  We do the join in reverse
        //  order so that we do not accidentially overwrite unmerged values.
        //
        while ( (Count1 >= 0) && (Count2 >= 0) )
            {
            //
            //  When 'Set1' is greater than 'Set2' we copy it to the end of 'Set1'.
            //
            if ( Set1[ Count1 ] > Set2[ Count2 ] )
                { Set1[ (Count3 --) ] = Set1[ (Count1 --) ]; }
            else
                {
                //
                //  When 'Set2' is greater than or equal to 'Set1' we copy it to the
                //  end of 'Set1'.
                //
                if ( Set1[ Count1 ] < Set2[ Count2 ] )
                    { Set1[ (Count3 --) ] = Set2[ (Count2 --) ]; }
                else
                    { Set1[ (Count3 --) ] = Set2[ (Count1 --, Count2 --) ]; }
                }
            }

        //
        //  Now if we run out of values in 'Set1' we copy 'Set2' to the start of 'Set1' so the
        //  values are in their final resting place.  If we run out of values in 'Set2' we just
        //  increment 'Count1' which the leaves the values in 'Set1' in the right place for the
        //  next step.
        //
        if ( (Count1 < 0) && (Count2 >= 0) )
            {
            for ( Count1=0;Count1 <= Count2;Count1 ++ )
                { Set1[ Count1 ] = Set2[ Count1 ]; }
            }
        else
            { Count1 ++; }

        //
        //  Now the values at the start of 'Set1' are already in the correct
        //  place.  So just copy the merged values back into place from the
        //  end of 'Set1'.
        //
        for ( Count3 ++;Count3 < MaxSize;Count1 ++,Count3 ++ )
            { Set1[ Count1 ] = Set1[ Count3 ]; }

        //
        //  Finally, we zero any junk we may have left at the end of 'Set1'.
        //
        ZeroValues( & Set1[ Count1 ],(MaxSize - Count1) );

        //
        //  Update the 'Used1' value.
        //
        (*Used1) = Count1;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Merge Caches.                                                                             */
            /*                                                                                            */
            /*  All the caches are merged into the set.                                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::MergeCaches
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CACHES                              *Caches,
            ROOT                                *Root
            )
        {
        AUTO FILE_SIZE NewSize = (Root -> Used + (Caches -> DeleteCache.Size()) + (Caches -> NewCache.Size()));
        AUTO VALUE *Set1;
        AUTO VALUE *Set2;
        AUTO VALUE *Set3;
        AUTO SBIT32 Size2;
        AUTO SBIT32 Size3;

        //
        //  We need to ensure that there is enough space in 'Set1' for the final result set.  If
        //  not, we will need to 'Resize()' it.  We then need to prepare 'Set1', 'Set2' and 'Set3'
        //  so we can merge them all into 'Set1'.
        //
        if
                (
                ((NewSize <= Root -> MaxSize) || (Resize( AutomaticHeapScope,Root,NewSize )))
                    &&
                (AutomaticHeapScope -> Update( Root -> Set,((VOID**) & Set1) ))
                    &&
                (Caches -> DeleteCache.ResultSet( & Set2,& Size2 ))
                    &&
                (Caches -> NewCache.ResultSet( & Set3,& Size3 ))
                )
            {
            //
            //  We apply the changes in the 'NewCache' and 'DeleteCache' to the 'Root -> Set'.
            //
            JoinSets( Root -> MaxSize,Set1,Set3,& Root -> Used,Size3 );

            DifferenceSets( Root -> MaxSize,Set1,Set2,& Root -> Used,Size2 );

            //
            //  We can now zero 'DeleteCache' and 'NewCache'.
            //
            Caches -> DeleteCache.DeleteAll();
            Caches -> NewCache.DeleteAll();

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Set Value.                                                                            */
            /*                                                                                            */
            /*  A new value is added to a set.                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::NewSetValue
            (
            VALUE                               *Set,
            FILE_SIZE                           *Used,
            VALUE                               *Value
            )
        {
        AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);

        //
        //  We way we add a 'Value' depends on the size of the 'Set'.
        //
        if ( (*Used) == 0 )
            {
            //
            //  When the 'Set' is empty the adding a 'Value' is trivial.
            //
            Set[0] = (*Value);

            (*Used) = 1;

            return True;
            }
        else
            {
            AUTO FILE_SIZE Offset;

            //
            //  When the 'Set' is larger we use a binary search to find any duplicates.
            //
            if ( ! FindSetValue( & Offset,Set,(*Used),Value ) )
                {
                AUTO FILE_SIZE Count;

                //
                //  We now move any greater values down the 'Set'.
                //
                for ( Count=(*Used) ++;Count > Offset;Count -- )
                    { Set[ Count ] = Set[ (Count-1) ]; }

                //
                //  We insert the new 'Value' into the 'Set' and exit.
                //
                Set[ Offset ] = (*Value);

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Value.                                                                                */
            /*                                                                                            */
            /*  A new value is added to a set.                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::NewValue
            (
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (NewValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Value.                                                                                */
            /*                                                                                            */
            /*  A new value is added to a set.                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::NewValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the set.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  We must have enough space to add the new 'Value'.
                //
                switch ( Root -> Type )
                    {
                    case SimpleRowSet:
                    case ExtendedRowSet:
                        {
                        if ( ! Resize( AutomaticHeapScope,Root,(Root -> Used + 1) ) )
                            { return False; }
                        }
                    }

                //
                //  Now we can add the new value.
                //
                switch ( Root -> Type )
                    {
                    case SimpleRowSet:
                        {
                        //
                        //  We now add the new 'Value'.
                        //
                        if ( NewSetValue( & Root -> Value,& Root -> Used,((VALUE*) & Value) ) )
                            { return True; }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorUnableToStoreValue )); }
                        }

                    case ExtendedRowSet:
                        {
                        AUTO VALUE *Set;

                        //
                        //  We will need to 'Update()' the associated 'Set'.
                        //
                        if ( AutomaticHeapScope -> Update( Root -> Set,((VOID**) & Set) ) )
                            {
                            if ( NewSetValue( Set,& Root -> Used,((VALUE*) & Value) ) )
                                { return True; }
                            else
                                { return (AutomaticHeapScope -> SetError( ErrorUnableToStoreValue )); }
                            }

                        break;
                        }

                    case ComplexRowSet:
                        {
                        AUTO CACHES *Caches;

                        //
                        //  We will need to 'Update()' the associated 'Caches'.
                        //
                        if ( AutomaticHeapScope -> Update( Root -> Caches,((VOID**) & Caches) ) )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                            AUTO FILE_ADDRESS Offset;
                            AUTO VALUE *Set;

                            //
                            //  We will need to 'MergeCaches()' if there is not enough space
                            //  to delete a value.
                            //
                            if ( Caches -> NewCache.Full() )
                                { MergeCaches( AutomaticHeapScope,Caches,Root ); }

                            //
                            //  We will need to 'View()' the associated 'Set'.
                            //
                            if ( AutomaticViewScope.View( Root -> Set,((VOID**) & Set) ) )
                                {
                                AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);

                                //
                                //  We search the 'Set' for the 'Value'.  If it is present
                                //  then it must not be in the 'DeleteCache' or it is an error.
                                //  If it is not present in the 'Set' we put it in the
                                //  'NewCache' or it is an error.
                                //
                                if ( FindSetValue( & Offset,Set,Root -> Used,((VALUE*) & Value) ) )
                                    {
                                    if ( Caches -> DeleteCache.DeleteValue( Value ) )
                                        { return True; }
                                    else
                                        { return (AutomaticHeapScope -> SetError( ErrorUnableToStoreValue )); }
                                    }
                                else
                                    {
                                    if ( Caches -> NewCache.NewValue( Value ) )
                                        { return True; }
                                    else
                                        { return (AutomaticHeapScope -> SetError( ErrorUnableToStoreValue )); }
                                    }
                                }
                            }

                        break;
                        }

                    default:
                        { FAILURE( "Unknown type" ); }
                    }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Normalize Sets.                                                                           */
            /*                                                                                            */
            /*  We Normalize the sets into a standard format.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::NormalizeSets
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope1,
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope2,
            AUTOMATIC_VIEW_SCOPE<HEAP>          *AutomaticViewScope2,
            ROOT                                *Root1,
            ROOT                                **Root2,
            VALUE                               **Set1,
            VALUE                               **Set2,
            SET<HEAP,LOCK,VALUE>                *Source
            )
        {
        //
        //  The placement of the data depends on the 'Type' of the 'Set'.
        //
        switch ( Root1 -> Type )
            {
            case SimpleRowSet:
                {
                (*Set1) = & Root1 -> Value;

                break;
                }

            case ExtendedRowSet:
                {
                //
                //  We will need to 'Update()' the associated 'Set'.
                //
                if ( ! AutomaticHeapScope1 -> Update( Root1 -> Set,((VOID**) Set1) ) )
                    { return False; }

                break;
                }

            case ComplexRowSet:
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope2 );
                AUTO CACHES *Caches1;

                //
                //  We will need to 'ExclusiveView()' the associated 'Caches'.
                //
                if ( AutomaticViewScope1.ExclusiveView( Root1 -> Caches,((VOID**) & Caches1) ) )
                    {
                    //
                    //  We will need to 'MergeCaches()' if there have been any changes.
                    //
                    if ( ((Caches1 -> DeleteCache.Size()) > 0) || ((Caches1 -> NewCache.Size()) > 0) )
                        {
                        //
                        //  When there have been changes we need to 'Update()' the
                        //  'Caches' so we can 'MergeCaches()'.
                        //
                        if
                                (
                                (! AutomaticHeapScope1 -> Update( Root1 -> Caches,((VOID**) & Caches1) ))
                                    ||
                                (! MergeCaches( AutomaticHeapScope1,Caches1,Root1 ))
                                )
                            { return False; }
                        }

                    //
                    //  We need to 'Update()' the associated 'Set'.
                    //
                    if ( ! AutomaticHeapScope1 -> Update( Root1 -> Set,((VOID**) Set1) ) )
                        { return False; }
                    }
                else
                    { return False; }

                break;
                }

            default:
                { FAILURE( "Unknown type" ); }
            }

        //
        //  The placement of the data depends on the 'Type' of the 'Set'.
        //
        switch ( (*Root2) -> Type )
            {
            case SimpleRowSet:
                {
                (*Set2) = & (*Root2) -> Value;

                break;
                }

            case ExtendedRowSet:
                {
                //
                //  We will need to 'View()' the associated 'Set'.
                //
                if ( ! AutomaticViewScope2 -> View( (*Root2) -> Set,((VOID**) Set2) ) )
                    { return False; }

                break;
                }

            case ComplexRowSet:
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope2 );
                AUTO CACHES *Caches2;

                //
                //  We will need to 'ExclusiveView()' the associated 'Caches'.
                //
                if ( AutomaticViewScope3.ExclusiveView( (*Root2) -> Caches,((VOID**) & Caches2) ) )
                    {
                    //
                    //  We will need to 'MergeCaches()' if there have been any changes.
                    //
                    if ( ((Caches2 -> DeleteCache.Size()) > 0) || ((Caches2 -> NewCache.Size()) > 0) )
                        {
                        //
                        //  When there have been changes we need to 'Update()' the
                        //  'Caches' so we can 'MergeCaches()'.
                        //
                        if
                                (
                                (! AutomaticHeapScope2 -> Update( Source -> Address,((VOID**) Root2) ))
                                    ||
                                (! AutomaticHeapScope2 -> Update( (*Root2) -> Caches,((VOID**) & Caches2) ))
                                    ||
                                (! MergeCaches( AutomaticHeapScope2,Caches2,(*Root2) ))
                                )
                            { return False; }
                        }

                    //
                    //  We need to 'View()' the associated 'Set'.
                    //
                    if ( ! AutomaticViewScope2 -> View( (*Root2) -> Set,((VOID**) Set2) ) )
                        { return False; }
                    }
                else
                    { return False; }

                break;
                }

            default:
                { FAILURE( "Unknown type" ); }
            }

        return True;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Resize.                                                                                   */
            /*                                                                                            */
            /*  Resize a row set.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Resize
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            ROOT                                *Root,
            FILE_SIZE                           Size
            )
        {
        AUTO BOOLEAN Result = False;

        //
        //  Let's see if we need to 'Resize()' the set.
        //
        if ( Size > Root -> MaxSize )
            {
            AUTO FILE_SIZE Space;
            AUTO VALUE *Set;

            //
            //  The way we 'Resize()' a set depends on its 'Type'.
            //
            switch ( Root -> Type )
                {
                case SimpleRowSet:
                    {
                    AUTO FILE_SIZE NewSize;

                    //
                    //  Compute the extra space we need in 'NewSize'.
                    //
                    for ( NewSize = MinExtendedSize;Size > NewSize;/* void */ )
                        { NewSize *= GrowthFactor; }

                    //
                    //  A 'SimpleRowSet' can only has space to store a single 'Value'.
                    //  When we need to expand this space we need to keep this 'Value'.
                    //
                    if ( ROCKALL_NEW<VALUE>::New( AutomaticHeapScope,& Root -> Set,& Set,NewSize,& Space,False,True ) )
                        {
                        AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);
                        AUTO FILE_SIZE Count;

                        //
                        //  We now update 'Root' to reflect the changes.
                        //
                        Root -> MaxSize = (Space / sizeof(VALUE));
                        Root -> Type = ExtendedRowSet;

                        //
                        //  Call the extra constructors.
                        //
                        for ( Count=NewSize;Count < Root -> MaxSize;Count ++ )
                            { PLACEMENT_NEW( & Set[ Count ],VALUE ); }

                        //
                        //  We will copy any original 'Value' into the new 'Set'.
                        //
                        if ( Root -> Used == 1 )
                            { Set[0] = Root -> Value; }

                        Result = True;
                        }

                    break;
                    }

                case ExtendedRowSet:
                case ComplexRowSet:
                    {
                    AUTO FILE_SIZE NewMaxSize;
                    AUTO FILE_SIZE NewSize;

                    //
                    //  Compute the extra space we need in 'NewMaxSize' and 'NewSize'.
                    //
                    for
                        (
                        NewMaxSize = (Root -> MaxSize * GrowthFactor), NewSize = (NewMaxSize * sizeof(VALUE));
                        Size > NewMaxSize;
                        NewMaxSize *= GrowthFactor, NewSize = (NewMaxSize * sizeof(VALUE))
                        );

                    //
                    //  We now 'Resize()' the 'ExpandedRowSet'.
                    //
                    if ( AutomaticHeapScope -> Resize( & Root -> Set,((VOID**) & Set),NewSize,& Space,True ) )
                        {
                        AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Set);
                        AUTO FILE_SIZE Count = Root -> MaxSize;

                        //
                        //  We now update the 'Root' to reflect the changes.
                        //
                        Root -> MaxSize = (Space / sizeof(VALUE));

                        //
                        //  Call the extra constructors.
                        //
                        for ( /* void */;Count < Root -> MaxSize;Count ++ )
                            { PLACEMENT_NEW( & Set[ Count ],VALUE ); }

                        Result = True;
                        }

                    break;
                    }

                default:
                    { FAILURE( "Unknown type" ); }
                }
            }
        else
            { return True; }

        //
        //  An 'ExpandedRowSet' is a simple array of 'VALUE'.  We will only allow this to
        //  grow to 'MaxExtendedSize' before we convert it to a 'ComplexRowSet'.
        //
        if ( (Result) && (Root -> Type == ExtendedRowSet) && (Root -> MaxSize > MaxExtendedSize) )
            {
            AUTO CACHES *Caches;

            //
            //  We now convert an 'ExpandedRowSet' into a 'ComplexRowSet'.  A 'ComplexRowSet'
            //  is simply an 'ExpandedRowSet' with a cache.
            //
            if ( ROCKALL_NEW<CACHES>::New( AutomaticHeapScope,& Root -> Caches,& Caches,1,NULL,False,True ) )
                { Root -> Type = ComplexRowSet; }
            }

        return Result;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  We return the size of a set.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Size
            (
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Size( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  We return the size of a set.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Size
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           *Size
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the queue.
            //
            if
                    (
                    (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  The way we compute the 'Size' depends on the 'Type' of the set.
                //
                switch ( Root -> Type )
                    {
                    case SimpleRowSet:
                    case ExtendedRowSet:
                        {
                        //
                        //  We return the 'Size'.
                        //
                        (*Size) = Root -> Used;

                        return True;
                        }

                    case ComplexRowSet:
                        {
                        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                        AUTO CACHES *Caches;

                        //
                        //  We will need to 'View()' the associated 'Caches'.
                        //
                        if ( AutomaticViewScope2.View( Root -> Caches,((VOID**) & Caches) ) )
                            {
                            //
                            //  We return the 'Size'.
                            //
                            (*Size) = (Root -> Used - (Caches -> DeleteCache.Size()) + (Caches -> NewCache.Size()));

                            return True;
                            }

                        break;
                        }

                    default:
                        { FAILURE( "Unknown type" ); }
                    }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Value.                                                                                    */
            /*                                                                                            */
            /*  The value of a set.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Value
            (
            FILE_SIZE                           MaxSize,
            FILE_SIZE                           *Size,
            VALUE                               *Values
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Value( & AutomaticHeapScope,MaxSize,Size,Values ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Value.                                                                                    */
            /*                                                                                            */
            /*  The value of a set.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Value
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           MaxSize,
            FILE_SIZE                           *Size,
            VALUE                               *Values
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()',
        //  'ExclusiveView()' and 'Update()' should be automatically optimized away for
        //  heaps that do not support this functionality by any decent compiler, due to
        //  the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'ExclusiveView()' the 'Root' as we need to view it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the queue.
            //
            if
                    (
                    (AutomaticHeapScope -> ExclusiveView( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  The way we view the data depends on the 'Type' of the set.
                //
                switch ( Root -> Type )
                    {
                    case SimpleRowSet:
                        {
                        AUTO FILE_SIZE Count;

                        //
                        //  We copy the values into the user supplied 'Set'.
                        //
                        for ( Count=0;(Count < MaxSize) && (Count < Root -> Used);Count ++ )
                            { Values[ Count ] = ((VALUE*) & Root -> Value)[ Count ]; }

                        //
                        //  We return the 'Size'.
                        //
                        (*Size) = Count;

                        return True;
                        }

                    case ExtendedRowSet:
                        {
                        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                        AUTO VALUE *AllValues;

                        //
                        //  We will need to 'View()' the associated 'ResultSet'.
                        //
                        if ( AutomaticViewScope2.View( Root -> Set,((VOID**) & AllValues) ) )
                            {
                            AUTO VALUE (*View2)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) AllValues);
                            AUTO FILE_SIZE Count;

                            //
                            //  We copy the 'ResultSet' into the supplied 'set'.
                            //
                            for ( Count=0;(Count < MaxSize) && (Count < Root -> Used);Count ++ )
                                { Values[ Count ] = AllValues[ Count ]; }

                            //
                            //  We return the 'Size'.
                            //
                            (*Size) = Count;

                            return True;
                            }

                        break;
                        }

                    case ComplexRowSet:
                        {
                        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                        AUTO CACHES *Caches;

                        //
                        //  We will need to 'ExclusiveView()' the 'Caches'.
                        //
                        if ( AutomaticViewScope2.ExclusiveView( Root -> Caches,((VOID**) & Caches) ) )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope );
                            AUTO VALUE *AllValues;

                            //
                            //  We will need to 'MergeCaches()' if there is any changes.
                            //
                            if ( ((Caches -> DeleteCache.Size()) > 0) || ((Caches -> NewCache.Size()) > 0) )
                                {
                                //
                                //  We will need to 'Update()' the 'Caches'.
                                //
                                if ( AutomaticHeapScope -> Update( Root -> Caches,((VOID**) & Caches) ) )
                                    { MergeCaches( AutomaticHeapScope,Caches,Root ); }
                                }

                            //
                            //  We will need to 'View()' the associated 'ResultSet'.
                            //
                            if ( AutomaticViewScope3.View( Root -> Set,((VOID**) & AllValues) ) )
                                {
                                AUTO VALUE (*View2)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) AllValues);
                                AUTO FILE_SIZE Count;

                                //
                                //  We copy the 'ResultSet' into the supplied 'set'.
                                //
                                for ( Count=0;(Count < MaxSize) && (Count < Root -> Used);Count ++ )
                                    { Values[ Count ] = AllValues[ Count ]; }

                                //
                                //  We return the 'Size'.
                                //
                                (*Size) = Count;

                                return True;
                                }
                            }

                        break;
                        }

                    default:
                        { FAILURE( "Unknown type" ); }
                    }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a set.                                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::View
            (
            VALUE                               **Set,
            FILE_SIZE                           *Size
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (View( & AutomaticHeapScope,Set,Size ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a set.                                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::View
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            VALUE                               **Set,
            FILE_SIZE                           *Size
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()',
        //  'ExclusiveView()' and 'Update()' should be automatically optimized away for
        //  heaps that do not support this functionality by any decent compiler, due to
        //  the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticHeapScope -> BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'ExclusiveView()' the 'Root' as we need to view it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the queue.
            //
            if
                    (
                    (AutomaticHeapScope -> ExclusiveView( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  The way we view the data depends on the 'Type' of the set.
                //
                switch ( Root -> Type )
                    {
                    case SimpleRowSet:
                        {
                        //
                        //  We return the 'Set' and the 'Size'.
                        //
                        (*Set) = ((VALUE*) & Root -> Value);

                        if ( Size != NULL )
                            { (*Size) = Root -> Used; }

                        return True;
                        }

                    case ExtendedRowSet:
                        {
                        //
                        //  We will need to 'View()' the associated 'Set'.
                        //
                        if ( AutomaticHeapScope -> View( Root -> Set,((VOID**) Set) ) )
                            {
                            AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) (*Set));

                            //
                            //  We return the 'Set' and the 'Size'.
                            //
                            if ( Size != NULL )
                                { (*Size) = Root -> Used; }

                            return True;
                            }

                        break;
                        }

                    case ComplexRowSet:
                        {
                        AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                        AUTO CACHES *Caches;

                        //
                        //  We will need to 'ExclusiveView()' the 'Caches'.
                        //
                        if ( AutomaticViewScope.ExclusiveView( Root -> Caches,((VOID**) & Caches) ) )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope3( AutomaticHeapScope );

                            //
                            //  We will need to 'MergeCaches()' if there is any changes.
                            //
                            if ( ((Caches -> DeleteCache.Size()) > 0) || ((Caches -> NewCache.Size()) > 0) )
                                {
                                //
                                //  We will need to 'Update()' the 'Caches'.
                                //
                                if ( AutomaticHeapScope -> Update( Root -> Caches,((VOID**) & Caches) ) )
                                    { MergeCaches( AutomaticHeapScope,Caches,Root ); }
                                }

                            //
                            //  We will need to 'View()' the associated 'Set'.
                            //
                            if ( AutomaticHeapScope -> View( Root -> Set,((VOID**) Set) ) )
                                {
                                AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) (*Set));

                                //
                                //  We return the 'Set' and the 'Size'.
                                //
                                if ( Size != NULL )
                                    { (*Size) = Root -> Used; }

                                return True;
                                }
                            }

                        break;
                        }

                    default:
                        { FAILURE( "Unknown type" ); }
                    }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk the set.                                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Walk
            (
            SET_WALK_FUNCTION                   WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Walk( & AutomaticHeapScope,WalkFunction,UserValue ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk the set.                                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN SET<HEAP,LOCK,VALUE>::Walk
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            SET_WALK_FUNCTION                   WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO VALUE *Set;
        AUTO FILE_SIZE Size;

        //
        //  We need to 'View()' the 'Set'.
        //
        if ( View( AutomaticHeapScope,& Set,& Size ) )
            {
            AUTO BOOLEAN Result = True;
            AUTO FILE_SIZE Count;

            //
            //  Walk the 'Set'.
            //
            for ( Count=0;(Result) && (Count < Size);Count ++ )
                {
                //
                //  We simply don't trust user supplied code so we prepare for trouble.
                //
#ifndef DISABLE_CRT_FEATURES
                try
                    {
#endif
                    //
                    //  Execute the user supplied walk function.
                    //
                    if ( ! WalkFunction( Set[ Count ],UserValue ) )
                        { Result = False; }
#ifndef DISABLE_CRT_FEATURES
                    }
                catch ( ... )
                    {
                    //
                    //  We will catch and skip all errors detected in the user supplied
                    //  walk function.
                    //
                    return (AutomaticHeapScope -> SetError( ErrorExceptionCaught ));
                    }
#endif
                }

            //
            //  End the view of the 'Set'.
            //
            if ( ! EndView( AutomaticHeapScope ) )
                { Result = False; }

            return Result;
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero Values.                                                                              */
            /*                                                                                            */
            /*  Zero a collection of values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> VOID SET<HEAP,LOCK,VALUE>::ZeroValues
            (
            VALUE                               *Values,
            FILE_SIZE                           Size
            )
        {
        AUTO FILE_SIZE Count;

        //
        //  We have to be careful as a 'VALUE' contains user types.  It is very hard to zero
        //  these as we would not wish to damage them.  So just to be safe we call the
        //  destructor, then we can zero them and then finally we call the constructor.
        //
        for ( Count=(Size-1);Count >= 0;Count -- )
            { PLACEMENT_DELETE( & Values[ Count ],VALUE ); }

        MemoryFunctions.Zero( Values,((SNATIVE) (Size * sizeof(VALUE))) );

        for ( Count=0;Count < Size;Count ++ )
            { PLACEMENT_NEW( & Values[ Count ],VALUE ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'SET'.                                                                     */
            /*                                                                                            */
            /*  The destructor for 'SET' cleans up the class.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> SET<HEAP,LOCK,VALUE>::~SET( VOID )
        {
        //
        //  We need to ensure that the class is 'Active'.
        //
        if ( Active )
            {
            AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;
            AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( & AutomaticHeapScope );
            AUTO TRANSACTIONAL_LOCK TransactionLock( & AutomaticHeapScope );

            //
            //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
            //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
            //  should be automatically optimized away for heaps that do not support this
            //  functionality by any decent compiler, due to the way the code has been written.
            //
            if
                    (
                    (Active)
                        &&
                    (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
                AUTO ROOT *Root;

                //
                //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
                //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
                //  above to prevent threads in the same transaction from accessing the queue.
                //
                if
                        (
                        (AutomaticHeapScope.Update( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                    //
                    //  We 'DeleteAll()' 'Set' values.
                    //
                    if ( DeleteAll( & AutomaticHeapScope,Root ) )
                        {
                        //
                        //  All we should have left now is a 'SimpleRowSet'.
                        //
                        if  ( Root -> Type == SimpleRowSet )
                            {
                            //
                            //  We zero 'Value' just to be tidy.
                            //
                            ZeroValues( & Root -> Value,1 );

                            //
                            //  It is not very hard to delete a 'SimpleRowSet'.
                            //
                            Root -> Used = 0;
                            Root -> Type = SimpleRowSet;
                            Root -> Signature = 0;
                            Root -> Set = NULL;
                            Root -> MaxSize = 0;
                            Root -> Caches = NULL;

                            //
                            //  We are no longer 'Active'.
                            //
                            Active = False;

                            //
                            //  Release the lock on the 'Root'.
                            //
                            AutomaticExclusiveLock2.Release();

                            //
                            //  Delete the 'Root' structure.
                            //
                            if ( ROCKALL_DELETE<ROOT>::Delete( & AutomaticHeapScope,Address ) )
                                {
                                //
                                //  Zero class members.
                                //
                                Address = NULL;
                                Active = False;
                                }
                            else
                                { FAILURE( "Unable to delete the root" ); }
                            }
                        else
                            { FAILURE( "Unable to delete simple row set" ); }
                        }
                    else
                        { FAILURE( "Unable to delete the root structure" ); }
                    }
                else
                    { FAILURE( "Unable to update the root structure" ); }
                }
            else
                { FAILURE( "Unable to update the root structure" ); }
            }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define SET ROCKALL::SET
#endif
#endif