#ifndef _STACK_HPP_
#define _STACK_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "AutomaticLockScope.hpp"
#include "AutomaticTransactionScope.hpp"
#include "AutomaticViewScope.hpp"
#include "Failure.hpp"
#include "FileLock.hpp"
#include "MemoryFunctions.hpp"
#include "MemoryLock.hpp"
#include "NoLock.hpp"
#include "PlacementNew.hpp"
#include "RockallDelete.hpp"
#include "RockallNew.hpp"
#include "TransactionLock.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Stack.                                                                                            */
        /*                                                                                                    */
        /*  A stack is a simple push and pop data structure for storing values.                               */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Stack.                                                                                    */
            /*                                                                                            */
            /*  A stack is a simple push and pop data structure for storing values.                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> class STACK
        {
            //
            //  Private constants.
            //
            STATIC CONST SBIT32 CacheSize       = 128;
            STATIC CONST SBIT32 DefaultSize     = (CacheSize * 4);
            STATIC CONST SBIT32 HalfCacheSize   = (CacheSize / 2);
            STATIC CONST SBIT32 ViewSize        = 8192;

            //
            //  The root signature.
            //
            STATIC CONST SBIT64 Signature1      = ((sizeof(LOCK) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature2      = (((Signature1 + sizeof(VALUE)) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature       = (0x0000400000000000 + Signature2);

            //
            //  Private types.
            //
            typedef TRANSACTION_LOCK<HEAP> TRANSACTIONAL_LOCK;

            typedef struct
                {
                VALUE                           Values[ CacheSize ];
                SBIT32                          Used;
                }
            CACHE;

            //
            //  Private types.
            //
            typedef struct
                {
                //
                //  Data locks.
                //
                LOCK                            ClassLock;

                //
                //  Protected data.
                //
                CACHE                           Cache;
                FILE_SIZE                       MaxSize;
                FILE_SIZE                       Signature;
                FILE_ADDRESS                    Stack;
                FILE_SIZE                       Used;
                }
            ROOT;

            //
            //  Private data.
            //
            BOOLEAN                             Active;
            FILE_ADDRESS                        Address;

            //
            //  Private static data.
            //
            STATIC MEMORY_FUNCTIONS<SNATIVE>    MemoryFunctions;

        public:
            //
            //  Public types.
            //
            typedef BOOLEAN (*STACK_WALK_FUNCTION)
                (
                VALUE                           & Value,
                VOID                            *UserValue
                );

            //
            //  Public functions.
            //
            STACK( VOID );
            STACK( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN DeleteAll( VOID );
            BOOLEAN DeleteAll( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN PeekValue( VALUE *Value );
            BOOLEAN PeekValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,VALUE *Value );

            BOOLEAN PopValue( VALUE *Value );
            BOOLEAN PopValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,VALUE *Value );

            BOOLEAN PushValue( CONST VALUE & Value );
            BOOLEAN PushValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST VALUE & Value );

            BOOLEAN Size( FILE_SIZE *Value );
            BOOLEAN Size( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_SIZE *Value );

            BOOLEAN Walk( STACK_WALK_FUNCTION WalkFunction,VOID *UserValue = NULL );

            BOOLEAN Walk
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                STACK_WALK_FUNCTION             WalkFunction,
                VOID                            *UserValue = NULL
                );

            ~STACK( VOID );

        private:
            //
            //  Private functions.
            //
            BOOLEAN DeleteAll( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,ROOT *Root );

            BOOLEAN MergeCache( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,ROOT *Root );

            VOID ZeroCache( CACHE *Cache );

            VOID ZeroValues( VALUE *Values,FILE_SIZE Size );

            //
            //  Disabled functionality.
            //
            STACK( CONST STACK & NoCopies );

            VOID operator=( CONST STACK & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Class Static Data.                                                                        */
            /*                                                                                            */
            /*  The class static data needs to be initialised before it's ready for use.                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> MEMORY_FUNCTIONS<SNATIVE> STACK<HEAP,LOCK,VALUE>::MemoryFunctions;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'STACK'.                                                                  */
            /*                                                                                            */
            /*  The constructor for 'STACK' prepares the class for use.                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> STACK<HEAP,LOCK,VALUE>::STACK( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        //
        //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
        //  we fail to highlight the issue.
        //
        if ( AutomaticHeapScope.Feature( (LOCK::Feature()) ) )
            {
            //
            //  Setup class members.
            //
            Active = False;
            Address = NULL;

            //
            //  Where necessary we automatically wrap all the changes within a new transaction.
            //
            if ( AutomaticHeapScope.BeginTransaction() )
                {
                AUTO ROOT *Root;

                //
                //  We create a new 'Root' on the heap and store the 'Address'.
                //
                if ( ROCKALL_NEW<ROOT>::New( & AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                    {
                    //
                    //  We now setup the 'Root' of the 'Stack'.
                    //
                    Root -> MaxSize = 0;
                    Root -> Signature = Signature;
                    Root -> Stack = NULL;
                    Root -> Used = 0;

                    //
                    //  Zero the 'Cache'.
                    //
                    ZeroCache( & Root -> Cache );

                    Active = True;
                    }
                }
            }
        else
            { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'STACK'.                                                                  */
            /*                                                                                            */
            /*  The constructor for 'STACK' prepares the class for use.                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> STACK<HEAP,LOCK,VALUE>::STACK
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        //
        //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
        //  we fail to highlight the issue.
        //
        if ( AutomaticHeapScope -> Feature( (LOCK::Feature()) ) )
            {
            AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

            //
            //  Setup class members.
            //
            Active = False;
            Address = NULL;

            //
            //  Where necessary we automatically wrap all the changes within a new transaction.
            //
            if ( AutomaticTransactionScope.BeginTransaction() )
                {
                AUTO ROOT *Root;

                //
                //  We create a new 'Root' on the heap and store the 'Address'.
                //
                if ( ROCKALL_NEW<ROOT>::New( AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                    {
                    //
                    //  We now setup the 'Root' of the 'Stack'.
                    //
                    Root -> MaxSize = 0;
                    Root -> Signature = Signature;
                    Root -> Stack = NULL;
                    Root -> Used = 0;

                    //
                    //  Zero the 'Cache'.
                    //
                    ZeroCache( & Root -> Cache );

                    Active = True;
                    }
                }
            }
        else
            { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the stack values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::DeleteAll( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (DeleteAll( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the stack values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::DeleteAll
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the stack.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Delete all the values.
                //
                return (DeleteAll( AutomaticHeapScope,Root ));
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the stack values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::DeleteAll
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            ROOT                                *Root
            )
        {
        //
        //  We delete the 'Stack' if it exists.
        //
        if ( Root -> MaxSize > 0 )
            {
            AUTO VALUE *Stack;

            //
            //  We will need to 'Update()' the 'Stack'.
            //
            if ( AutomaticHeapScope -> Update( Root -> Stack,((VOID**) & Stack) ) )
                {
                AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Stack);
                AUTO FILE_SIZE Count;

                //
                //  Call the extra destructors.
                //
                for ( Count=(Root -> MaxSize-1);Count >= DefaultSize;Count -- )
                    { PLACEMENT_DELETE( & Stack[ Count ],VALUE ); }

                //
                //  Delete the 'Stack'.
                //
                if ( ROCKALL_DELETE<VALUE>::Delete( AutomaticHeapScope,Root -> Stack,DefaultSize ) )
                    {
                    Root -> MaxSize = 0;
                    Root -> Stack = NULL;
                    Root -> Used = 0;

                    //
                    //  Finally, we zero the 'Cache'.
                    //
                    ZeroCache( & Root -> Cache );

                    return True;
                    }
                else
                    { FAILURE( "Unable to delete the stack values" ); }
                }

            return False;
            }
        else
            {
            //
            //  We zero the 'Cache'.
            //
            ZeroCache( & Root -> Cache );

            return True;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Merge Cache.                                                                              */
            /*                                                                                            */
            /*  Merge the cache into the stack.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::MergeCache
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            ROOT                                *Root
            )
        {
        AUTO BOOLEAN Result = True;

        //
        //  Let's see if the 'Cache' is empty.  If so, we will try to refill it.
        //
        if ( Root -> Cache.Used <= 0 )
            {
            //
            //  Let's see if the 'Stack' is empty.  If not, we refill the 'Cache'.
            //
            if ( Root -> Used > 0 )
                {
                AUTO VALUE *Stack;

                //
                //  We will need to 'Update()' the 'Stack'.
                //
                if ( AutomaticHeapScope -> Update( Root -> Stack,((VOID**) & Stack) ) )
                    {
                    AUTO SBIT32 FillSize = ((Root -> Used > HalfCacheSize) ? HalfCacheSize : ((SBIT32) Root -> Used));
                    AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Stack);
                    AUTO SBIT32 Count;

                    //
                    //  We refill the 'Cache' from the 'Stack'.
                    //
                    for ( Count=(FillSize-1);Count >= 0;Count -- )
                        { Root -> Cache.Values[ Count ] = Stack[ (-- Root -> Used) ]; }

                    //
                    //  We now zero the empty space just to be tidy.
                    //
                    ZeroValues( & Stack[ Root -> Used ],FillSize );

                    //
                    //  Update the size of the 'Cache'.
                    //
                    Root -> Cache.Used = FillSize;
                    }
                else
                    { Result = False; }
                }
            else
                { Result = False; }
            }

        //
        //  Let's see if the 'Cache' is full.  If so, we empty it.
        //
        if ( Root -> Cache.Used >= CacheSize )
            {
            //
            //  Let's see if we have enough space on the 'Stack'.
            //
            if ( (Root -> Used + HalfCacheSize) > Root -> MaxSize )
                {
                //
                //  We will create a new 'Stack' if needed.
                //
                if ( Root -> MaxSize <= 0 )
                    {
                    AUTO FILE_SIZE Space;
                    AUTO VALUE *Stack;

                    //
                    //  Create a new 'Stack' and compute 'MaxSize' from the space we are
                    //  actually allocated, which might be more than we requested.
                    //
                    if
                            (
                            ROCKALL_NEW<VALUE>::New
                                (
                                AutomaticHeapScope,
                                & Root -> Stack,
                                & Stack,
                                DefaultSize,
                                & Space,
                                False,
                                True
                                )
                            )
                        {
                        AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Stack);
                        AUTO SBIT32 Count;

                        //
                        //  Update 'MaxSize'.
                        //
                        Root -> MaxSize = (Space / sizeof(VALUE));

                        //
                        //  Call the extra constructors.
                        //
                        for ( Count=DefaultSize;Count < Root -> MaxSize;Count ++ )
                            { PLACEMENT_NEW( & Stack[ Count ],VALUE ); }
                        }
                    else
                        { Result = False; }
                    }
                else
                    {
                    AUTO FILE_SIZE NewSize = ((Root -> MaxSize * GrowthFactor) * sizeof(VALUE));
                    AUTO FILE_SIZE Space;
                    AUTO VALUE *Stack;

                    //
                    //  Resize the 'Stack' and compute 'MaxSize' from the space we are
                    //  actually allocated, which might be more than we requested.
                    //
                    if ( AutomaticHeapScope -> Resize( & Root -> Stack,((VOID**) & Stack),NewSize,& Space,True ) )
                        {
                        AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Stack);
                        AUTO FILE_SIZE Count = Root -> MaxSize;

                        //
                        //  Compute the new 'MaxSize'.
                        //
                        Root -> MaxSize = (Space / sizeof(VALUE));

                        //
                        //  Call the extra constructors.
                        //
                        for ( /* void */;Count < Root -> MaxSize;Count ++ )
                            { PLACEMENT_NEW( & Stack[ Count ],VALUE ); }
                        }
                    else
                        { Result = False; }
                    }
                }

            //
            //  We now push half of the 'Cache' onto the 'Stack'.
            //
            if ( Result )
                {
                AUTO VALUE *Stack;

                //
                //  We will need to 'Update()' the 'Stack'.
                //
                if ( AutomaticHeapScope -> Update( Root -> Stack,((VOID**) & Stack) ) )
                    {
                    AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Stack);
                    AUTO SBIT32 Count;

                    //
                    //  Now we empty half the 'Cache'.
                    //
                    for ( Count=0;Count < HalfCacheSize;Count ++ )
                        { Stack[ (Root -> Used ++) ] = Root -> Cache.Values[ Count ]; }

                    //
                    //  We move any remaining values down the 'Cache'.
                    //
                    for ( Count=0;Count < HalfCacheSize;Count ++ )
                        { Root -> Cache.Values[ Count ] = Root -> Cache.Values[ (HalfCacheSize + Count) ]; }

                    //
                    //  We now zero the empty space just to be tidy.
                    //
                    ZeroValues( & Root -> Cache.Values[ HalfCacheSize ],HalfCacheSize );

                    //
                    //  Update the size of the 'Cache'.
                    //
                    Root -> Cache.Used = HalfCacheSize;
                    }
                else
                    { Result = False; }
                }
            }

        return Result;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Peek Value.                                                                               */
            /*                                                                                            */
            /*  Peek at the top of the stack.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::PeekValue( VALUE *Value )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (PeekValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Peek Value.                                                                               */
            /*                                                                                            */
            /*  Peek at the top of the stack.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::PeekValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            VALUE                               *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()',
        //  'ExclusiveView()' and 'Update()' should be automatically optimized away for
        //  heaps that do not support this functionality by any decent compiler, due to
        //  the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'ExclusiveView()' the 'Root' as we need to view it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the stack.
            //
            if
                    (
                    (AutomaticHeapScope -> ExclusiveView( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Let's make sure the 'Cache' is not empty or refill it.
                //
                if
                        (
                            (Root -> Cache.Used > 0)
                                ||
                            (
                            (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                                &&
                            (MergeCache( AutomaticHeapScope,Root ))
                            )
                        )
                    {
                    //
                    //  We peek at the top value in the 'Cache' and return it.
                    //
                    (*Value) = Root -> Cache.Values[ (Root -> Cache.Used-1) ];

                    return True;
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorNoValueAvailable )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Pop Value.                                                                                */
            /*                                                                                            */
            /*  Pop a value off the stack.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::PopValue( VALUE *Value )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (PopValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Pop Value.                                                                                */
            /*                                                                                            */
            /*  Pop a value off the stack.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::PopValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            VALUE                               *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the stack.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Let's make sure the 'Cache' is not empty or refill it.
                //
                if ( (Root -> Cache.Used > 0) || (MergeCache( AutomaticHeapScope,Root )) )
                    {
                    AUTO VALUE *Current = & Root -> Cache.Values[ (-- Root -> Cache.Used) ];

                    //
                    //  We pop the top value on the 'Cache' and return it.  We will also
                    //  zero the 'VALUE' just to be tidy.
                    //
                    (*Value) = (*Current);

                    ZeroValues( Current,1 );

                    return True;
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorNoValueAvailable )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Push Value.                                                                               */
            /*                                                                                            */
            /*  Push a value on the stack.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::PushValue( CONST VALUE & Value )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (PushValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Push Value.                                                                               */
            /*                                                                                            */
            /*  Push a value on the stack.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::PushValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the stack.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Let's make sure the 'Cache' is not full or empty it.
                //
                if ( (Root -> Cache.Used < CacheSize) || (MergeCache( AutomaticHeapScope,Root )) )
                    {
                    //
                    //  We push the value on to the top of the 'Cache'.
                    //
                    Root -> Cache.Values[ (Root -> Cache.Used ++) ] = Value;

                    return True;
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToStoreValue )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the stack.                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::Size( FILE_SIZE *Value )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Size( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the stack.                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::Size
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the stack.
            //
            if
                    (
                    (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  Update 'Value' with the size of the 'Cache' and the 'Stack'.
                //
                (*Value) = (Root -> Cache.Used + Root -> Used);

                return True;
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk the stack.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::Walk
            (
            STACK_WALK_FUNCTION                 WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Walk( & AutomaticHeapScope,WalkFunction,UserValue ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk the stack.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN STACK<HEAP,LOCK,VALUE>::Walk
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            STACK_WALK_FUNCTION                 WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the stack.
            //
            if
                    (
                    (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  If the 'PopCache' is not empty then 'Walk()' all the values.
                //
                if ( Root -> Cache.Used > 0 )
                    {
                    AUTO SBIT32 Count;

                    for ( Count=(Root -> Cache.Used-1);Count >= 0;Count -- )
                        {
                        //
                        //  We simply don't trust user supplied code so we prepare for trouble.
                        //
#ifndef DISABLE_CRT_FEATURES
                        try
                            {
#endif
                            //
                            //  Execute the user supplied walk function.
                            //
                            if ( ! WalkFunction( Root -> Cache.Values[ Count ],UserValue ) )
                                { return False; }
#ifndef DISABLE_CRT_FEATURES
                            }
                        catch ( ... )
                            {
                            //
                            //  We will catch and skip all errors detected in the user supplied
                            //  walk function.
                            //
                            return (AutomaticHeapScope -> SetError( ErrorExceptionCaught ));
                            }
#endif
                        }
                    }

                //
                //  If the 'MaxSize' is not zero then 'Walk()' all the values.
                //
                if ( Root -> MaxSize > 0 )
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope ) ;
                    AUTO VALUE *Stack;

                    //
                    //  We will need to 'View()' the 'Stack'.
                    //
                    if ( AutomaticViewScope2.View( Root -> Stack,((VOID**) & Stack) ) )
                        {
                        AUTO FILE_SIZE Count;

                        //
                        //  Walk the 'Stack'.
                        //
                        for ( Count=(Root -> Used-1);Count >= 0;Count -- )
                            {
                            //
                            //  We simply don't trust user supplied code so we prepare for trouble.
                            //
#ifndef DISABLE_CRT_FEATURES
                            try
                                {
#endif
                                //
                                //  Execute the user supplied walk function.
                                //
                                if ( ! WalkFunction( Stack[ Count ],UserValue ) )
                                    { return False; }
#ifndef DISABLE_CRT_FEATURES
                                }
                            catch ( ... )
                                {
                                //
                                //  We will catch and skip all errors detected in the user supplied
                                //  walk function.
                                //
                                return (AutomaticHeapScope -> SetError( ErrorExceptionCaught ));
                                }
#endif
                            }
                        }
                    }

                return True;
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero Values.                                                                              */
            /*                                                                                            */
            /*  Zero a collection of values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> VOID STACK<HEAP,LOCK,VALUE>::ZeroCache( CACHE *Cache )
        {
        //
        //  We have to be careful as a 'VALUE' contains user types.  It is very hard to zero
        //  these as we would not wish to damage them.  So just to be safe we call the
        //  destructor, then we can zero them and then finally we call the constructor.
        //
        ZeroValues( Cache -> Values,Cache -> Used );

        Cache -> Used = 0;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero Values.                                                                              */
            /*                                                                                            */
            /*  Zero a collection of values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> VOID STACK<HEAP,LOCK,VALUE>::ZeroValues
            (
            VALUE                               *Values,
            FILE_SIZE                           Size
            )
        {
        AUTO FILE_SIZE Count;

        //
        //  We have to be careful as a 'VALUE' contains user types.  It is very hard to zero
        //  these as we would not wish to damage them.  So just to be safe we call the
        //  destructor, then we can zero them and then finally we call the constructor.
        //
        for ( Count=(Size-1);Count >= 0;Count -- )
            { PLACEMENT_DELETE( & Values[ Count ],VALUE ); }

        MemoryFunctions.Zero( Values,((SNATIVE) (Size * sizeof(VALUE))) );

        for ( Count=0;Count < Size;Count ++ )
            { PLACEMENT_NEW( & Values[ Count ],VALUE ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'STACK'.                                                                   */
            /*                                                                                            */
            /*  The destructor for 'STACK' cleans up the class.                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> STACK<HEAP,LOCK,VALUE>::~STACK( VOID )
        {
        //
        //  We need to ensure that the class is 'Active'.
        //
        if ( Active )
            {
            AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;
            AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( & AutomaticHeapScope );
            AUTO TRANSACTIONAL_LOCK TransactionLock( & AutomaticHeapScope );

            //
            //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
            //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
            //  should be automatically optimized away for heaps that do not support this
            //  functionality by any decent compiler, due to the way the code has been written.
            //
            if
                    (
                    (Active)
                        &&
                    (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
                AUTO ROOT *Root;

                //
                //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
                //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
                //  above to prevent threads in the same transaction from accessing the stack.
                //
                if
                        (
                        (AutomaticHeapScope.Update( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                    //
                    //  We 'DeleteAll()' 'Stack' values.
                    //
                    if ( DeleteAll( & AutomaticHeapScope,Root ) )
                        {
                        //
                        //  Zero the 'Root' structure.
                        //
                        Root -> Used = 0;
                        Root -> Stack = NULL;
                        Root -> Signature = 0;
                        Root -> MaxSize = 0;

                        //
                        //  We are no longer 'Active'.
                        //
                        Active = False;

                        //
                        //  Release the lock on the 'Root'.
                        //
                        AutomaticExclusiveLock2.Release();

                        //
                        //  Delete the 'Root' structure.
                        //
                        if ( ROCKALL_DELETE<ROOT>::Delete( & AutomaticHeapScope,Address ) )
                            {
                            //
                            //  Zero class members.
                            //
                            Address = NULL;
                            Active = False;
                            }
                        else
                            { FAILURE( "Unable to delete the root" ); }
                        }
                    else
                        { FAILURE( "Unable to delete the root structure" ); }
                    }
                else
                    { FAILURE( "Unable to update the root structure" ); }
                }
            else
                { FAILURE( "Unable to update the root structure" ); }
            }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define STACK ROCKALL::STACK
#endif
#endif