#ifndef _BASE_TYPES_HPP_
#define _BASE_TYPES_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "CompilerFlags.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Standard Base Types.                                                                              */
        /*                                                                                                    */
        /*  The standard base types ensure consistency in the presence of different compilers, hardware       */
        /*  platforms and operating systems.                                                                  */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Standard Base Types.                                                                      */
            /*                                                                                            */
            /*  The standard base types are merely redefinitions of existing compiler supported types.    */
            /*                                                                                            */
            /**********************************************************************************************/

    typedef bool                                BOOLEAN;
    typedef char                                CHAR;
    typedef wchar_t                             WCHAR;
    typedef short                               SHORT;
    typedef int                                 INT;
    typedef long                                LONG;

    typedef signed char                         SCHAR;
    typedef signed short                        SWCHAR;
    typedef signed short                        SSHORT;
    typedef signed int                          SINT;
    typedef signed long                         SLONG;

    typedef unsigned char                       UCHAR;
    typedef unsigned short                      UWCHAR;
    typedef unsigned short                      USHORT;
    typedef unsigned int                        UINT;
    typedef unsigned long                       ULONG;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Fixed Size Types.                                                                         */
            /*                                                                                            */
            /*  The fixed size types provide a specific precision regardless of the underlying platform.  */
            /*                                                                                            */
            /**********************************************************************************************/

    #ifdef WINDOWS
    typedef signed __int8                       SBIT8;
    typedef signed __int16                      SBIT16;
    typedef signed __int32                      SBIT32;
    typedef signed __int32 __w64                SSBIT32;
    typedef signed __int64                      SBIT64;

    typedef unsigned __int8                     UBIT8;
    typedef unsigned __int16                    UBIT16;
    typedef unsigned __int32                    UBIT32;
    typedef unsigned __int64                    UBIT64;
    #endif

            /**********************************************************************************************/
            /*                                                                                            */
            /*  The Native Types.                                                                         */
            /*                                                                                            */
            /*  The native types provide a natural precision given the underlying hardware platform.      */
            /*                                                                                            */
            /**********************************************************************************************/

    #ifdef WINDOWS
    #ifdef WINDOWS32
    typedef signed __int32 __w64                SNATIVE;
    typedef unsigned __int32 __w64              UNATIVE;
    #endif

    #ifdef WINDOWS64
    typedef signed __int64 __w64                SNATIVE;
    typedef unsigned __int64 __w64              UNATIVE;
    #endif
    #endif

            /**********************************************************************************************/
            /*                                                                                            */
            /*  The File Types.                                                                           */
            /*                                                                                            */
            /*  The file types provide the precision for any associated file system.                      */
            /*                                                                                            */
            /**********************************************************************************************/

    #ifdef WINDOWS
    typedef SBIT64                              FILE_ADDRESS;
    typedef FILE_ADDRESS                        FILE_SIZE;
    #endif
    }
#endif