#ifndef _COMPILER_FLAGS_HPP_
#define _COMPILER_FLAGS_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "ReservedWords.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Standard Compiler Flags.                                                                          */
        /*                                                                                                    */
        /*  All compilers tend to have a collection of flags they set when code is compiled.  We              */
        /*  standardize all the compiler flags here into a standard set used throughout the code.             */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Standard Compiler Flags.                                                                  */
            /*                                                                                            */
            /*  All specific compiler flags are converted into a standard set of compiler flags used      */
            /*  throughout the code.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    #ifdef _WIN32
    #ifndef _M_X64
    #define CACHE_SIZE                          64
    #define WINDOWS                             1
    #define WINDOWS32                           1

    #ifdef _M_IX86
    #define PLATFORM_32BIT                      1
    #endif
    #endif
    #endif

    #ifdef _WIN32
    #ifdef _M_X64
    #define CACHE_SIZE                          64
    #define WINDOWS                             1
    #define WINDOWS64                           1

    #ifdef _M_X64
    #define PLATFORM_64BIT                      1
    #endif
    #endif
    #endif

    #ifdef WINDOWS
    #define CACHE_LINE                          __declspec(align(CACHE_SIZE))
    #define CACHE_MASK                          (CACHE_SIZE-1)

    #ifdef _DEBUG
    #define DEBUGGING                           1
    #endif

    #define INLINE                              __forceinline

    #ifndef NULL
    #define NULL                                0
    #endif
    #endif

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Special Linkage Macros.                                                                   */
            /*                                                                                            */
            /*  We need to select the correct linkage based on the current compiling environment.         */
            /*                                                                                            */
            /**********************************************************************************************/

    #ifdef WINDOWS
    #ifdef COMPILING_ROCKALL_DB_DLL
    #define ROCKALL_DB_LINKAGE __declspec(dllexport)
    #else
    #define ROCKALL_DB_LINKAGE __declspec(dllimport)
    #endif
    #endif

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Disable Selected Compiler Warnings.                                                       */
            /*                                                                                            */
            /*  Disable any unhelpful compiler warnings.                                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    #ifdef WINDOWS
    #pragma warning(disable : 4073)
    #pragma warning(disable : 4355)
    #pragma warning(disable : 4996)

    #define TO_DO( Message ) message ( "TO DO -------->>>>>>>>> " Message )
    #endif
    }
#endif