#ifndef _ERROR_HPP_
#define _ERROR_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "ReservedWords.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Global Errors.                                                                                    */
        /*                                                                                                    */
        /*  The global errors class is used to report errors.                                                 */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
    typedef enum
        {
        //
        //
        //  Global errors.
        //
        //  All the error codes appear within the following enumeration.  We group the errors codes
        //  into sections that relate to the part of the system in which the error typically occurs.  A
        //  small number of error codes appear in multiple areas and so appear here as part of the
        //  global error codes.
        //
        NoError                                 =    NULL,          // No error code is available.
        ErrorAddressNotInUse                    =    1000,          // The Address is not in use.
        ErrorDemonstrationLimit                 =    1001,          // This version of the software is limited.
        ErrorExceptionCaught                    =    1002,          // An exception was caught during processing.
        ErrorFunctionalityUnavailable           =    1003,          // The functionality is not available.
        ErrorHeapCorrupt                        =    1004,          // The heap has been damaged.
        ErrorIncorrectParameter                 =    1005,          // A parameter to the call is invalid.
        ErrorNoActiveHeap                       =    1006,          // There is no active heap available.
        ErrorRequestDeclined                    =    1007,          // The request was declined.
        ErrorUnsupported                        =    1008,          // The call in this context is unsupported.

        //
        //  Functionality: Operating System.
        //
        ErrorAdjustingFileSize                  =    2000,          // A file size adjustment was refused.
        ErrorFileCreateOrOpen                   =    2001,          // A file create or open request was refused.
        ErrorFileInformation                    =    2002,          // A file information request was refused.
        ErrorFileTransfer                       =    2003,          // A file transfer was refused.
        ErrorNoAvailableMemory                  =    2004,          // There is no memory space available.
        ErrorPostFailure                        =    2005,          // A completion port post was refused.
        ErrorThreadLocalStore                   =    2006,          // A thread local store request was refused.

        //
        //  Functionality: Transactional Heaps.
        //
        ErrorActiveRegion                       =    3000,          // The region is still in use.
        ErrorActiveTransaction                  =    3001,          // There is already an active transaction.
        ErrorCreatingFile                       =    3002,          // We were unable to create the file.
        ErrorDeadlock                           =    3003,          // The lock was declined to avoid a deadlock.
        ErrorFileAlreadyActive                  =    3004,          // There is already an active file.
        ErrorLockingRootPage                    =    3005,          // We were unable to lock the root page.
        ErrorNoActiveTransaction                =    3006,          // We were unable to find an active transaction.
        ErrorNoActiveRegion                     =    3007,          // We were unable to find an active region.
        ErrorNoCreateOrOpenFile                 =    3008,          // We were unable to find an active file.
        ErrorNoExistingLock                     =    3009,          // We were unable to find an active lock.
        ErrorNoSpaceForRootPage                 =    3010,          // We were unable to get space for the root page.
        ErrorOpeningFile                        =    3011,          // We were unable to open the file.
        ErrorReadingSystemValue                 =    3012,          // We were unable to read a system value.
        ErrorSettingSystemValue                 =    3013,          // We were unable to set a system value.
        ErrorTransactionTooBig                  =    3014,          // A transaction has reached its memory limit.
        ErrorUnableToLockRegion                 =    3015,          // We were unable to recover the file.
        ErrorUnableToLockRoot                   =    3016,          // We were unable to recover the file.
        ErrorUnableToRecoverFile                =    3017,          // We were unable to recover the file.
        ErrorUpdateToReadOnlyFile               =    3018,          // We were unable to recover the file.

        //
        //  Functionality: Rockall-DB Library.
        //
        ErrorDuplicateKey                       =    9000,          // A duplicate was found.
        ErrorExistingTransaction                =    9001,          // There is already an active transaction.
        ErrorKeyNotFound                        =    9002,          // The key was not found.
        ErrorNoTransaction                      =    9003,          // There was no active transaction.
        ErrorNoValueAvailable                   =    9004,          // We were unable to return a value.
        ErrorStringConversion                   =    9005,          // The string has a conversion error.
        ErrorStringOverflow                     =    9006,          // The string has overflowed.
        ErrorUnableToStoreValue                 =    9007,          // We were unable store the value.
        ErrorUnableToViewRoot                   =    9008,          // We were unable to view the root.
        ErrorValueNotFound                      =    9009,          // The value was not found.

        //
        //  The last error code.
        //
        LastErrorCode
        }
    GLOBAL_ERROR_NUMBER;

    BOOLEAN Error( GLOBAL_ERROR_NUMBER ErrorNumber );

    BOOLEAN LastError( GLOBAL_ERROR_NUMBER *ErrorNumber = NULL );

    BOOLEAN ResetError( GLOBAL_ERROR_NUMBER ErrorNumber = NoError );
    }
#endif