#ifndef _EXPORT_ASYNCHRONOUS_CALLBACK_HPP_
#define _EXPORT_ASYNCHRONOUS_CALLBACK_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "PublicConstants.hpp"
#include "VirtualAsynchronousCallback.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Export Asynchronous Callback.                                                                     */
        /*                                                                                                    */
        /*  An export asynchronous callback is an externally visible asynchronous callback mechanism.         */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Export Asynchronous Callback Linkage.                                                     */
            /*                                                                                            */
            /*  The export asynchronous callback linkage functions are responsible for managing the       */
            /*  asynchronous callback mechanism.                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    extern "C"
        {
        ROCKALL_DB_LINKAGE VOID CreateAsynchronousCallback( VIRTUAL_ASYNCHRONOUS_CALLBACK **AsynchronousCallback );

        ROCKALL_DB_LINKAGE VOID DeleteAsynchronousCallback( VIRTUAL_ASYNCHRONOUS_CALLBACK **AsynchronousCallback );
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Export Asynchronous Callback.                                                             */
            /*                                                                                            */
            /*  An export asynchronous callback is an externally visible asynchronous callback            */
            /*  mechanism.                                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    class EXPORT_ASYNCHRONOUS_CALLBACK : public VIRTUAL_ASYNCHRONOUS_CALLBACK
        {
            //
            //  Private data.
            //
            VIRTUAL_ASYNCHRONOUS_CALLBACK       *AsynchronousCallback;

        public:
            //
            //  Public virtual functions.
            //
            EXPORT_ASYNCHRONOUS_CALLBACK( VOID )
                { CreateAsynchronousCallback( & AsynchronousCallback ); }

            VIRTUAL SBIT32 NumberOfCPUs( VOID )
                { return ((AsynchronousCallback != NULL) ? AsynchronousCallback -> NumberOfCPUs() : 1); }

            VIRTUAL BOOLEAN Schedule( VOID *Function,VOID *Type,SBIT32 Size )
                { return ((AsynchronousCallback != NULL) && (AsynchronousCallback -> Schedule( Function,Type,Size ))); }

            VIRTUAL VOID Sleep( SBIT32 Time )
                { if ( AsynchronousCallback != NULL ) { AsynchronousCallback -> Sleep( Time ); } }

            VIRTUAL BOOLEAN WaitAll( VOID )
                { return ((AsynchronousCallback != NULL) && (AsynchronousCallback -> WaitAll())); }

            VIRTUAL ~EXPORT_ASYNCHRONOUS_CALLBACK( VOID )
                { DeleteAsynchronousCallback( & AsynchronousCallback ); }

        private:
            //
            //  Disabled functionality.
            //
            EXPORT_ASYNCHRONOUS_CALLBACK( CONST EXPORT_ASYNCHRONOUS_CALLBACK & NoCopies );

            VOID operator=( CONST EXPORT_ASYNCHRONOUS_CALLBACK & NoCopies );
        };
    }
#endif