#ifndef _EXPORT_HEAP_HPP_
#define _EXPORT_HEAP_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Export Heap.                                                                                      */
        /*                                                                                                    */
        /*  The export heap interface is the standard interface for the Rockall-DB suite of heaps             */
        /*  and databases.                                                                                    */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Export Heap.                                                                              */
            /*                                                                                            */
            /*  The export heap interface provides prototypes for all the publically available            */
            /*  functionality within 'RockallDB.dll'.                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    class EXPORT_HEAP : public VIRTUAL_HEAP
        {
        protected:
            //
            //  Protected data.
            //
            BOOLEAN                             Active;
            VIRTUAL_HEAP                        *Heap;
            VOID                                *UserValue;

        public:
            //
            //  The Rockall-DB core functionality.
            //
            EXPORT_HEAP( VOID )
                { Active = False; Heap = NULL; UserValue = NULL; }

            VIRTUAL BOOLEAN Delete( FILE_ADDRESS Address,BOOLEAN NoDestroy = False,BOOLEAN Zero = False )
                { return ((Active) && (Heap -> Delete( Address,NoDestroy,Zero ))); }

            VIRTUAL BOOLEAN Delete( VOID *Data,BOOLEAN NoDestroy = False,BOOLEAN Zero = False )
                { return ((Active) && (Heap -> Delete( Data,NoDestroy,Zero ))); }

            VIRTUAL BOOLEAN DeleteAll( BOOLEAN NoDestroy = False,BOOLEAN Zero = False )
                { return ((Active) && (Heap -> DeleteAll( NoDestroy,Zero ))); }

            VIRTUAL BOOLEAN Details( FILE_ADDRESS Address,BOOLEAN *Destroy = NULL,FILE_SIZE *Space = NULL )
                { return ((Active) && (Heap -> Details( Address,Destroy,Space ))); }

            VIRTUAL BOOLEAN Details( VOID *Data,BOOLEAN *Destroy = NULL,FILE_SIZE *Space = NULL )
                { return ((Active) && (Heap -> Details( Data,Destroy,Space ))); }

            VIRTUAL SBIT32 Feature( VOID )
                { return ((Active) ? (Heap -> Feature()) : 0); }

            VIRTUAL BOOLEAN GetError( SBIT32 *ErrorNumber )
                { return ((Active) && (Heap -> GetError( ErrorNumber ))); }

            VIRTUAL BOOLEAN New
                    (
                    FILE_ADDRESS                *Address,
                    VOID                        **Data,
                    FILE_SIZE                   Size,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return ((Active) && (Heap -> New( Address,Data,Size,Space,Destroy,Zero ))); }

            VIRTUAL BOOLEAN New
                    (
                    VOID                        **Data,
                    FILE_SIZE                   Size,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return ((Active) && (Heap -> New( NULL,Data,Size,Space,Destroy,Zero ))); }

            VIRTUAL BOOLEAN Resize
                    (
                    FILE_ADDRESS                *Address,
                    VOID                        **Data,
                    FILE_SIZE                   NewSize,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Zero = False
                    )
                { return ((Active) && (Heap -> Resize( Address,Data,NewSize,Space,Zero ))); }

            VIRTUAL BOOLEAN Resize
                    (
                    VOID                        **Data,
                    FILE_SIZE                   NewSize,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Zero = False
                    )
                { return ((Active) && (Heap -> Resize( NULL,Data,NewSize,Space,Zero ))); }

            VIRTUAL BOOLEAN SetError( SBIT32 ErrorNumber )
                { return ((Active) && (Heap -> SetError( ErrorNumber ))); }

            VIRTUAL BOOLEAN Size( FILE_SIZE *Space )
                { return ((Active) && (Heap -> Size( Space ))); }

            VIRTUAL BOOLEAN Walk( HEAP_WALK_FUNCTION WalkFunction,VOID *UserValue = NULL )
                { return ((Active) && (Heap -> Walk( WalkFunction,UserValue ))); }

            //
            //  The Rockall-DB transactional functionality.
            //
            VIRTUAL BOOLEAN CreateFile( CHAR *FileName,ENCRYPTION_FUNCTION Function = NULL,VOID *UserValue = NULL )
                { return ((Active) && (Heap -> CreateFile( FileName,Function,UserValue ))); }

            VIRTUAL BOOLEAN CreateFile( WCHAR *FileName,ENCRYPTION_FUNCTION Function = NULL,VOID *UserValue = NULL )
                { return ((Active) && (Heap -> CreateFile( FileName,Function,UserValue ))); }

            VIRTUAL BOOLEAN CreateRegion( SBIT32 *NewRegionID )
                { return ((Active) && (Heap -> CreateRegion( NewRegionID ))); }

            VIRTUAL BOOLEAN BeginTransaction
                    (
                    FILE_ADDRESS                *NewTransactionID = NULL,
                    EXPORT_MEMORY_LOCK          *UserLock = NULL
                    )
                { return ((Active) && (Heap -> BeginTransaction( NewTransactionID,UserLock ))); }

            VIRTUAL BOOLEAN ChangeRegion( SBIT32 NewRegionID,SBIT32 *OldRegionID = NULL )
                { return ((Active) && (Heap -> ChangeRegion( NewRegionID,OldRegionID ))); }

            VIRTUAL BOOLEAN ExclusiveView( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL )
                { return ((Active) && (Heap -> ExclusiveView( Address,Data,Space ))); }

            VIRTUAL BOOLEAN EndExclusiveView( FILE_ADDRESS Address,BOOLEAN Reset = True )
                { return ((Active) && (Heap -> EndExclusiveView( Address,Reset ))); }

            VIRTUAL BOOLEAN JoinTransaction( FILE_ADDRESS TransactionID,EXPORT_MEMORY_LOCK *UserLock = NULL )
                { return ((Active) && (Heap -> JoinTransaction( TransactionID,UserLock ))); }

            VIRTUAL BOOLEAN LeaveTransaction( VOID )
                { return ((Active) && (Heap -> LeaveTransaction())); }

            VIRTUAL BOOLEAN Update( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL )
                { return ((Active) && (Heap -> Update( Address,Data,Space ))); }

            VIRTUAL BOOLEAN View( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL )
                { return ((Active) && (Heap -> View( Address,Data,Space ))); }

            VIRTUAL BOOLEAN EndView( FILE_ADDRESS Address,BOOLEAN Reset = True )
                { return ((Active) && (Heap -> EndView( Address,Reset ))); }

            VIRTUAL BOOLEAN EndTransaction( BOOLEAN Abort = False,BOOLEAN Commit = False )
                { return ((Active) && (Heap -> EndTransaction( Abort,Commit ))); }

            VIRTUAL BOOLEAN DeleteRegion( SBIT32 RegionID )
                { return ((Active) && (Heap -> DeleteRegion( RegionID ))); }

            VIRTUAL BOOLEAN CloseFile( VOID )
                { return ((Active) && (Heap -> CloseFile())); }

            //
            //  The Rockall-DB database functionality.
            //
            VIRTUAL BOOLEAN CreateShadowFile( CHAR *FileName,SBIT32 *ID = NULL,BOOLEAN Wait = True )
                { return ((Active) && (Heap -> CreateShadowFile( FileName,ID,Wait ))); }

            VIRTUAL BOOLEAN CreateShadowFile( WCHAR *FileName,SBIT32 *ID = NULL,BOOLEAN Wait = True )
                { return ((Active) && (Heap -> CreateShadowFile( FileName,ID,Wait ))); }

            VIRTUAL BOOLEAN CreateSnapshotFile( WCHAR *FileName )
                { return ((Active) && (Heap -> CreateSnapshotFile( FileName ))); }

            VIRTUAL BOOLEAN CreateSnapshotFile( CHAR *FileName )
                { return ((Active) && (Heap -> CreateSnapshotFile( FileName ))); }

            VIRTUAL BOOLEAN OpenFile
                    (
                    CHAR                        *FileName,
                    BOOLEAN                     ReadOnly = False,
                    ENCRYPTION_FUNCTION         Function = NULL,
                    VOID                        *UserValue = NULL
                    )
                { return ((Active) && (Heap -> OpenFile( FileName,ReadOnly,Function,UserValue ))); }

            VIRTUAL BOOLEAN OpenFile
                    (
                    WCHAR                       *FileName,
                    BOOLEAN                     ReadOnly = False,
                    ENCRYPTION_FUNCTION         Function = NULL,
                    VOID                        *UserValue = NULL
                    )
                { return ((Active) && (Heap -> OpenFile( FileName,ReadOnly,Function,UserValue ))); }

            VIRTUAL BOOLEAN Commit( VOID )
                { return ((Active) && (Heap -> Commit())); }

            VIRTUAL BOOLEAN FileSize( FILE_SIZE *Size )
                { return ((Active) && (Heap -> FileSize( Size ))); }

            VIRTUAL BOOLEAN GetUserValue( SBIT32 ID,FILE_ADDRESS *Value )
                { return ((Active) && (Heap -> GetUserValue( ID,Value ))); }

            VIRTUAL BOOLEAN SetUserValue( SBIT32 ID,FILE_ADDRESS Value )
                { return ((Active) && (Heap -> SetUserValue( ID,Value ))); }

            VIRTUAL BOOLEAN Touch( FILE_ADDRESS Address )
                { return ((Active) && (Heap -> Touch( Address ))); }

            VIRTUAL BOOLEAN CloseShadowFile( SBIT32 ID = 0 )
                { return ((Active) && (Heap -> CloseShadowFile( ID ))); }

            VIRTUAL ~EXPORT_HEAP( VOID )
                { Active = False; }

        private:
            //
            //  Disabled functionality.
            //
            EXPORT_HEAP( CONST EXPORT_HEAP & NoCopies );

            VOID operator=( CONST EXPORT_HEAP & NoCopies );
        };
    }
#endif