#ifndef _EXPORT_MEMORY_LOCK_HPP_
#define _EXPORT_MEMORY_LOCK_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "PublicConstants.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Export Memory Lock.                                                                               */
        /*                                                                                                    */
        /*  An export memory lock is an externally visible shareable lock.                                    */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Export Lock Linkage.                                                                      */
            /*                                                                                            */
            /*  The export lock linkage functions are responsible for managing external locks.            */
            /*                                                                                            */
            /**********************************************************************************************/

    extern "C"
        {
        ROCKALL_DB_LINKAGE VOID CreateMemoryLock( BOOLEAN *Active,BOOLEAN *Delete,VOID **MemoryLock );

        ROCKALL_DB_LINKAGE BOOLEAN ClaimMemoryLock( BOOLEAN Active,VOID *MemoryLock,SBIT16 Type,BOOLEAN Wait );

        ROCKALL_DB_LINKAGE VOID RegisterMemoryLock( BOOLEAN *Active,BOOLEAN *Delete,VOID *NewLock,VOID **MemoryLock );

        ROCKALL_DB_LINKAGE VOID ReleaseMemoryLock( BOOLEAN Active,VOID *MemoryLock,SBIT16 Type );

        ROCKALL_DB_LINKAGE SBIT32 TotalMemoryLocks( VOID );

        VOID UpdateMemoryLock( BOOLEAN *Active,BOOLEAN *Delete,VOID *NewLock,VOID **MemoryLock );

        ROCKALL_DB_LINKAGE VOID DeleteMemoryLock( BOOLEAN *Active,BOOLEAN *Delete,VOID **MemoryLock );
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Export Memory Lock.                                                                       */
            /*                                                                                            */
            /*  An export memory lock is an externally visible shareable lock.                            */
            /*                                                                                            */
            /**********************************************************************************************/

    class EXPORT_MEMORY_LOCK
        {
            //
            //  Private data.
            //
            BOOLEAN                             Active;
            BOOLEAN                             Delete;
            VOID                                *MemoryLock;

        public:
            //
            //  Public inline functions.
            //
            EXPORT_MEMORY_LOCK( VOID )
                { CreateMemoryLock( & Active,& Delete,& MemoryLock ); }

            EXPORT_MEMORY_LOCK( VOID *NewLock )
                { Active = (NewLock != NULL); Delete = False; MemoryLock = NewLock; }

            INLINE BOOLEAN ClaimExclusiveLock( BOOLEAN Wait = True )
                { return (ClaimMemoryLock( Active,MemoryLock,0,Wait )); }

            INLINE BOOLEAN ClaimSharedLock( BOOLEAN Wait = True )
                { return (ClaimMemoryLock( Active,MemoryLock,1,Wait )); }

            INLINE VOID ReleaseExclusiveLock( VOID )
                { ReleaseMemoryLock( Active,MemoryLock,0 ); }

            INLINE VOID ReleaseSharedLock( VOID )
                { ReleaseMemoryLock( Active,MemoryLock,1 ); }

            ~EXPORT_MEMORY_LOCK( VOID )
                { DeleteMemoryLock( & Active,& Delete,& MemoryLock ); }

            //
            //  Public inline functions.
            //
            INLINE VOID *GetLock( VOID )
                { return ((Active) ? MemoryLock : NULL); }

            INLINE VOID RegisterLock( VOID *NewLock )
                { RegisterMemoryLock( & Active,& Delete,NewLock,& MemoryLock ); }

            INLINE VOID UpdateLock( VOID *NewLock )
                { UpdateMemoryLock( & Active,& Delete,NewLock,& MemoryLock ); }

        private:
            //
            //  Disabled functionality.
            //
            EXPORT_MEMORY_LOCK( CONST EXPORT_MEMORY_LOCK & NoCopies );

            VOID operator=( CONST EXPORT_MEMORY_LOCK & NoCopies );
        };
    }
#endif