#ifndef _EXPORT_THREAD_LOCAL_STORE_HPP_
#define _EXPORT_THREAD_LOCAL_STORE_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "PublicConstants.hpp"
#include "VirtualThreadLocalStore.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Export Thread Local Store.                                                                        */
        /*                                                                                                    */
        /*  An export thread local store is an externally visible thread local store value.                   */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Export Thread Local Store Linkage.                                                        */
            /*                                                                                            */
            /*  The export thread local store linkage functions are responsible for managing the          */
            /*  thread local store value.                                                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    extern "C"
        {
        ROCKALL_DB_LINKAGE VOID CreateThreadLocalStore( VIRTUAL_THREAD_LOCAL_STORE **ThreadLocalStore );

        ROCKALL_DB_LINKAGE VOID DeleteThreadLocalStore( VIRTUAL_THREAD_LOCAL_STORE **ThreadLocalStore );
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Export Thread Local Store.                                                                */
            /*                                                                                            */
            /*  An export thread local store is an externally visible thread local store value.           */
            /*                                                                                            */
            /**********************************************************************************************/

    class EXPORT_THREAD_LOCAL_STORE : public VIRTUAL_THREAD_LOCAL_STORE
        {
            //
            //  Private data.
            //
            VIRTUAL_THREAD_LOCAL_STORE          *ThreadLocalStore;

        public:
            //
            //  Public virtual functions.
            //
            EXPORT_THREAD_LOCAL_STORE( VOID )
                { CreateThreadLocalStore( & ThreadLocalStore ); }

            VIRTUAL VOID *Get( VOID )
                { return ((ThreadLocalStore != NULL) ? (ThreadLocalStore -> Get()) : NULL); }

            VIRTUAL BOOLEAN Set( VOID *Value )
                { return ((ThreadLocalStore != NULL) && (ThreadLocalStore -> Set( Value ))); }

            VIRTUAL ~EXPORT_THREAD_LOCAL_STORE( VOID )
                { DeleteThreadLocalStore( & ThreadLocalStore ); }

        private:
            //
            //  Disabled functionality.
            //
            EXPORT_THREAD_LOCAL_STORE( CONST EXPORT_THREAD_LOCAL_STORE & NoCopies );

            VOID operator=( CONST EXPORT_THREAD_LOCAL_STORE & NoCopies );
        };
    }
#endif