#ifndef _PUBLIC_CONSTANTS_HPP_
#define _PUBLIC_CONSTANTS_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "CompilerFlags.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Public Constants.                                                                                 */
        /*                                                                                                    */
        /*  It is better to place constants close to where they are used.  Nonetheless, a small number        */
        /*  of constants are global and so need to be gathered together in a common header.                   */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Public Constants.                                                                         */
            /*                                                                                            */
            /*  The public constants are available throughout the code.  Care should be taken to          */
            /*  only put truly global constants in this file.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    STATIC CONST SBIT32 ByteBits                = 8;

    STATIC CONST SBIT32 Bit8Bits                = (sizeof(SBIT8) * ByteBits);
    STATIC CONST SBIT32 Bit16Bits               = (sizeof(SBIT16) * ByteBits);
    STATIC CONST SBIT32 Bit32Bits               = (sizeof(SBIT32) * ByteBits);
    STATIC CONST SBIT32 Bit64Bits               = (sizeof(SBIT64) * ByteBits);
    STATIC CONST SBIT32 NativeBits              = (sizeof(SNATIVE) * ByteBits);

    STATIC CONST SBIT32 Bit8Bytes               = sizeof(SBIT8);
    STATIC CONST SBIT32 Bit16Bytes              = sizeof(SBIT16);
    STATIC CONST SBIT32 Bit32Bytes              = sizeof(SBIT32);
    STATIC CONST SBIT32 Bit64Bytes              = sizeof(SBIT64);
    STATIC CONST SBIT32 NativeBytes             = sizeof(SNATIVE);

    STATIC CONST SBIT32 Bit8Mask                = (sizeof(SBIT8)-1);
    STATIC CONST SBIT32 Bit16Mask               = (sizeof(SBIT16)-1);
    STATIC CONST SBIT32 Bit32Mask               = (sizeof(SBIT32)-1);
    STATIC CONST SBIT32 Bit64Mask               = (sizeof(SBIT64)-1);
    STATIC CONST SBIT32 NativeMask              = (sizeof(SNATIVE)-1);

    STATIC CONST SBIT64 ByteSex                 = 0x0001020304050607;

    STATIC CONST BOOLEAN False                  = false;
    STATIC CONST BOOLEAN True                   = true;

    STATIC CONST SBIT32 FastGrowthFactor        = 8;
    STATIC CONST SBIT32 GrowthFactor            = 2;
    STATIC CONST SBIT32 ShrinkFactor            = 8;

    STATIC CONST SBIT32 Guard00ForSBit32        = 0xDeadBe0f;
    STATIC CONST SBIT32 Guard01ForSBit32        = 0xDeadBe1f;
    STATIC CONST SBIT32 Guard02ForSBit32        = 0xDeadBe2f;
    STATIC CONST SBIT32 Guard03ForSBit32        = 0xDeadBe3f;
    STATIC CONST SBIT32 Guard04ForSBit32        = 0xDeadBe4f;
    STATIC CONST SBIT32 Guard05ForSBit32        = 0xDeadBe5f;
    STATIC CONST SBIT32 Guard06ForSBit32        = 0xDeadBe7f;
    STATIC CONST SBIT32 Guard07ForSBit32        = 0xDeadBe7f;
    STATIC CONST SBIT32 Guard08ForSBit32        = 0xDeadBe8f;
    STATIC CONST SBIT32 Guard09ForSBit32        = 0xDeadBe9f;

    STATIC CONST SBIT64 Guard00ForSBit64        = 0xDeadBe0fDeadBe0f;
    STATIC CONST SBIT64 Guard01ForSBit64        = 0xDeadBe1fDeadBe1f;
    STATIC CONST SBIT64 Guard02ForSBit64        = 0xDeadBe2fDeadBe2f;
    STATIC CONST SBIT64 Guard03ForSBit64        = 0xDeadBe3fDeadBe3f;
    STATIC CONST SBIT64 Guard04ForSBit64        = 0xDeadBe4fDeadBe4f;
    STATIC CONST SBIT64 Guard05ForSBit64        = 0xDeadBe5fDeadBe5f;
    STATIC CONST SBIT64 Guard06ForSBit64        = 0xDeadBe7fDeadBe7f;
    STATIC CONST SBIT64 Guard07ForSBit64        = 0xDeadBe7fDeadBe7f;
    STATIC CONST SBIT64 Guard08ForSBit64        = 0xDeadBe8fDeadBe8f;
    STATIC CONST SBIT64 Guard09ForSBit64        = 0xDeadBe9fDeadBe9f;

    STATIC CONST SBIT32 HashMaskSBit32          = 0x3fffffff;
    STATIC CONST SBIT64 HashMaskSBit64          = 0x3fffffffffffffff;

    STATIC CONST SBIT32 Kilobyte                = 1024;
    STATIC CONST SBIT32 Megabyte                = (Kilobyte * Kilobyte);
    STATIC CONST SBIT32 Gigabyte                = (Kilobyte * Megabyte);

    STATIC CONST SBIT32 MaxSBit8                = 0x7f;
    STATIC CONST SBIT64 MaxSBit16               = 0x7fff;
    STATIC CONST SBIT32 MaxSBit32               = 0x7fffffff;
    STATIC CONST SBIT64 MaxSBit64               = 0x7fffffffffffffff;

    STATIC CONST SBIT32 MinSBit8                = 0x80;
    STATIC CONST SBIT64 MinSBit16               = 0x8000;
    STATIC CONST SBIT32 MinSBit32               = 0x80000000;
    STATIC CONST SBIT64 MinSBit64               = 0x8000000000000000;

    STATIC CONST SBIT32 MaxThreads              = 1024;

    STATIC CONST SBIT8 NoValue                  = -1;

    STATIC CONST SBIT32 PrimeSBit32             = 49157;
    STATIC CONST SBIT64 PrimeSBit64             = 1610612741;

    #ifdef PLATFORM_32BIT
    STATIC CONST SBIT32 Guard00ForSNative       = Guard00ForSBit32;
    STATIC CONST SBIT32 Guard01ForSNative       = Guard01ForSBit32;
    STATIC CONST SBIT32 Guard02ForSNative       = Guard02ForSBit32;
    STATIC CONST SBIT32 Guard03ForSNative       = Guard03ForSBit32;
    STATIC CONST SBIT32 Guard04ForSNative       = Guard04ForSBit32;
    STATIC CONST SBIT32 Guard05ForSNative       = Guard05ForSBit32;
    STATIC CONST SBIT32 Guard06ForSNative       = Guard06ForSBit32;
    STATIC CONST SBIT32 Guard07ForSNative       = Guard07ForSBit32;
    STATIC CONST SBIT32 Guard08ForSNative       = Guard08ForSBit32;
    STATIC CONST SBIT32 Guard09ForSNative       = Guard09ForSBit32;

    STATIC CONST SBIT32 HashMaskSNative         = HashMaskSBit32;
    STATIC CONST SBIT32 MaxSNative              = MaxSBit32;
    STATIC CONST SBIT32 MaxString               = (1 << 30);
    STATIC CONST SBIT32 MinSNative              = MinSBit32;
    STATIC CONST SBIT32 PrimeSNative            = PrimeSBit32;
    #endif
    #ifdef PLATFORM_64BIT
    STATIC CONST SBIT64 Guard00ForSNative       = Guard00ForSBit64;
    STATIC CONST SBIT64 Guard01ForSNative       = Guard01ForSBit64;
    STATIC CONST SBIT64 Guard02ForSNative       = Guard02ForSBit64;
    STATIC CONST SBIT64 Guard03ForSNative       = Guard03ForSBit64;
    STATIC CONST SBIT64 Guard04ForSNative       = Guard04ForSBit64;
    STATIC CONST SBIT64 Guard05ForSNative       = Guard05ForSBit64;
    STATIC CONST SBIT64 Guard06ForSNative       = Guard06ForSBit64;
    STATIC CONST SBIT64 Guard07ForSNative       = Guard07ForSBit64;
    STATIC CONST SBIT64 Guard08ForSNative       = Guard08ForSBit64;
    STATIC CONST SBIT64 Guard09ForSNative       = Guard09ForSBit64;

    STATIC CONST SBIT64 HashMaskSNative         = HashMaskSBit64;
    STATIC CONST SBIT64 MaxSNative              = MaxSBit64;
    STATIC CONST SBIT64 MaxString               = 0x0000010000000000;
    STATIC CONST SBIT64 MinSNative              = MinSBit64;
    STATIC CONST SBIT64 PrimeSNative            = PrimeSBit64;
    #endif
    }
#endif