#ifndef _SUB_SET_HPP_
#define _SUB_SET_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "Failure.hpp"
#include "MemoryFunctions.hpp"
#include "PlacementNew.hpp"
#include "ReservedWords.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Sub Set.                                                                                          */
        /*                                                                                                    */
        /*  A sub-set is a portion of a set and is essentially a cache.  We use sub-sets to save us           */
        /*  from having to update large data structures repeatedly.  When a sub-set becomes full we           */
        /*  update the main data structure.                                                                   */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Sub Set.                                                                                  */
            /*                                                                                            */
            /*  A sub-set is a portion of a set and is essentially a cache.  We use sub-sets to save      */
            /*  us from having to update large data structures repeatedly.  When a sub-set becomes        */
            /*  full we update the main data structure.                                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> class SUB_SET
        {
            //
            //  Private data.
            //
            BOOLEAN                             Active;
            VALUE                               SubSet[ MAX_SIZE ];
            SBIT32                              Used;

            //
            //  Static private data.
            //
            STATIC MEMORY_FUNCTIONS<SNATIVE>    MemoryFunctions;

        public:
            //
            //  Public functions.
            //
            SUB_SET( VOID );

            BOOLEAN DeleteAll( VOID );

            BOOLEAN DeleteValue( CONST VALUE & Value );

            BOOLEAN FindValue( CONST VALUE & Value );

            BOOLEAN PopValue( VALUE *Value );

            BOOLEAN PushValue( CONST VALUE & Value );

            BOOLEAN ResultSet( VALUE **Set,SBIT32 *Size );

            ~SUB_SET( VOID );

            //
            //  Public inline functions.
            //
            INLINE BOOLEAN Full( VOID )
                { return (Used >= MAX_SIZE); }

            INLINE BOOLEAN NewValue( CONST VALUE & Value )
                { return (PushValue( Value )); }

            INLINE SBIT32 Size( VOID )
                { return Used; }

        private:
            //
            //  Private functions.
            //
            BOOLEAN FindValue( VALUE *Value,SBIT32 *Offset );

            VOID Zero( VALUE *Set,SBIT32 Size );

            //
            //  Disabled functionality.
            //
            SUB_SET( CONST SUB_SET & NoCopies );

            VOID operator=( CONST SUB_SET & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Class Static Data.                                                                        */
            /*                                                                                            */
            /*  The class static data needs to be initialised before it's ready for use.                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> MEMORY_FUNCTIONS<SNATIVE> SUB_SET<MAX_SIZE,VALUE>::MemoryFunctions;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'SUB_SET'.                                                                */
            /*                                                                                            */
            /*  The constructor for 'SUB_SET' prepares the class for use.                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> SUB_SET<MAX_SIZE,VALUE>::SUB_SET( VOID )
        {
        //
        //  We make sure that 'MAX_SIZE' makes sense.
        //
        if ( MAX_SIZE > 1 )
            {
            //
            //  Setup the class members.
            //
            Active = True;
            Used = 0;

            //
            //  Zero the 'SubSet'.
            //
            Zero( SubSet,MAX_SIZE );
            }
        else
            { FAILURE( "The max size must be greater than 1" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all of the sub-set.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> BOOLEAN SUB_SET<MAX_SIZE,VALUE>::DeleteAll( VOID )
        {
        //
        //  We ensure we are 'Active' before we delete the 'SubSet'.
        //
        if ( Active )
            {
            //
            //  Zero the 'SubSet' and 'Used'.
            //
            Zero( SubSet,MAX_SIZE );

            Used = 0;

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Value.                                                                             */
            /*                                                                                            */
            /*  Delete a value from the sub-set.                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> BOOLEAN SUB_SET<MAX_SIZE,VALUE>::DeleteValue( CONST VALUE & Value )
        {
        //
        //  We ensure we are 'Active' and that the 'SubSet' is not empty and has
        //  not overflowed.
        //
        if ( (Active) && (Used > 0) && (Used <= MAX_SIZE) )
            {
            AUTO SBIT32 Offset;

            //
            //  We find the 'Value' in the 'SubSet'.
            //
            if ( FindValue( ((VALUE*) & Value),& Offset ) )
                {
                AUTO FILE_SIZE Count;

                //
                //  We slide any remaining values up the array.
                //
                for ( Count=Offset;Count < (Used-1);Count ++ )
                    { SubSet[ Count ] = SubSet[ (Count+1) ]; }

                //
                //  We zero the last value as it is now free.
                //
                Zero( & SubSet[ (-- Used) ],1 );

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Value.                                                                               */
            /*                                                                                            */
            /*  Find a value in the sub-set.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> BOOLEAN SUB_SET<MAX_SIZE,VALUE>::FindValue( CONST VALUE & Value )
        {
        //
        //  We ensure we are 'Active' and that the 'SubSet' is not empty and has
        //  not overflowed.
        //
        if ( (Active) && (Used > 0) && (Used <= MAX_SIZE) )
            {
            AUTO SBIT32 Offset;

            //
            //  We find the 'Value' in the 'SubSet'.
            //
            return (FindValue( ((VALUE*) & Value),& Offset ));
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Value.                                                                               */
            /*                                                                                            */
            /*  Find a key and value pair in the sub-set.                                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> BOOLEAN SUB_SET<MAX_SIZE,VALUE>::FindValue
            (
            VALUE                               *Value,
            SBIT32                              *Offset
            )
        {
        //
        //  We way we find a 'Value' depends on number of elements 'Used'.
        //
        switch ( Used )
            {
            case 0:
                {
                //
                //  When the 'SubSet' is empty we can immediately exit. We also
                //  update the 'Offset' so we know where the 'Value' would fit in
                //  the 'SubSet'.
                //

                //
                (*Offset) = 0;

                break;
                }

            case 1: case 2: case 3: case 4: case 5: case 6: case 7:
                {
                AUTO SBIT32 Count;

                //
                //  When the 'SubSet' is small we can find the 'Value' with a simple scan.
                //
                for ( Count=0;Count < Used;Count ++ )
                    {
                    AUTO VALUE *Current = & SubSet[ Count ];

                    //
                    //  When we match the 'Value' (or scan past it) we exit the loop.
                    //  The 'SubSet' is kept in sorted order so there is no need to
                    //  go beyond a 'SubSet' entry that exceeds the 'Value'.
                    //
                    if ( (*Current) >= (*Value) )
                        {
                        (*Offset) = Count;

                        return ((*Value) == (*Current));
                        }
                    }

                //
                //  Update the 'Offset' so we know where the 'Value' would fit in
                //  the 'SubSet'.
                //
                (*Offset) = Used;

                break;
                }

            default:
                {
                AUTO SBIT32 Maximum = (Used-1);
                AUTO SBIT32 Minimum = 0;

                //
                //  We find the 'Value' using a binary search as 'SubSet' is stored
                //  in sorted order.
                //
                while ( Maximum >= Minimum )
                    {
                    AUTO SBIT32 MidPoint = ((Maximum + Minimum) / 2);
                    AUTO VALUE *MidValue = & SubSet[ MidPoint ];

                    //
                    //  Let's see if the 'MidValue' is greater than the 'Value'.
                    //
                    if ( (*MidValue) < (*Value) )
                        { Minimum = (MidPoint + 1); }
                    else
                        {
                        //
                        //  Let's see if the 'MidValue' value is less than the 'Value'.
                        //
                        if ( (*MidValue) > (*Value) )
                            { Maximum = (MidPoint - 1); }
                        else
                            {
                            //
                            //  We have found the target 'Value'.
                            //
                            (*Offset) = MidPoint;

                            return True;
                            }
                        }
                    }

                //
                //  Update the 'Offset' so we know where the 'Value' would fit in
                //  the 'SubSet'.
                //
                (*Offset) = Minimum;

                break;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Pop Value.                                                                                */
            /*                                                                                            */
            /*  Pop the highest value in the sub-set.                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> BOOLEAN SUB_SET<MAX_SIZE,VALUE>::PopValue( VALUE *Value )
        {
        //
        //  We ensure we are 'Active' and that the 'SubSet' is not empty and has
        //  not overflowed.
        //
        if ( (Active) && (Used > 0) && (Used <= MAX_SIZE) )
            {
            AUTO VALUE *Current = & SubSet[ (-- Used) ];

            //
            //  We return the highest 'Value' in the 'SubSet' and 'Zero' it.
            //
            (*Value) = (*Current);

            Zero( Current,1 );

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Pusk Value.                                                                               */
            /*                                                                                            */
            /*  A new value is pushed into the sub-set.                                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> BOOLEAN SUB_SET<MAX_SIZE,VALUE>::PushValue( CONST VALUE & Value )
        {
        //
        //  We ensure we are 'Active' and that the 'SubSet' is not empty and has
        //  not overflowed.
        //
        if ( (Active) && (Used > 0) && (Used <= MAX_SIZE) )
            {
            AUTO SBIT32 Offset;

            //
            //  We find the 'Value' in the 'SubSet'.
            //
            if ( ! FindValue( ((VALUE*) & Value),& Offset ) )
                {
                AUTO FILE_SIZE Count;

                //
                //  We slide any other values down the array.
                //
                for ( Count=(Used ++);Count > Offset;Count -- )
                    { SubSet[ Count ] = SubSet[ (Count-1) ]; }

                //
                //  We insert the new 'Value'.
                //
                SubSet[ Offset ] = Value;

                return True;
                }
            }
        else
            {
            //
            //  We can simply copy in the new 'Value' when the 'SubSet' is empty.
            //
            if ( Used == 0 )
                {
                SubSet[ (Used ++) ] = Value;

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Result Set.                                                                               */
            /*                                                                                            */
            /*  We return the result set.                                                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> BOOLEAN SUB_SET<MAX_SIZE,VALUE>::ResultSet
            (
            VALUE                               **Set,
            SBIT32                              *Size
            )
        {
        //
        //  We ensure we are 'Active' and that the 'SubSet' has not overflowed.
        //
        if ( (Active) && (Used >= 0) && (Used <= MAX_SIZE) )
            {
            (*Set) = SubSet;
            (*Size) = Used;

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero.                                                                                     */
            /*                                                                                            */
            /*  Zero a collection of values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> VOID SUB_SET<MAX_SIZE,VALUE>::Zero
            (
            VALUE                               *Set,
            SBIT32                              Size
            )
        {
        //
        //  We make sure that 'Size' makes sense.
        //
        if ( (Size >= 0) && (Size <= MAX_SIZE) )
            {
            AUTO SBIT32 Count;

            //
            //  We have to be careful as a 'VALUE' is a user type.  It is very hard to zero
            //  these as we would not wish to damage them.  So just to be safe we call the
            //  destructor, then we zero them and finally we call the constructor again to
            //  rebuild them.
            //
            for ( Count=(Size-1);Count >= 0;Count -- )
                { PLACEMENT_DELETE( & Set[ Count ],VALUE ); }

            MemoryFunctions.Zero( Set,(Size * sizeof(VALUE)) );

            for ( Count=0;Count < Size;Count ++ )
                { PLACEMENT_NEW( & Set[ Count ],VALUE ); }
            }
        else
            { FAILURE( "Wrong size for set" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'SUB_SET'.                                                                 */
            /*                                                                                            */
            /*  The destructor for 'SUB_SET' cleans up the class.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <SBIT32 MAX_SIZE,class VALUE> SUB_SET<MAX_SIZE,VALUE>::~SUB_SET( VOID )
        {
        //
        //  Delete the sub-set.
        //
        DeleteAll();

        //
        //  Zero the class members.
        //
        Used = 0;
        Active = False;
        }
    }
#endif