#ifndef _SUB_STRING_HPP_
#define _SUB_STRING_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "AutomaticTransactionScope.hpp"
#include "AutomaticViewScope.hpp"
#include "BaseTypes.hpp"
#include "Failure.hpp"
#include "MemoryFunctions.hpp"
#include "ReservedWords.hpp"
#include "String.hpp"
#include "StringFunctions.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Sub-String.                                                                                       */
        /*                                                                                                    */
        /*  A sub-string is a common core class which implements most of the string functionality.            */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  A Class Forward Reference.                                                                */
            /*                                                                                            */
            /*  A class reference is sometimes made before a class is fully specified.  If this is        */
            /*  the case a forward reference to the class is made here.                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE,FILE_SIZE MAX_SIZE> class FIXED_STRING;

    template <class HEAP,class TYPE> class FLEXIBLE_STRING;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Sub-String.                                                                               */
            /*                                                                                            */
            /*  A sub-string is a common core class which implements most of the string functionality.    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> class SUB_STRING
        {
            //
            //  Private constants.
            //
            STATIC CONST SBIT32 Mask            = ((1 << (sizeof(TYPE) * 8)) - 1);

        protected:
            //
            //  Protected types.
            //
            typedef enum
                {
                NoStringType                    = 0,
                FixedStringType                 = 1,
                FlexibleStringType              = 2,
                EndOfStringTypes
                }
            STRING_TYPE;

            //
            //  Protected data.
            //
            FILE_SIZE                           Type;

            //
            //  Protected static data.
            //
            STATIC MEMORY_FUNCTIONS<SNATIVE>    MemoryFunctions;
            STATIC STRING_FUNCTIONS<CHAR>       StringFunctions1;
            STATIC STRING_FUNCTIONS<TYPE>       StringFunctions2;
            STATIC STRING_FUNCTIONS<WCHAR>      StringFunctions3;

        public:
            //
            //  Public functions.
            //
            SUB_STRING( VOID );

            BOOLEAN Assignment( CONST SUB_STRING<HEAP,CHAR> & String );
            BOOLEAN Assignment( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST SUB_STRING<HEAP,CHAR> & String );

            BOOLEAN Assignment( CONST SUB_STRING<HEAP,WCHAR> & String );
            BOOLEAN Assignment( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST SUB_STRING<HEAP,WCHAR> & String );

            BOOLEAN Assignment( CONST CHAR *Text );
            BOOLEAN Assignment( CONST CHAR *Text,FILE_SIZE Size );
            BOOLEAN Assignment( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST CHAR *Text );
            BOOLEAN Assignment( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST CHAR *Text,FILE_SIZE Size );

            BOOLEAN Assignment( CONST WCHAR *Text );
            BOOLEAN Assignment( CONST WCHAR *Text,FILE_SIZE Size );
            BOOLEAN Assignment( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST WCHAR *Text );
            BOOLEAN Assignment( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST WCHAR *Text,FILE_SIZE Size );

            BOOLEAN Compare( CONST TYPE *Text,SBIT16 *Result );
            BOOLEAN Compare( CONST TYPE *Text,FILE_SIZE Size,SBIT16 *Result );
            BOOLEAN Compare
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                CONST TYPE                      *Text,
                FILE_SIZE                       Size,
                SBIT16                          *Result
                );

            BOOLEAN Compare( CONST SUB_STRING<HEAP,TYPE> & String,SBIT16 *Result );
            BOOLEAN Compare
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                CONST SUB_STRING<HEAP,TYPE>     & String,
                SBIT16                          *Result
                );

            BOOLEAN Join( CONST TYPE *Text );
            BOOLEAN Join( CONST TYPE *Text,FILE_SIZE Size );
            BOOLEAN Join( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST TYPE *Text,FILE_SIZE Size );

            BOOLEAN Join( CONST SUB_STRING<HEAP,TYPE> & String );
            BOOLEAN Join( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST SUB_STRING<HEAP,TYPE> & String );

            BOOLEAN Lower( VOID );
            BOOLEAN Lower( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN Upper( VOID );
            BOOLEAN Upper( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN Value( FILE_SIZE MaxSize,FILE_SIZE *Size,TYPE *Text );

            BOOLEAN Value
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                FILE_SIZE                       MaxSize,
                FILE_SIZE                       *Size,
                TYPE                            *Text
                );

            ~SUB_STRING( VOID );

            //
            //  Public operators.
            //
            VOID operator=( CONST SUB_STRING<HEAP,CHAR> & String );
            VOID operator=( CONST SUB_STRING<HEAP,WCHAR> & String );
            VOID operator=( CONST CHAR *Text );
            VOID operator=( CONST WCHAR *Text );

            VOID operator+=( CONST SUB_STRING<HEAP,TYPE> & String );
            VOID operator+=( CONST TYPE *Text );

            bool operator==( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator!=( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator< ( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator<=( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator> ( CONST SUB_STRING<HEAP,TYPE> & String );
            bool operator>=( CONST SUB_STRING<HEAP,TYPE> & String );

            bool operator==( CONST TYPE *Text );
            bool operator!=( CONST TYPE *Text );
            bool operator< ( CONST TYPE *Text );
            bool operator<=( CONST TYPE *Text );
            bool operator> ( CONST TYPE *Text );
            bool operator>=( CONST TYPE *Text );

            //
            //  Public internal functions.
            //
            BOOLEAN Resize( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,STRING<TYPE> **String,FILE_SIZE Size );

            BOOLEAN Update( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,STRING<TYPE> **String );

            BOOLEAN View( AUTOMATIC_VIEW_SCOPE<HEAP> *AutomaticViewScope,STRING<TYPE> **String );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Class Static Data.                                                                        */
            /*                                                                                            */
            /*  The class static data needs to be initialised before it's ready for use.                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> MEMORY_FUNCTIONS<SNATIVE> SUB_STRING<HEAP,TYPE>::MemoryFunctions;
    template <class HEAP,class TYPE> STRING_FUNCTIONS<CHAR> SUB_STRING<HEAP,TYPE>::StringFunctions1;
    template <class HEAP,class TYPE> STRING_FUNCTIONS<TYPE> SUB_STRING<HEAP,TYPE>::StringFunctions2;
    template <class HEAP,class TYPE> STRING_FUNCTIONS<WCHAR> SUB_STRING<HEAP,TYPE>::StringFunctions3;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'SUB_STRING'.                                                             */
            /*                                                                                            */
            /*  The constructor for 'SUB_STRING' prepares the class for use.                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> SUB_STRING<HEAP,TYPE>::SUB_STRING( VOID )
        {
        //
        //  Setup the class members.
        //
        Type = NoStringType;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            CONST SUB_STRING<HEAP,CHAR>         & String
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Assignment( & AutomaticHeapScope,String ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST SUB_STRING<HEAP,CHAR>         & String
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'Update()' as we need to modify it.  The function calls to 'BeginTransaction()'
        //  and 'Update()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO SUB_STRING<HEAP,CHAR> *NewValue = ((SUB_STRING<HEAP,CHAR>*) & String);
            AUTO STRING<TYPE> *String1;
            AUTO STRING<CHAR> *String2;

            //
            //  We 'Update()' the target and 'View()' the 'NewValue'.
            //
            if
                    (
                    (Update( AutomaticHeapScope,& String1 ))
                        &&
                    (NewValue -> View( & AutomaticViewScope,& String2 ))
                    )
                {
                AUTO FILE_SIZE MinCopy = (String2 -> Used);
                AUTO FILE_SIZE MaxCopy = (MinCopy);
                AUTO FILE_SIZE Count;

                //
                //  Let's see if there is enough space to store the 'Text'.
                //
                if ( MaxCopy > (String1 -> MaxSize-1) )
                    {
                    //
                    //  We will 'Resize()' so it is large enough for the 'Text'.
                    //
                    if ( ! Resize( AutomaticHeapScope,& String1,(MaxCopy+1) ) )
                        {
                        //
                        //  If we can't 'Resize()' and we can't truncate then
                        //  we are forced to exit.
                        //
                        if ( AutomaticHeapScope -> GetTruncate() )
                            { MinCopy = (String1 -> MaxSize-1); }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorStringOverflow )); }
                        }
                    }

                //
                //  We now copy the value.
                //
                if ( String1 -> MaxSize > 0 )
                    {
                    //
                    //  Copy the 'Text' to the target 'String'.
                    //
                    for ( Count=0;Count < MinCopy;Count ++ )
                        { String1 -> Text[ Count ] = ((TYPE) String2 -> Text[ Count ]); }

                    //
                    //  Zero the end of the 'String' if necessary.
                    //
                    for ( Count=MinCopy;Count < String1 -> Used;Count ++ )
                        { String1 -> Text[ Count ] = NULL; }

                    //
                    //  Update the size 'Used'.
                    //
                    String1 -> Used = MinCopy;
                    }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            CONST SUB_STRING<HEAP,WCHAR>        & String
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Assignment( & AutomaticHeapScope,String ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST SUB_STRING<HEAP,WCHAR>        & String
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'Update()' as we need to modify it.  The function calls to 'BeginTransaction()'
        //  and 'Update()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO SUB_STRING<HEAP,WCHAR> *NewValue = ((SUB_STRING<HEAP,WCHAR>*) & String);
            AUTO STRING<TYPE> *String1;
            AUTO STRING<WCHAR> *String2;

            //
            //  We 'Update()' the target and 'View()' the 'NewValue'.
            //
            if
                    (
                    (Update( AutomaticHeapScope,& String1 ))
                        &&
                    (NewValue -> View( & AutomaticViewScope,& String2 ))
                    )
                {
                AUTO FILE_SIZE MinCopy = (String2 -> Used);
                AUTO FILE_SIZE MaxCopy = (MinCopy);
                AUTO FILE_SIZE Count;

                //
                //  Let's see if there is enough space to store the 'Text'.
                //
                if ( MaxCopy > (String1 -> MaxSize-1) )
                    {
                    //
                    //  We will 'Resize()' so it is large enough for the 'Text'.
                    //
                    if ( ! Resize( AutomaticHeapScope,& String1,(MaxCopy+1) ) )
                        {
                        //
                        //  If we can't 'Resize()' and we can't truncate then
                        //  we are forced to exit.
                        //
                        if ( AutomaticHeapScope -> GetTruncate() )
                            { MinCopy = (String1 -> MaxSize-1); }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorStringOverflow )); }
                        }
                    }

                //
                //  We now copy the value.
                //
                if ( String1 -> MaxSize > 0 )
                    {
                    //
                    //  We automatically convert to the 'TYPE' of the 'String'.  When the
                    //  'TYPE' of the destination 'String' is larger than 'WCHAR' it is just
                    //  a simple copy.  When the 'TYPE' of the destination 'String' is smaller
                    //  than 'WCHAR' we may have to to truncate the characters.
                    //
                    if ( sizeof(TYPE) >= sizeof(WCHAR) )
                        {
                        //
                        //  Copy the 'NewValue' to the target 'String'.
                        //
                        for ( Count=0;Count < MinCopy;Count ++ )
                            { String1 -> Text[ Count ] = ((TYPE) String2 -> Text[ Count ]); }
                        }
                    else
                        {
                        //
                        //  When we can't 'Truncate' we have to ensure that the 'Text' is suitable
                        //  before copying it to the 'String'.
                        //
                        if ( ! AutomaticHeapScope -> GetTruncate() )
                            {
                            //
                            //  We scan the 'Text' and ensure it is suitable.
                            //
                            for ( Count=0;Count < MinCopy;Count ++ )
                                {
                                if ( (String2 -> Text[ Count ] & ~Mask) != 0 )
                                    { return (AutomaticHeapScope -> SetError( ErrorStringConversion )); }
                                }

                            //
                            //  We can now copy the 'Text'.
                            //
                            for ( Count=0;Count < MinCopy;Count ++ )
                                { String1 -> Text[ Count ] = ((TYPE) String2 -> Text[ Count ]); }
                            }
                        else
                            {
                            //
                            //  Copy the 'Text' to the target 'String' converting the characters.
                            //
                            for ( Count=0;Count < MinCopy;Count ++ )
                                {
                                AUTO WCHAR Current = String2 -> Text[ Count ];

                                if ( (Current & ~Mask) == 0 )
                                    { String1 -> Text[ Count ] = ((TYPE) Current); }
                                else
                                    { String1 -> Text[ Count ] = ((TYPE) '?'); }
                                }
                            }
                        }

                    //
                    //  Zero the end of the 'String' if necessary.
                    //
                    for ( Count=MinCopy;Count < String1 -> Used;Count ++ )
                        { String1 -> Text[ Count ] = NULL; }

                    //
                    //  Update the size 'Used'.
                    //
                    String1 -> Used = MinCopy;
                    }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            CONST CHAR                          *Text
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Assignment( & AutomaticHeapScope,Text,StringFunctions1.Length( ((CHAR*) Text) ) ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            CONST CHAR                          *Text,
            FILE_SIZE                           Size
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Assignment( & AutomaticHeapScope,Text,Size ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST CHAR                          *Text
            )
        { return (Assignment( AutomaticHeapScope,Text,StringFunctions1.Length( ((CHAR*) Text) ) )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST CHAR                          *Text,
            FILE_SIZE                           Size
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'Update()' as we need to modify it.  The function calls to 'BeginTransaction()'
        //  and 'Update()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO STRING<TYPE> *String;

            //
            //  We need to 'Update()'.
            //
            if ( Update( AutomaticHeapScope,& String ) )
                {
                AUTO FILE_SIZE MinCopy = (Size);
                AUTO FILE_SIZE MaxCopy = (MinCopy);
                AUTO FILE_SIZE Count;

                //
                //  Let's see if there is enough space to store the 'Text'.
                //
                if ( MaxCopy > (String -> MaxSize-1) )
                    {
                    //
                    //  We will 'Resize()' so it is large enough for the 'Text'.
                    //
                    if ( ! Resize( AutomaticHeapScope,& String,(MaxCopy+1) ) )
                        {
                        //
                        //  If we can't 'Resize()' and we can't truncate then
                        //  we are forced to exit.
                        //
                        if ( AutomaticHeapScope -> GetTruncate() )
                            { MinCopy = (String -> MaxSize-1); }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorStringOverflow )); }
                        }
                    }

                //
                //  We now copy the value.
                //
                if ( String -> MaxSize > 0 )
                    {
                    //
                    //  Copy the 'Text' to the target 'String'.
                    //
                    for ( Count=0;Count < MinCopy;Count ++ )
                        { String -> Text[ Count ] = ((TYPE) Text[ Count ]); }

                    //
                    //  Zero the end of the 'String' if necessary.
                    //
                    for ( Count=MinCopy;Count < String -> Used;Count ++ )
                        { String -> Text[ Count ] = NULL; }

                    //
                    //  Update the size 'Used'.
                    //
                    String -> Used = MinCopy;
                    }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            CONST WCHAR                         *Text
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Assignment( & AutomaticHeapScope,Text,StringFunctions3.Length( ((WCHAR*) Text) ) ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            CONST WCHAR                         *Text,
            FILE_SIZE                           Size
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Assignment( & AutomaticHeapScope,Text,Size ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST WCHAR                         *Text
            )
        { return (Assignment( AutomaticHeapScope,Text,StringFunctions3.Length( ((WCHAR*) Text) ) )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Assignment.                                                                               */
            /*                                                                                            */
            /*  Assign a new value to a string.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Assignment
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST WCHAR                         *Text,
            FILE_SIZE                           Size
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'Update()' as we need to modify it.  The function calls to 'BeginTransaction()'
        //  and 'Update()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO STRING<TYPE> *String;

            //
            //  We need to 'Update()'.
            //
            if ( Update( AutomaticHeapScope,& String ) )
                {
                AUTO FILE_SIZE MinCopy = (Size);
                AUTO FILE_SIZE MaxCopy = (MinCopy);
                AUTO FILE_SIZE Count;

                //
                //  Let's see if there is enough space to store the 'Text'.
                //
                if ( MaxCopy > (String -> MaxSize-1) )
                    {
                    //
                    //  We will 'Resize()' so it is large enough for the 'Text'.
                    //
                    if ( ! Resize( AutomaticHeapScope,& String,(MaxCopy+1) ) )
                        {
                        //
                        //  If we can't 'Resize()' and we can't truncate then
                        //  we are forced to exit.
                        //
                        if ( AutomaticHeapScope -> GetTruncate() )
                            { MinCopy = (String -> MaxSize-1); }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorStringOverflow )); }
                        }
                    }

                //
                //  We now copy the value.
                //
                if ( String -> MaxSize > 0 )
                    {
                    //
                    //  We automatically convert to the 'TYPE' of the 'String'.  When the
                    //  'TYPE' of the destination 'String' is larger than 'WCHAR' it is just
                    //  a simple copy.  When the 'TYPE' of the destination 'String' is smaller
                    //  than 'WCHAR' we may have to to truncate the characters.
                    //
                    if ( sizeof(TYPE) >= sizeof(WCHAR) )
                        {
                        //
                        //  Copy the 'Text' to the target 'String'.
                        //
                        for ( Count=0;Count < MinCopy;Count ++ )
                            { String -> Text[ Count ] = ((TYPE) Text[ Count ]); }
                        }
                    else
                        {
                        //
                        //  When we can't 'Truncate' we have to ensure that the 'Text' is suitable
                        //  before copying it to the 'String'.
                        //
                        if ( ! AutomaticHeapScope -> GetTruncate() )
                            {
                            //
                            //  We scan the 'Text' and ensure it is suitable.
                            //
                            for ( Count=0;Count < MinCopy;Count ++ )
                                {
                                if ( (Text[ Count ] & ~Mask) != 0 )
                                    { return (AutomaticHeapScope -> SetError( ErrorStringConversion )); }
                                }

                            //
                            //  We can now copy the 'Text'.
                            //
                            for ( Count=0;Count < MinCopy;Count ++ )
                                { String -> Text[ Count ] = ((TYPE) Text[ Count ]); }
                            }
                        else
                            {
                            //
                            //  Copy the 'Text' to the target 'String' converting the characters.
                            //
                            for ( Count=0;Count < MinCopy;Count ++ )
                                {
                                AUTO WCHAR Current = Text[ Count ];

                                if ( (Current & ~Mask) == 0 )
                                    { String -> Text[ Count ] = ((TYPE) Current); }
                                else
                                    { String -> Text[ Count ] = ((TYPE) '?'); }
                                }
                            }
                        }

                    //
                    //  Zero the end of the 'String' if necessary.
                    //
                    for ( Count=MinCopy;Count < String -> Used;Count ++ )
                        { String -> Text[ Count ] = NULL; }

                    //
                    //  Update the size 'Used'.
                    //
                    String -> Used = MinCopy;
                    }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Compare.                                                                                  */
            /*                                                                                            */
            /*  Compare two strings.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Compare
            (
            CONST TYPE                          *Text,
            SBIT16                              *Result
            )
        { return (Compare( Text,StringFunctions2.Length( ((TYPE*) Text) ),Result )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Compare.                                                                                  */
            /*                                                                                            */
            /*  Compare two strings.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Compare
            (
            CONST TYPE                          *Text,
            FILE_SIZE                           Size,
            SBIT16                              *Result
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Compare( & AutomaticHeapScope,Text,Size,Result ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Compare.                                                                                  */
            /*                                                                                            */
            /*  Compare two strings.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Compare
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST TYPE                          *Text,
            FILE_SIZE                           Size,
            SBIT16                              *Result
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'View()' as we need to examine it.  The function calls to 'BeginTransaction()'
        //  and 'View()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO STRING<TYPE> *String;

            //
            //  We need to 'View()' so we can compare the value.
            //
            if ( View( & AutomaticViewScope,& String ) )
                {
                AUTO FILE_SIZE MinSize = ((String -> Used <= Size) ? String -> Used : Size);
                AUTO FILE_SIZE Count;

                //
                //  We compare the values to make sure they match.
                //
                for ( Count=0,(*Result)=0;((*Result) == 0) && (Count < MinSize);Count ++ )
                    {
                    if ( String -> Text[ Count ] != Text[ Count ] )
                        { (*Result) = ((String -> Text[ Count ] < Text[ Count ]) ? -1 : 1); }
                    }

                //
                //  When the values match the shorter string is the smallest.
                //
                if ( ((*Result) == 0) && (String -> Used != Size) )
                    { (*Result) = ((String -> Used < Size) ? -1 : 1); }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Compare.                                                                                  */
            /*                                                                                            */
            /*  Compare two strings.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Compare
            (
            CONST SUB_STRING<HEAP,TYPE>         & String,
            SBIT16                              *Result
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Compare( & AutomaticHeapScope,String,Result ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Compare.                                                                                  */
            /*                                                                                            */
            /*  Compare two strings.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Compare
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST SUB_STRING<HEAP,TYPE>         & String,
            SBIT16                              *Result

            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'View()' as we need to examine it.  The function calls to 'BeginTransaction()'
        //  and 'View()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
            AUTO SUB_STRING<HEAP,TYPE> *NewValue = ((SUB_STRING<HEAP,TYPE>*) & String);
            AUTO STRING<TYPE> *String1;
            AUTO STRING<TYPE> *String2;

            //
            //  We need to 'View()' so we can compare the two values.
            //
            if
                    (
                    (View( & AutomaticViewScope1,& String1 ))
                        &&
                    (NewValue -> View( & AutomaticViewScope2,& String2 ))
                    )
                {
                AUTO FILE_SIZE MinSize = ((String1 -> Used <= String2 -> Used) ? String1 -> Used : String2 -> Used);
                AUTO FILE_SIZE Count;

                //
                //  We compare the values to make sure they match.
                //
                for ( Count=0,(*Result)=0;((*Result) == 0) && (Count < MinSize);Count ++ )
                    {
                    if ( String1 -> Text[ Count ] != String2 -> Text[ Count ] )
                        { (*Result) = ((String1 -> Text[ Count ] < String2 -> Text[ Count ]) ? -1 : 1); }
                    }

                //
                //  When the values match the shorter string is the smallest.
                //
                if ( ((*Result) == 0) && (String1 -> Used != String2 -> Used) )
                    { (*Result) = ((String1 -> Used < String2 -> Used) ? -1 : 1); }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  Join two strings.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Join
            (
            CONST TYPE                          *Text
            )
        { return (Join( Text,StringFunctions2.Length( ((TYPE*) Text) ) )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  Join two strings.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Join
            (
            CONST TYPE                          *Text,
            FILE_SIZE                           Size
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Join( & AutomaticHeapScope,Text,Size ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  Join two strings.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Join
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST TYPE                          *Text,
            FILE_SIZE                           Size
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'Update()' as we need to modify it.  The function calls to 'BeginTransaction()'
        //  and 'Update()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO STRING<TYPE> *String;

            //
            //  We need to 'Update()'.
            //
            if ( Update( AutomaticHeapScope,& String ) )
                {
                AUTO FILE_SIZE MinCopy = (Size);
                AUTO FILE_SIZE MaxCopy = (String -> Used + MinCopy);
                AUTO FILE_SIZE Count;

                //
                //  Let's see if there is enough space to store the 'Text'.
                //
                if ( MaxCopy > (String -> MaxSize-1) )
                    {
                    //
                    //  We will 'Resize()' so it is large enough for the 'Text'.
                    //
                    if ( ! Resize( AutomaticHeapScope,& String,(MaxCopy+1) ) )
                        {
                        //
                        //  If we can't 'Resize()' and we can't truncate then
                        //  we are forced to exit.
                        //
                        if ( AutomaticHeapScope -> GetTruncate() )
                            { MinCopy = (String -> MaxSize - String -> Used - 1); }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorStringOverflow )); }
                        }
                    }

                //
                //  Add the new 'Text' to the 'String'.
                //
                for ( Count=0;Count < MinCopy;Count ++ )
                    { String -> Text[ (String -> Used + Count) ] = Text[ Count ]; }

                //
                //  Update the size 'Used'.
                //
                String -> Used += MinCopy;

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  Join two strings.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Join
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Join( & AutomaticHeapScope,String ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join.                                                                                     */
            /*                                                                                            */
            /*  Join two strings.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Join
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'Update()' as we need to modify it.  The function calls to 'BeginTransaction()'
        //  and 'Update()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
            AUTO SUB_STRING<HEAP,TYPE> *NewValue = ((SUB_STRING<HEAP,TYPE>*) & String);
            AUTO STRING<TYPE> *String1;
            AUTO STRING<TYPE> *String2;

            //
            //  We need to 'View()' so we can compare the two values.
            //
            if
                    (
                    (View( & AutomaticViewScope1,& String1 ))
                        &&
                    (NewValue -> View( & AutomaticViewScope2,& String2 ))
                    )
                {
                AUTO FILE_SIZE MinCopy = (String2 -> Used);
                AUTO FILE_SIZE MaxCopy = (String1 -> Used + MinCopy);
                AUTO FILE_SIZE Count;

                //
                //  Let's see if there is enough space to store the 'Text'.
                //
                if ( MaxCopy > (String1 -> MaxSize-1) )
                    {
                    //
                    //  We will 'Resize()' so it is large enough for the 'Text'.
                    //
                    if ( ! Resize( AutomaticHeapScope,& String1,(MaxCopy+1) ) )
                        {
                        //
                        //  If we can't 'Resize()' and we can't truncate then
                        //  we are forced to exit.
                        //
                        if ( AutomaticHeapScope -> GetTruncate() )
                            { MinCopy = (String1 -> MaxSize - String1 -> Used - 1); }
                        else
                            { return (AutomaticHeapScope -> SetError( ErrorStringOverflow )); }
                        }
                    }

                //
                //  Add the new 'Text' to the 'String'.
                //
                for ( Count=0;Count < MinCopy;Count ++ )
                    { String1 -> Text[ (String1 -> Used + Count) ] = String2 -> Text[ Count ]; }

                //
                //  Update the size 'Used'.
                //
                String1 -> Used += MinCopy;

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Lower.                                                                                    */
            /*                                                                                            */
            /*  We convert a string to lower case.                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Lower( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Lower( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Lower.                                                                                    */
            /*                                                                                            */
            /*  We convert a string to lower case.                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Lower
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'Update()' as we need to modify it.  The function calls to 'BeginTransaction()'
        //  and 'Update()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO STRING<TYPE> *String;

            //
            //  We 'Update()' the target.
            //
            if ( Update( AutomaticHeapScope,& String ) )
                {
                AUTO FILE_SIZE Count;
                STATIC CONST TYPE Delta = (((TYPE) 'a') - ((TYPE) 'A'));
                STATIC CONST TYPE A = ((TYPE) 'A');
                STATIC CONST TYPE Z = ((TYPE) 'Z');

                //
                //  Convert all ascii lower case to lower case.
                //
                for ( Count=0;Count < String -> Used;Count ++ )
                    {
                    AUTO TYPE *Current = & String -> Text[ Count ];

                    if ( ((*Current) >= A) && ((*Current) <= Z) )
                        { (*Current) += Delta; }
                    }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID SUB_STRING<HEAP,TYPE>::operator=
            (
            CONST SUB_STRING<HEAP,CHAR>         & String
            )
        {
        if ( ! Assignment( String ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID SUB_STRING<HEAP,TYPE>::operator=
            (
            CONST SUB_STRING<HEAP,WCHAR>        & String
            )
        {
        if ( ! Assignment( String ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID SUB_STRING<HEAP,TYPE>::operator=
            (
            CONST CHAR                          *Text
            )
        {
        if ( ! Assignment( Text ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '='.                                                                             */
            /*                                                                                            */
            /*  The operator '='.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID SUB_STRING<HEAP,TYPE>::operator=
            (
            CONST WCHAR                         *Text
            )
        {
        if ( ! Assignment( Text ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '+='.                                                                            */
            /*                                                                                            */
            /*  The operator '+='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID SUB_STRING<HEAP,TYPE>::operator+=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        if ( ! Join( String ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '+='.                                                                            */
            /*                                                                                            */
            /*  The operator '+='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> VOID SUB_STRING<HEAP,TYPE>::operator+=
            (
            CONST TYPE                          *Text
            )
        {
        if ( ! Join( Text ) )
            { FAILURE( "String assignment failure" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '=='.                                                                            */
            /*                                                                                            */
            /*  The operator '=='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator==
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( String,& Result ) )
            { return (Result == 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '!='.                                                                            */
            /*                                                                                            */
            /*  The operator '!='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator!=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( String,& Result ) )
            { return (Result != 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<'.                                                                             */
            /*                                                                                            */
            /*  The operator '<'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator<
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( String,& Result ) )
            { return (Result < 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<='.                                                                            */
            /*                                                                                            */
            /*  The operator '<='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator<=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( String,& Result ) )
            { return (Result <= 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>'.                                                                             */
            /*                                                                                            */
            /*  The operator '>'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator>
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( String,& Result ) )
            { return (Result > 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>='.                                                                            */
            /*                                                                                            */
            /*  The operator '>='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator>=
            (
            CONST SUB_STRING<HEAP,TYPE>         & String
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( String,& Result ) )
            { return (Result >= 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '=='.                                                                            */
            /*                                                                                            */
            /*  The operator '=='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator==
            (
            CONST TYPE                          *Text
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( Text,& Result ) )
            { return (Result == 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '!='.                                                                            */
            /*                                                                                            */
            /*  The operator '!='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator!=
            (
            CONST TYPE                          *Text
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( Text,& Result ) )
            { return (Result != 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<'.                                                                             */
            /*                                                                                            */
            /*  The operator '<'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator<
            (
            CONST TYPE                          *Text
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( Text,& Result ) )
            { return (Result < 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '<='.                                                                            */
            /*                                                                                            */
            /*  The operator '<='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator<=
            (
            CONST TYPE                          *Text
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( Text,& Result ) )
            { return (Result <= 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>'.                                                                             */
            /*                                                                                            */
            /*  The operator '>'.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator>
            (
            CONST TYPE                          *Text
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( Text,& Result ) )
            { return (Result > 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  operator '>='.                                                                            */
            /*                                                                                            */
            /*  The operator '>='.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> bool SUB_STRING<HEAP,TYPE>::operator>=
            (
            CONST TYPE                          *Text
            )
        {
        AUTO SBIT16 Result;

        if ( Compare( Text,& Result ) )
            { return (Result >= 0); }
        else
            { return ((FAILURE( "String compare failure" )), false); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Resize.                                                                                   */
            /*                                                                                            */
            /*  Resize a string.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Resize
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            STRING<TYPE>                        **String,
            FILE_SIZE                           Size
            )
        {
        //
        //  It would be nice to use a C++ 'virtual' pointer but this uses v-tables and
        //  pointers.  The strings supported by this class can also be written to files
        //  and so we need to implement them using a 'Type' instead of a v-table.
        //
        switch ( Type )
            {
            case FixedStringType:
                { return ((FIXED_STRING<HEAP,TYPE,1>*) this) -> Resize( AutomaticHeapScope,String,Size ); }

            case FlexibleStringType:
                { return ((FLEXIBLE_STRING<HEAP,TYPE>*) this) -> Resize( AutomaticHeapScope,String,Size ); }

            default:
                { return False; }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update.                                                                                   */
            /*                                                                                            */
            /*  Update a string.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Update
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            STRING<TYPE>                        **String
            )
        {
        //
        //  It would be nice to use a C++ 'virtual' pointer but this uses v-tables and
        //  pointers.  The strings supported by this class can also be written to files
        //  and so we need to implement them using a 'Type' instead of a v-table.
        //
        switch ( Type )
            {
            case FixedStringType:
                { return ((FIXED_STRING<HEAP,TYPE,1>*) this) -> Update( AutomaticHeapScope,String ); }

            case FlexibleStringType:
                { return ((FLEXIBLE_STRING<HEAP,TYPE>*) this) -> Update( AutomaticHeapScope,String ); }

            default:
                { return False; }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Upper.                                                                                    */
            /*                                                                                            */
            /*  We convert a string to upper case.                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Upper( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Upper( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Upper.                                                                                    */
            /*                                                                                            */
            /*  We convert a string to upper case.                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Upper
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'Update()' as we need to modify it.  The function calls to 'BeginTransaction()'
        //  and 'Update()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO STRING<TYPE> *String;

            //
            //  We 'Update()' the target.
            //
            if ( Update( AutomaticHeapScope,& String ) )
                {
                AUTO FILE_SIZE Count;
                STATIC CONST TYPE Delta = (((TYPE) 'a') - ((TYPE) 'A'));
                STATIC CONST TYPE A = ((TYPE) 'a');
                STATIC CONST TYPE Z = ((TYPE) 'z');

                //
                //  Convert all ascii lower case to upper case.
                //
                for ( Count=0;Count < String -> Used;Count ++ )
                    {
                    AUTO TYPE *Current = & String -> Text[ Count ];

                    if ( ((*Current) >= A) && ((*Current) <= Z) )
                        { (*Current) -= Delta; }
                    }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Value.                                                                                    */
            /*                                                                                            */
            /*  The value of a string.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Value
            (
            FILE_SIZE                           MaxSize,
            FILE_SIZE                           *Size,
            TYPE                                *Text
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Value( & AutomaticHeapScope,MaxSize,Size,Text ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Value.                                                                                    */
            /*                                                                                            */
            /*  The value of a string.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::Value
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           MaxSize,
            FILE_SIZE                           *Size,
            TYPE                                *Text
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

        //
        //  We need to 'BeginTransaction()' to ensure that we have a transaction.  Next we
        //  'View()' as we need to examine it.  The function calls to 'BeginTransaction()'
        //  and 'View()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has
        //  been written.
        //
        if ( AutomaticTransactionScope.BeginTransaction() )
            {
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO STRING<TYPE> *String;

            //
            //  We need to 'View()' so we can see the value.
            //
            if ( View( & AutomaticViewScope,& String ) )
                {
                AUTO FILE_SIZE Count;

                //
                //  We copy the value to 'Text'.
                //
                for ( Count=0;(Count < (MaxSize-1)) && (Count < String -> Used);Count ++ )
                    { Text[ Count ] = String -> Text[ Count ]; }

                //
                //  We return the 'Size'.
                //
                (*Size) = Count;

                //
                //  We always add a 'NULL' to the 'Text'.
                //
                Text[ Count ] = NULL;

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View a string.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> BOOLEAN SUB_STRING<HEAP,TYPE>::View
            (
            AUTOMATIC_VIEW_SCOPE<HEAP>          *AutomaticViewScope,
            STRING<TYPE>                        **String
            )
        {
        //
        //  It would be nice to use a C++ 'virtual' pointer but this uses v-tables and
        //  pointers.  The strings supported by this class can also be written to files
        //  and so we need to implement them using a 'Type' instead of a v-table.
        //
        switch ( Type )
            {
            case FixedStringType:
                { return ((FIXED_STRING<HEAP,TYPE,1>*) this) -> View( AutomaticViewScope,String ); }

            case FlexibleStringType:
                { return ((FLEXIBLE_STRING<HEAP,TYPE>*) this) -> View( AutomaticViewScope,String ); }

            default:
                { return False; }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'SUB_STRING'.                                                              */
            /*                                                                                            */
            /*  The destructor for 'SUB_STRING' cleans up the class.                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class TYPE> SUB_STRING<HEAP,TYPE>::~SUB_STRING( VOID )
        {
        //
        //  Zero the class members.
        //
        Type = NoStringType;
        }
    }
#endif