#ifndef _TRANSACTION_LOCK_HPP_
#define _TRANSACTION_LOCK_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "BaseTypes.hpp"
#include "ExportMemoryLock.hpp"
#include "PublicConstants.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Transaction Lock.                                                                                 */
        /*                                                                                                    */
        /*  A transaction lock only works in connection with the transactional heaps.                         */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  A Class Forward Reference.                                                                */
            /*                                                                                            */
            /*  A class reference is sometimes made before a class is fully specified.  If this is        */
            /*  the case a forward reference to the class is made here.                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> class AUTOMATIC_HEAP_SCOPE;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Transaction Lock.                                                                         */
            /*                                                                                            */
            /*  A transaction lock only works in connection with the transactional heaps.                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> class TRANSACTION_LOCK : public EXPORT_MEMORY_LOCK
        {
            //
            //  Private data.
            //
            HEAP                                *Heap;

        public:
            //
            //  Public inline functions.
            //
            TRANSACTION_LOCK( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope ) : EXPORT_MEMORY_LOCK( NULL )
                { Heap = (AutomaticHeapScope -> GetHeap()); }

            TRANSACTION_LOCK( HEAP *NewHeap ) : EXPORT_MEMORY_LOCK( NULL )
                { Heap = NewHeap; }

            INLINE BOOLEAN ClaimExclusiveLock( BOOLEAN Wait = True )
                {
                if ( Heap -> Feature( RockallTransactionalSupport ) )
                    { return (EXPORT_MEMORY_LOCK::ClaimExclusiveLock( Wait )); }
                else
                    { return True; }
                }

            INLINE BOOLEAN ClaimSharedLock( BOOLEAN Wait = True )
                {
                if ( Heap -> Feature( RockallTransactionalSupport ) )
                    { return (EXPORT_MEMORY_LOCK::ClaimSharedLock( Wait )); }
                else
                    { return True; }
                }

            INLINE VOID ReleaseExclusiveLock( VOID )
                {
                if ( Heap -> Feature( RockallTransactionalSupport ) )
                    { EXPORT_MEMORY_LOCK::ReleaseExclusiveLock(); }
                }

            INLINE VOID ReleaseSharedLock( VOID )
                {
                if ( Heap -> Feature( RockallTransactionalSupport ) )
                    { EXPORT_MEMORY_LOCK::ReleaseSharedLock(); }
                }

            ~TRANSACTION_LOCK( VOID )
                { Heap = NULL; }

            //
            //  Public inline functions.
            //
            INLINE VOID *GetLock( VOID )
                {
                if ( Heap -> Feature( RockallTransactionalSupport ) )
                    { return (EXPORT_MEMORY_LOCK::GetLock()); }
                else
                    { return NULL; }
                }

            INLINE VOID RegisterLock( VOID *NewLock )
                {
                if ( Heap -> Feature( RockallTransactionalSupport ) )
                    { EXPORT_MEMORY_LOCK::RegisterLock( NewLock ); }
                }

            INLINE VOID UpdateLock( VOID *NewLock )
                {
                if ( Heap -> Feature( RockallTransactionalSupport ) )
                    { EXPORT_MEMORY_LOCK::UpdateLock( NewLock ); }
                }

        private:
            //
            //  Disabled functionality.
            //
            TRANSACTION_LOCK( CONST TRANSACTION_LOCK & NoCopies );

            VOID operator=( CONST TRANSACTION_LOCK & NoCopies );
        };
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define TRANSACTION_LOCK ROCKALL::TRANSACTION_LOCK
#endif
#endif