#ifndef _ASYNCHRONOUS_CALLBACKS_HPP_
#define _ASYNCHRONOUS_CALLBACKS_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "ExportAsynchronousCallback.hpp"
#include "PublicConstants.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Asynchronous Callbacks.                                                                           */
        /*                                                                                                    */
        /*  An asynchronous callback can be scheduled at any point and will be executed when suitable         */
        /*  resources become available.                                                                       */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Asynchronous Callbacks.                                                                   */
            /*                                                                                            */
            /*  An asynchronous callback can be scheduled at any point and will be executed when          */
            /*  suitable resources become available.                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> class ASYNCHRONOUS_CALLBACKS
        {
            //
            //  Private types.
            //
            typedef VOID (*ASYNCHRONOUS_CALLBACK_FUNCTION)( TYPE *Type );

            //
            //  Private data.
            //
            EXPORT_ASYNCHRONOUS_CALLBACK        AsynchronousCallback;

        public:
            //
            //  Public functions.
            //
            ASYNCHRONOUS_CALLBACKS( VOID ) { /* void */ }

            SBIT32 NumberOfCPUs( VOID )
                { return (AsynchronousCallback.NumberOfCPUs()); }

            BOOLEAN Schedule( ASYNCHRONOUS_CALLBACK_FUNCTION Function,TYPE Type )
                { return (AsynchronousCallback.Schedule( Function,((VOID*) & Type),((SBIT32) sizeof(TYPE)) )); }

            BOOLEAN Schedule( ASYNCHRONOUS_CALLBACK_FUNCTION Function,VOID *Type,SBIT32 Size )
                { return (AsynchronousCallback.Schedule( ((VOID*) Function),Type,Size )); }

            VOID Sleep( SBIT32 Time )
                { AsynchronousCallback.Sleep( Time ); }

            BOOLEAN WaitAll( VOID )
                { return (AsynchronousCallback.WaitAll()); }

            ~ASYNCHRONOUS_CALLBACKS( VOID ) { /* void */ }

        private:
            //
            //  Disabled functionality.
            //
            ASYNCHRONOUS_CALLBACKS( CONST ASYNCHRONOUS_CALLBACKS & NoCopies );

            VOID operator=( CONST ASYNCHRONOUS_CALLBACKS & NoCopies );
        };
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define ASYNCHRONOUS_CALLBACKS ROCKALL::ASYNCHRONOUS_CALLBACKS
#endif
#endif