#ifndef _ROCKALL_DELETE_HPP_
#define _ROCKALL_DELETE_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "BaseTypes.hpp"
#include "DatabaseHeap.hpp"
#include "Failure.hpp"
#include "MultiThreadedHeap.hpp"
#include "PlacementNew.hpp"
#include "PublicConstants.hpp"
#include "SingleThreadedHeap.hpp"
#include "TransactionalHeap.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Rockall Delete.                                                                                   */
        /*                                                                                                    */
        /*  The standard C++ new and delete operators do not support Rockall-DB very well.  A number of       */
        /*  C++ compilers have exciting variations further compounding the situation.  A key issue            */
        /*  is that a most C++ compilers do not support multiple heaps well.  Consequently, this class        */
        /*  tries to help in this area.                                                                       */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Delete.                                                                           */
            /*                                                                                            */
            /*  The standard C++ new and delete operators do not support Rockall-DB very well.  A         */
            /*  number of C++ compilers have exciting variations further compounding the situation.       */
            /*  A key issue is that a most C++ compilers do not support multiple heaps well.              */
            /*  Consequently, this class tries to help in this area.                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> class ROCKALL_DELETE
        {
            //
            //  Private types.
            //
            typedef AUTOMATIC_HEAP_SCOPE<SINGLE_THREADED_HEAP> AUTOMATIC_HEAP_TYPE_1;
            typedef AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> AUTOMATIC_HEAP_TYPE_2;
            typedef AUTOMATIC_HEAP_SCOPE<TRANSACTIONAL_HEAP> AUTOMATIC_HEAP_TYPE_3;
            typedef AUTOMATIC_HEAP_SCOPE<DATABASE_HEAP> AUTOMATIC_HEAP_TYPE_4;
            typedef AUTOMATIC_HEAP_SCOPE<VIRTUAL_HEAP> AUTOMATIC_HEAP_TYPE_5;

        public:
            //
            //  Public functions.
            //
            ROCKALL_DELETE( VOID ) { /* void */ }

            ~ROCKALL_DELETE( VOID ) { /* void */ }

            //
            //  Public static functions.
            //
            STATIC BOOLEAN Delete
                (
                VIRTUAL_HEAP                    *Heap,
                FILE_ADDRESS                    Address,
                FILE_SIZE                       Size = 1,
                BOOLEAN                         Zero = False
                );

            STATIC BOOLEAN Delete
                (
                VIRTUAL_HEAP                    *Heap,
                TYPE                            *Data,
                FILE_SIZE                       Size = 1,
                BOOLEAN                         Zero = False
                );

            //
            //  Public static inline functions.
            //
            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_1       *Heap,
                    FILE_ADDRESS                Address,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Address,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_2       *Heap,
                    FILE_ADDRESS                Address,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Address,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_3       *Heap,
                    FILE_ADDRESS                Address,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Address,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_4       *Heap,
                    FILE_ADDRESS                Address,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Address,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_5       *Heap,
                    FILE_ADDRESS                Address,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Address,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_1       *Heap,
                    TYPE                        *Data,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Data,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_2       *Heap,
                    TYPE                        *Data,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Data,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_3       *Heap,
                    TYPE                        *Data,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Data,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_4       *Heap,
                    TYPE                        *Data,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Data,Size,Zero )); }

            STATIC INLINE BOOLEAN Delete
                    (
                    AUTOMATIC_HEAP_TYPE_5       *Heap,
                    TYPE                        *Data,
                    FILE_SIZE                   Size = 1,
                    BOOLEAN                     Zero = False
                    )
                { return (Delete( (Heap -> GetHeap()),Data,Size,Zero )); }

        private:
            //
            //  Disabled functionality.
            //
            ROCKALL_DELETE( CONST ROCKALL_DELETE & NoCopies );

            VOID operator=( CONST ROCKALL_DELETE & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete.                                                                                   */
            /*                                                                                            */
            /*  Delete an allocation.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> BOOLEAN ROCKALL_DELETE<TYPE>::Delete
            (
            VIRTUAL_HEAP                        *Heap,
            FILE_ADDRESS                        Address,
            FILE_SIZE                           Size,
            BOOLEAN                             Zero
            )
        {
        AUTO TYPE *Data;

        //
        //  When we have 'RockallTransactionalSupport' we will need to call 'Update()'
        //  to claim the required locks and get the 'Address' of the 'Data'.  If not,
        //  then we can simply cast 'Address' to 'Data' as they are known to be the same.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return ((Heap -> Update( Address,((VOID**) & Data) )) && (Delete( Heap,Data,Size,Zero ))); }
        else
            { return (Delete( Heap,((TYPE*) Address),Size,Zero )); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete.                                                                                   */
            /*                                                                                            */
            /*  Delete an allocation.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> BOOLEAN ROCKALL_DELETE<TYPE>::Delete
            (
            VIRTUAL_HEAP                        *Heap,
            TYPE                                *Data,
            FILE_SIZE                           Size,
            BOOLEAN                             Zero
            )
        {
        AUTO FILE_SIZE Count;

        //
        //  Call the destructors.
        //
        for ( Count=(Size-1);Count >= 0;Count -- )
            { PLACEMENT_DELETE( & Data[ Count ],TYPE ); }

        //
        //  Delete the existing space.  We set the 'NoDestroy' flag by default as the
        //  destructor will have alreday been executed above and it would obviously be an
        //  error if it were executed again.
        //
        if ( ! Heap -> Delete( Data,True,Zero ) )
            { FAILURE( "Delete after destructor must not fail" ); }

        return True;
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define ROCKALL_DELETE ROCKALL::ROCKALL_DELETE
#endif
#endif
