#ifndef _ROCKALL_NEW_HPP_
#define _ROCKALL_NEW_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "BaseTypes.hpp"
#include "DatabaseHeap.hpp"
#include "Failure.hpp"
#include "MultiThreadedHeap.hpp"
#include "PlacementNew.hpp"
#include "PublicConstants.hpp"
#include "SingleThreadedHeap.hpp"
#include "TransactionalHeap.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Rockall New.                                                                                      */
        /*                                                                                                    */
        /*  The standard C++ new and delete operators do not support Rockall-DB very well.  A number of       */
        /*  C++ compilers have exciting variations further compounding the situation.  A key issue            */
        /*  is that a most C++ compilers do not support multiple heaps well.  Consequently, this class        */
        /*  tries to help in this area.                                                                       */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall New.                                                                              */
            /*                                                                                            */
            /*  The standard C++ new and delete operators do not support Rockall-DB very well.  A         */
            /*  number of C++ compilers have exciting variations further compounding the situation.       */
            /*  A key issue is that a most C++ compilers do not support multiple heaps well.              */
            /*  Consequently, this class tries to help in this area.                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> class ROCKALL_NEW
        {
            //
            //  Private types.
            //
            typedef AUTOMATIC_HEAP_SCOPE<SINGLE_THREADED_HEAP> AUTOMATIC_HEAP_TYPE_1;
            typedef AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> AUTOMATIC_HEAP_TYPE_2;
            typedef AUTOMATIC_HEAP_SCOPE<TRANSACTIONAL_HEAP> AUTOMATIC_HEAP_TYPE_3;
            typedef AUTOMATIC_HEAP_SCOPE<DATABASE_HEAP> AUTOMATIC_HEAP_TYPE_4;
            typedef AUTOMATIC_HEAP_SCOPE<VIRTUAL_HEAP> AUTOMATIC_HEAP_TYPE_5;

        public:
            //
            //  Public functions.
            //
            ROCKALL_NEW( VOID ) { /* void */ }

            ~ROCKALL_NEW( VOID ) { /* void */ }

            //
            //  Public static functions.
            //
            STATIC BOOLEAN New
                (
                VIRTUAL_HEAP                    *Heap,
                FILE_ADDRESS                    *Address,
                TYPE                            **Data,
                FILE_SIZE                       Size = 1,
                FILE_SIZE                       *Space = NULL,
                BOOLEAN                         Destroy = False,
                BOOLEAN                         Zero = False
                );

            //
            //  Public static inline functions.
            //
            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_1       *Heap,
                    FILE_ADDRESS                *Address,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),Address,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_2       *Heap,
                    FILE_ADDRESS                *Address,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),Address,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_3       *Heap,
                    FILE_ADDRESS                *Address,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),Address,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_4       *Heap,
                    FILE_ADDRESS                *Address,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),Address,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_5       *Heap,
                    FILE_ADDRESS                *Address,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),Address,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    VIRTUAL_HEAP                *Heap,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( Heap,NULL,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_1       *Heap,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),NULL,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_2       *Heap,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),NULL,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_3       *Heap,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),NULL,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_4       *Heap,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),NULL,Data,Size,Space,Destroy,Zero )); }

            STATIC INLINE BOOLEAN New
                    (
                    AUTOMATIC_HEAP_TYPE_5       *Heap,
                    TYPE                        **Data,
                    FILE_SIZE                   Size = 1,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (New( (Heap -> GetHeap()),NULL,Data,Size,Space,Destroy,Zero )); }

        private:
            //
            //  Disabled functionality.
            //
            ROCKALL_NEW( CONST ROCKALL_NEW & NoCopies );

            VOID operator=( CONST ROCKALL_NEW & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New.                                                                                      */
            /*                                                                                            */
            /*  A new allocation.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> BOOLEAN ROCKALL_NEW<TYPE>::New
            (
            VIRTUAL_HEAP                        *Heap,
            FILE_ADDRESS                        *Address,
            TYPE                                **Data,
            FILE_SIZE                           Size,
            FILE_SIZE                           *Space,
            BOOLEAN                             Destroy,
            BOOLEAN                             Zero
            )
        {
        //
        //  Create new allocation.
        //
        if ( Heap -> New( Address,((VOID**) Data),(sizeof(TYPE) * Size),Space,Destroy,Zero ) )
            {
            AUTO FILE_SIZE Count = 0;

#ifndef DISABLE_CRT_FEATURES
            try
                {
#endif
                //
                //  Call the constructors.
                //
                for ( Count=0;Count < Size;Count ++ )
                    { PLACEMENT_NEW( & ((*Data)[ Count ]),TYPE ); }
#ifndef DISABLE_CRT_FEATURES
                }
            catch(...)
                {
                //
                //  Call the destructors.
                //
                for ( /* void */;Count >= 0;Count -- )
                    { PLACEMENT_DELETE( & ((*Data)[ Count ]),TYPE ); }

                if ( ! Heap -> Delete( Data ) )
                    { FAILURE( "Unable to undo 'New()'" ); }

                return False;
                }
#endif

            return True;
            }
        else
            { return False; }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define ROCKALL_NEW ROCKALL::ROCKALL_NEW
#endif
#endif
